@@ -1,5 +1,5 @@
 /***************************************************************************
- game.c Copyright (C) 1999,2000,01,02 Christoph Reichenbach
+ game.c Copyright (C) 1999 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -25,25 +25,29 @@
 
 ***************************************************************************/
 
+/* Attempt to guess if recent version of Platform SDK */
+#ifdef _WIN32
+#	pragma message("******************** IMPORTANT MESSAGE ********************")
+#	pragma message("You must have installed a recent Platform and DirectX SDK")
+#	pragma message("for this build to be successful. Download MS SDKs from:")
+#	pragma message("www.microsoft.com/msdownload/platformsdk/sdkupdate")
+#	pragma message("***********************************************************")
+#	include <windows.h>
+#	if (WINVER < 0x500)
+#		error *** BUILD FAILED: Need more recent SDKs or fix your SDK paths in Tools Options ***
+#	endif
+#endif
+
 #include <sciresource.h>
 #include <engine.h>
 #include <versions.h>
 #include <kernel.h>
 #include <kdebug.h>
 
-#ifdef HAVE_SDL
-#ifndef _MSC_VER
-#include <SDL/SDL.h>
-#else 
-#include <SDL.h>
-#endif
-#endif
-
-#if !defined (_WIN32) && !defined (__BEOS__)
+#if !defined (_WIN32) && !defined (_DREAMCAST)
 #include <sys/resource.h>
 #endif
 
-
 /* Structures and data from vm.c: */
 extern calls_struct_t *send_calls;
 extern int send_calls_allocated;
@@ -60,10 +64,9 @@ _init_vocabulary(state_t *s) /* initialize vocabulary and related resources */
 
 	if ((s->parser_words = vocab_get_words(s->resmgr, &(s->parser_words_nr)))) {
 		s->parser_suffices = vocab_get_suffices(s->resmgr, &(s->parser_suffices_nr));
-		s->parser_branches = vocab_get_branches(s->resmgr, &(s->parser_branches_nr));
-
-		/* Now build a GNF grammar out of this */
-		s->parser_rules = vocab_build_gnf(s->parser_branches, s->parser_branches_nr);
+		if ((s->parser_branches = vocab_get_branches(s->resmgr, &(s->parser_branches_nr))))
+		    /* Now build a GNF grammar out of this */
+		    s->parser_rules = vocab_build_gnf(s->parser_branches, s->parser_branches_nr);
 
 	} else {
 		sciprintf("Assuming that this game does not use a parser.\n");
@@ -127,6 +130,7 @@ _reset_graphics_input(state_t *s)
 	resource_t *resource;
 	int font_nr;
 	gfx_color_t transparent	= {0};
+
 	sciprintf("Initializing graphics\n");
 
 	if (s->resmgr->sci_version <= SCI_VERSION_01) {
@@ -161,7 +165,10 @@ _reset_graphics_input(state_t *s)
 	s->drop_views = NULL; /* And, consequently, no list for dropped views */
 
 	s->priority_first = 42; /* Priority zone 0 ends here */
-	s->priority_last = 200; /* The highest priority zone (15) starts here */
+
+	if (s->version < SCI_VERSION_FTU_PRIORITY_14_ZONES)
+		s->priority_last = 200; else
+		s->priority_last = 190;
 
 	font_nr = -1;
 	do {
@@ -175,7 +182,9 @@ _reset_graphics_input(state_t *s)
 
 	s->visual = gfxw_new_visual(s->gfx_state, font_nr);
 
-	s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
+	if (s->version < SCI_VERSION_FTU_NORMAL_PORTS)
+		s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 0, 320, 200), s->ega_colors[0], transparent); else
+			s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
 	s->titlebar_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 0, 320, 10), s->ega_colors[0], s->ega_colors[15]);
 	s->picture_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
 	s->titlebar_port->color.mask |= GFX_MASK_PRIORITY;
@@ -203,6 +212,12 @@ _reset_graphics_input(state_t *s)
 int
 game_init_graphics(state_t *s)
 {
+#ifndef WITH_PIC_SCALING
+	if (s->gfx_state->options->pic0_unscaled == 0)
+		sciprintf("WARNING: Pic scaling was disabled; your version of FreeSCI has no support for scaled pic drawing built in.\n");
+
+	s->gfx_state->options->pic0_unscaled = 1;
+#endif
 	return _reset_graphics_input(s);
 }
 
@@ -230,7 +245,7 @@ static int
 suggested_script(resource_t *res, unsigned int class)
 {
 	int offset;
-	if (class >= res->size >> 2)
+	if (!res || class >= res->size >> 2)
 		return -1;
 
 	offset = 2 + (class << 2);
@@ -340,7 +355,7 @@ script_init_engine(state_t *s, sci_version_t version)
 
 		}
 	}
-	scir_unlock_resource(s->resmgr, vocab996);
+	scir_unlock_resource(s->resmgr, vocab996, sci_vocab, 996);
 	vocab996 = NULL;
 
 	s->_heap = heap_new();
@@ -490,7 +505,7 @@ game_init(state_t *s)
 	heap_ptr game_obj; /* Address of the game object */
 	int i;
 
-	if (!script_instantiate(s, 0, 0)) {
+	if (!script_instantiate(s, 0, 1)) {
 		sciprintf("game_init(): Could not instantiate script 0\n");
 		return 1;
 	}
@@ -536,6 +551,8 @@ game_init(state_t *s)
 
 	s->successor = NULL; /* No successor */
 	s->status_bar_text = NULL; /* Status bar is blank */
+	s->status_bar_foreground = 0;
+	s->status_bar_background = 15;
 
 	fprintf(stderr," Script 0 at %04x\n", script0);
 
@@ -545,6 +562,10 @@ game_init(state_t *s)
 	/* Global variables are script 0's local variables */
 
 	sci_get_current_time(&(s->game_start_time)); /* Get start time */
+#ifdef ARM_WINCE	
+	s->game_start_time.tv_usec = 0;
+	s->game_start_time.tv_sec = 0;
+#endif	
 	memcpy(&(s->last_wait_time), &(s->game_start_time), sizeof(GTimeVal));
 	/* Use start time as last_wait_time */
 
@@ -553,8 +574,8 @@ game_init(state_t *s)
 
 	srand(time(NULL)); /* Initialize random number generator */
 
-	memset(s->hunk, sizeof(s->hunk), 0); /* Sets hunk to be unused */
-	memset(s->clone_list, sizeof(s->clone_list), 0); /* No clones */
+	memset(s->hunk, 0, sizeof(s->hunk)); /* Sets hunk to be unused */
+	memset(s->clone_list, 0, sizeof(s->clone_list)); /* No clones */
 
 	/*	script_dissect(0, s->selector_names, s->selector_names_nr); */
 	game_obj = script0 + GET_HEAP(s->scripttable[0].export_table_offset + 2);
@@ -634,10 +655,3 @@ game_exit(state_t *s)
 	return 0;
 }
 
-void
-other_libs_exit()
-{
-#ifdef HAVE_SDL
-	SDL_Quit();
-#endif
-}
