@@ -1,5 +1,5 @@
 /***************************************************************************
- Kfile.c Copyright (C) 1999,2000,01,02 Christoph Reichenbach
+ Kfile.c Copyright (C) 1999 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -33,13 +33,22 @@
 #  endif
 #  define WIN32_LEAN_AND_MEAN
 #  include <windows.h>
-#else
+#elif defined (__MORPHOS__)
+#  include <sys/stat.h>
+#elif defined (_DREAMCAST)
+#  include <dc.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
 #  include <unistd.h>
+#endif
+
+#ifdef HAVE_DIRENT_H
 #  include <dirent.h>
 #endif
 
 #ifndef O_BINARY
-#define O_BINARY 0
+#  define O_BINARY 0
 #endif
 
 static int _savegame_indices_nr = -1; /* means 'uninitialized' */
@@ -61,45 +70,41 @@ f_open_mirrored(state_t *s, char *fname)
 {
 	int fd;
 	char *buf = NULL;
-	struct stat fstate;
+	int fsize;
 
 	chdir(s->resource_dir);
 	fd = sci_open(fname, O_RDONLY | O_BINARY);
-	if (!fd) {
+	if (!IS_VALID_FD(fd)) {
 		chdir(s->work_dir);
 		return NULL;
 	}
 
-	fstat(fd, &fstate);
-	if (fstate.st_size) {
-		buf = sci_malloc(fstate.st_size);
-		read(fd, buf, fstate.st_size);
+	fsize = sci_fd_size(fd);
+	if (fsize > 0) {
+		buf = sci_malloc(fsize);
+		read(fd, buf, fsize);
 	}
 
 	close(fd);
 
 	chdir(s->work_dir);
 
-	/* Visual C++ doesn't allow to specify O_BINARY with creat() */
-#ifdef _MSC_VER
-	fd = _open(fname, _O_CREAT | _O_BINARY | _O_RDWR, _S_IREAD | _S_IWRITE);
-#else
-	fd = creat(fname, 0600);
-#endif
+	fd = sci_create_file(fname);
 
-	if (!fd && buf) {
+
+	if (!IS_VALID_FD(fd) && buf) {
 		free(buf);
 		sciprintf("kfile.c: f_open_mirrored(): Warning: Could not create '%s' in '%s' (%d bytes to copy)\n",
-			  fname, s->work_dir, fstate.st_size);
+			  fname, s->work_dir, fsize);
 		return NULL;
 	}
 
-	if (fstate.st_size) {
+	if (fsize > 0) {
 		int ret;
-		ret = write(fd, buf, fstate.st_size);
-		if (ret < fstate.st_size) {
+		ret = write(fd, buf, fsize);
+		if (ret < fsize) {
 			sciprintf("kfile.c: f_open_mirrored(): Warning: Could not write all %ld bytes to '%s' in '%s' (only wrote %ld)\n",
-				  (long)fstate.st_size, fname, s->work_dir, ret);
+				  (long)fsize, fname, s->work_dir, ret);
 		}
 
 		free(buf);
@@ -111,8 +116,8 @@ f_open_mirrored(state_t *s, char *fname)
 }
 
 
-#define _K_FILE_MODE_OPEN_OR_FAIL 0
-#define _K_FILE_MODE_OPEN_OR_CREATE 1
+#define _K_FILE_MODE_OPEN_OR_CREATE 0
+#define _K_FILE_MODE_OPEN_OR_FAIL 1
 #define _K_FILE_MODE_CREATE 2
 
 
@@ -142,7 +147,7 @@ file_open(state_t *s, char *filename, int mode)
 	}
 	if (!file) { /* Failed */
 		SCIkdebug(SCIkFILE, "file_open() failed\n");
-		s->acc = 0;
+		s->acc = 0xffff;
 		return;
 	}
 
@@ -332,8 +337,7 @@ kFGets(state_t *s, int funct_nr, int argc, heap_ptr argp)
   int handle = UPARAM(2);
 
   fgets_wrapper(s, dest, maxsize, handle);
-  s->acc=UPARAM(0);
-
+  s->acc = UPARAM(0);
 }
 
 
@@ -343,11 +347,19 @@ kFGets(state_t *s, int funct_nr, int argc, heap_ptr argp)
 void
 kGetCWD(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
-  heap_ptr offset = UPARAM(0);
-  char *targetaddr = (char *) s->heap + offset;
+	char *wd = sci_getcwd();
+	heap_ptr offset = UPARAM(0);
+	char *targetaddr = (char *) s->heap + offset;
 
-  s->acc = offset;
-  getcwd(targetaddr, PATH_MAX + 1);
+	s->acc = offset;
+	strncpy(targetaddr, wd, MAX_SAVE_DIR_SIZE - 1);
+	targetaddr[MAX_SAVE_DIR_SIZE - 1] = 0; /* Terminate */
+
+	SCIkdebug(SCIkFILE, "Copying cwd='%s'(%d chars) to %p"
+		  " (heap starts at %p, offset=%04x)\n",
+		  wd, strlen(wd), targetaddr, s->heap, offset);
+
+	free(wd);
 }
 
 #define K_DEVICE_INFO_GET_DEVICE 0
@@ -363,7 +375,6 @@ kDeviceInfo_Win32(state_t *s, int funct_nr, int argc, heap_ptr argp)
   char dir_buffer [MAX_PATH], dir_buffer2 [MAX_PATH];
   int mode = UPARAM(0);
 
-  CHECK_THIS_KERNEL_FUNCTION;
 
   switch(mode) {
 
@@ -438,7 +449,6 @@ kDeviceInfo_Unix(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
   int mode = UPARAM(0);
 
-  CHECK_THIS_KERNEL_FUNCTION;
 
   switch(mode) {
 
@@ -503,6 +513,17 @@ kGetSaveDir(state_t *s, int funct_nr, int argc, heap_ptr argp)
   s->acc = s->save_dir + 2; /* +2 to step over heap block size */
 }
 
+#if defined(_DREAMCAST) || defined(ARM_WINCE) || defined(_GP32)
+
+void
+kCheckFreeSpace(state_t *s, int funct_nr, int argc, heap_ptr argp)
+{
+	/* There is no practical way to detect enough memory card space,
+	so we just attempt to save no matter what. */
+	s->acc = 1;
+}
+
+#else /* ! defined(_DREAMCAST) && !defined(ARM_WINCE) && !defined(_GP32) */
 
 void
 kCheckFreeSpace(state_t *s, int funct_nr, int argc, heap_ptr argp)
@@ -515,11 +536,12 @@ kCheckFreeSpace(state_t *s, int funct_nr, int argc, heap_ptr argp)
   int failed = 0;
   int pathlen;
 
+  memset(buf, 0, sizeof(buf));
   strcpy(testpath, path);
   strcat(testpath, "freesci.foo");
   pathlen = strlen(testpath);
 
-  while ((fd = open(testpath, O_RDONLY)) > -1) {
+  while (IS_VALID_FD(fd = open(testpath, O_RDONLY))) {
     close(fd);
     if (testpath[pathlen - 2] == 'z') { /* Failed. */
       SCIkwarn(SCIkWARNING, "Failed to find non-existing file for free space test\n");
@@ -540,9 +562,9 @@ kCheckFreeSpace(state_t *s, int funct_nr, int argc, heap_ptr argp)
       ++testpath[pathlen - 1];
   }
 
-  fd = creat(testpath, 0600);
+  fd = sci_create_file(testpath);
 
-  if (fd == -1) {
+  if (!IS_VALID_FD(fd)) {
     SCIkwarn(SCIkWARNING,"Could not test for disk space: %s\n", strerror(errno));
     SCIkwarn(SCIkWARNING,"Test path was '%s'\n", testpath);
     free(testpath);
@@ -563,8 +585,19 @@ kCheckFreeSpace(state_t *s, int funct_nr, int argc, heap_ptr argp)
   free(testpath);
 }
 
+#endif
 
 /* Returns a dynamically allocated pointer to the name of the requested save dir */
+#ifdef _DREAMCAST
+
+char *
+_k_get_savedir_name(int nr)
+{
+	return strdup("/ram");
+}
+
+#else /* !_DREAMCAST */
+
 char *
 _k_get_savedir_name(int nr)
 {
@@ -579,17 +612,13 @@ _k_get_savedir_name(int nr)
 	return savedir_name;
 }
 
+#endif
 
 int
 _k_check_file(char *filename, int minfilesize)
      /* Returns 0 if the file exists and is big enough */
 {
-	struct stat state;
-
-	if (stat(filename, &state))
-		return 1;
-
-	return (state.st_size < minfilesize);
+	return (sci_file_size(filename) < minfilesize);
 }
 
 int
@@ -618,7 +647,7 @@ _k_find_savegame_by_name(char *game_id_file, char *name)
 				fclose(idfile);
 			}
 
-			chdir("..");
+			chdir(G_DIR_PARENT_S);
 		}
 		free(buf);
 	}
@@ -656,7 +685,7 @@ test_savegame(state_t *s, char *savegame_id, char *savegame_name, int savegame_n
 		} else if (_k_check_file(game_id_file, 1))
 			retval = 0;
 
-		chdir ("..");
+		chdir (G_DIR_PARENT_S);
 	}
 	free(game_id_file);
 	return retval;
@@ -671,7 +700,6 @@ kCheckSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	TEST_DIR_OR_QUIT(workdir);
 
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (soundserver_dead) {
 		sciprintf("Soundserver is dead- cannot save game state!");
@@ -680,12 +708,24 @@ kCheckSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		return;
 	}
 
-	if (savedir_nr > MAX_SAVEGAME_NR-1) {
+	if ((savedir_nr < 0) || (savedir_nr > MAX_SAVEGAME_NR-1)) {
 		_chdir_restoredir(workdir);
 		s->acc = 0;
 		return;
 	}
 
+	savedir_nr = _savegame_indices[savedir_nr].id;
+
+	if ((savedir_nr < 0) || (savedir_nr > MAX_SAVEGAME_NR-1)) {
+		_chdir_restoredir(workdir);
+		s->acc = 0;
+		return;
+	}
+
+#ifdef _DREAMCAST
+	dc_retrieve_savegame((char *) s->game_name, savedir_nr);
+#endif
+
 	s->acc = test_savegame(s, (buf = _k_get_savedir_name(savedir_nr)), NULL, 0);
 
 	_chdir_restoredir(workdir);
@@ -693,6 +733,16 @@ kCheckSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 }
 
 
+#ifdef _DREAMCAST
+static long
+get_file_mtime(int fd)
+{
+	/* FIXME (Dreamcast): Not yet implemented */
+	return 0;
+}
+
+#else
+
 #define get_file_mtime(fd) get_file_mtime_Unix(fd)
 /* Returns the time of the specified file's last modification
 ** Parameters: (int) fd: The file descriptor of the file in question
@@ -713,7 +763,7 @@ get_file_mtime_Unix(int fd) /* returns the  */
 
 	return fd_stat.st_ctime;
 }
-
+#endif
 
 static int
 _savegame_index_struct_compare(const void *a, const void *b)
@@ -722,11 +772,56 @@ _savegame_index_struct_compare(const void *a, const void *b)
 		- ((struct _savegame_index_struct *)a)->timestamp;
 }
 
+#ifdef _DREAMCAST
+
 static void
-update_savegame_indices(char *gfname)
+update_savegame_indices(char *game_id)
 {
 	int i;
-	int gfname_len = strlen(gfname);
+	char *vmu;
+
+	_savegame_indices_nr = 0;
+
+	if (!(vmu = dc_get_first_vmu())) {
+		sciprintf("%s, L%d: No VMU found!\n", __FILE__, __LINE__);
+		return;
+	}
+
+	fs_chdir(vmu);
+	sci_free(vmu);
+
+	for (i = 0; i < MAX_SAVEGAME_NR; i++) {
+		char *catname = dc_get_cat_name(game_id, i);
+		sci_dir_t dirent;
+		sci_init_dir(&dirent);
+
+		if (sci_find_first(&dirent, catname)) {
+			_savegame_indices[_savegame_indices_nr].id = i;
+			_savegame_indices[_savegame_indices_nr++].timestamp = 0;
+		}
+
+		sci_finish_find(&dirent);
+
+		sci_free(catname);
+	}
+
+	fs_chdir("/ram");
+
+	qsort(_savegame_indices, _savegame_indices_nr, sizeof(struct _savegame_index_struct),
+	      _savegame_index_struct_compare);
+
+}
+
+#else /* !_DREAMCAST */
+
+static void
+update_savegame_indices(char *game_id)
+{
+	int i;
+	char *gfname = sci_malloc(strlen(game_id) + strlen(FREESCI_ID_SUFFIX) + 1);
+
+	strcpy(gfname, game_id);
+	strcat(gfname, FREESCI_ID_SUFFIX);
 
 	_savegame_indices_nr = 0;
 
@@ -736,22 +831,25 @@ update_savegame_indices(char *gfname)
 
 		if (!chdir(dirname)) {
 
-			if ((fd = sci_open(gfname, O_RDONLY)) > 0) {
+			if (IS_VALID_FD(fd = sci_open(gfname, O_RDONLY))) {
 				_savegame_indices[_savegame_indices_nr].id = i;
 				_savegame_indices[_savegame_indices_nr++].timestamp = get_file_mtime(fd);
 				close(fd);
 			}
-			chdir("..");
+			chdir(G_DIR_PARENT_S);
 		}
 
 		free(dirname);
 	}
 
+	sci_free(gfname);
+
 	qsort(_savegame_indices, _savegame_indices_nr, sizeof(struct _savegame_index_struct),
 	      _savegame_index_struct_compare);
 
 }
 
+#endif
 
 void
 kGetSaveFiles(state_t *s, int funct_nr, int argc, heap_ptr argp)
@@ -768,9 +866,7 @@ kGetSaveFiles(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	strcpy(gfname, game_id);
 	strcat(gfname, FREESCI_ID_SUFFIX); /* This file is used to identify in-game savegames */
 
-	update_savegame_indices(gfname);
-
-	CHECK_THIS_KERNEL_FUNCTION;
+	update_savegame_indices(game_id);
 
 	SCIkASSERT(UPARAM(0) >= 800);
 	SCIkASSERT(nametarget >= 800);
@@ -782,6 +878,9 @@ kGetSaveFiles(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		char *savedir_name = _k_get_savedir_name(_savegame_indices[i].id);
 		FILE *idfile;
 
+#ifdef _DREAMCAST
+		dc_retrieve_savegame(game_id, _savegame_indices[i].id);
+#endif
 		if (!chdir(savedir_name)) {
 
 
@@ -804,7 +903,7 @@ kGetSaveFiles(state_t *s, int funct_nr, int argc, heap_ptr argp)
 					fclose(idfile);
 				}
 			}
-			chdir("..");
+			chdir(G_DIR_PARENT_S);
 		}
 		free(savedir_name);
 	}
@@ -839,7 +938,7 @@ kSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	strcpy(game_id_file_name, game_id);
 	strcat(game_id_file_name, FREESCI_ID_SUFFIX);
 
-	update_savegame_indices(game_id_file_name);
+	update_savegame_indices(game_id);
 
 	if (savedir_nr >= 0 && savedir_nr < _savegame_indices_nr)
 		/* Overwrite */
@@ -872,9 +971,13 @@ kSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 	savegame_dir = _k_get_savedir_name(savedir_id);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 	s->acc = 1;
 
+#ifdef _GP32
+	/* Write save files without updating FAT to speed up the process. */
+	smNoFATUpdate("dev0:\\");
+#endif
+
 	if (gamestate_save(s, savegame_dir)) {
 		sciprintf("Saving the game failed.\n");
 		s->acc = 0;
@@ -894,9 +997,19 @@ kSaveGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 			s->acc = 0;
 		}
 
-		chdir ("..");
+		chdir (G_DIR_PARENT_S);
 	}
 	free(game_id_file_name);
+
+#ifdef _DREAMCAST
+	if (dc_store_savegame(game_id, game_description, savedir_id) < 0)
+		s->acc = 0;
+#elif defined(_GP32)
+	/* Update FAT. */
+	smFATUpdate("dev0:\\");
+#endif
+
+	sci_free(savegame_dir);
 	_chdir_restoredir(workdir);
 }
 
@@ -910,17 +1023,12 @@ kRestoreGame(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	TEST_DIR_OR_QUIT(workdir);
 
 	if (_savegame_indices_nr < 0) {
-		char *game_id_file_name = sci_malloc(strlen(game_id) + strlen(FREESCI_ID_SUFFIX) + 1);
-
-		strcpy(game_id_file_name, game_id);
-		strcat(game_id_file_name, FREESCI_ID_SUFFIX);
 		SCIkwarn(SCIkWARNING, "Savegame index list not initialized!\n");
-		update_savegame_indices(game_id_file_name);
+		update_savegame_indices(game_id);
 	}
 
 	savedir_nr = _savegame_indices[savedir_nr].id;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (savedir_nr > -1) {
 		char *savedir_name = _k_get_savedir_name(savedir_nr);
@@ -965,7 +1073,7 @@ kValidPath(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	chdir(cpath);
 }
 
-#define K_FILEIO_OPEN 		0
+#define K_FILEIO_OPEN		0
 #define K_FILEIO_CLOSE		1
 #define K_FILEIO_READ_RAW	2
 #define K_FILEIO_WRITE_RAW	3
@@ -1009,7 +1117,7 @@ first_file(state_t *s, char *dir, char *mask, heap_ptr buffer)
 		return;
 	}
 
-	if (strcmp(dir, ".")) {
+	if (strcmp(dir, G_DIR_CURRENT_S)) {
 		sciprintf("%s L%d: Non-local first_file: Not implemented yet\n",
 			  __FILE__, __LINE__);
 		s->acc = 0;
@@ -1033,7 +1141,6 @@ kFileIO(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
   int func_nr = UPARAM(0);
 
-  CHECK_THIS_KERNEL_FUNCTION;
 
   switch (func_nr) {
 
@@ -1113,7 +1220,7 @@ kFileIO(state_t *s, int funct_nr, int argc, heap_ptr argp)
 #ifndef _WIN32
 	if (strcmp(mask, "*.*")==0) strcpy(mask, "*"); /* For UNIX */
 #endif
-	first_file(s, ".", mask, buf);
+	first_file(s, G_DIR_CURRENT_S, mask, buf);
 
 	break;
     }
