@@ -1,5 +1,5 @@
 /***************************************************************************
- kgraphics.c Copyright (C) 1999,2000,01,02 Christoph Reichenbach
+ kgraphics.c Copyright (C) 1999,2000,01 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -25,6 +25,9 @@
 
 ***************************************************************************/
 
+#define PIC_IS_NEW_PIC 1
+#define PIC_IS_NEW_PORT 2
+
 #include <sciresource.h>
 #include <engine.h>
 #include <gfx_widgets.h>
@@ -81,11 +84,6 @@
   if (s->visual) \
        s->visual->print(GFXW(s->visual), 0);
 
-static inline void
-handle_gfx_error(int line, char *file)
-{
-	exit(1); /* FIXME: Replace this with a longjmp() eventually */
-}
 
 #define GFX_ASSERT(x) { \
 	int val = !!(x); \
@@ -94,8 +92,7 @@ handle_gfx_error(int line, char *file)
 			SCIkwarn(SCIkWARNING, "GFX subsystem returned error on \"" #x "\"!\n"); \
 		else {\
                         SCIkwarn(SCIkERROR, "GFX subsystem fatal error condition on \"" #x "\"!\n"); \
-                        BREAKPOINT(); \
-			handle_gfx_error(__LINE__, __FILE__); \
+			vm_handle_fatal_error(s, __LINE__, __FILE__); \
 	        } \
         }\
 }
@@ -103,9 +100,9 @@ handle_gfx_error(int line, char *file)
 #define ASSERT(x) { \
        int val = !!(x); \
        if (!val) { \
-                        SCIkwarn(SCIkERROR, "GFX subsystem fatal error condition on \"" #x "\"!\n"); \
+                        SCIkwarn(SCIkERROR, "Fatal error condition on \"" #x "\"!\n"); \
                         BREAKPOINT(); \
-			handle_gfx_error(__LINE__, __FILE__); \
+			vm_handle_fatal_error(s, __LINE__, __FILE__); \
        } \
 }
 
@@ -153,20 +150,26 @@ reparentize_primary_widget_lists(state_t *s, gfxw_port_t *newport)
 	}
 }
 
-inline int
+int
 _find_view_priority(state_t *s, int y)
 {
-	if (s->version <= SCI_VERSION_LTU_PRIORITY_OB1)
-		++y;
+  /*        if (s->version <= SCI_VERSION_LTU_PRIORITY_OB1)
+	    ++y; */
 
 	if (s->pic_priority_table) { /* SCI01 priority table set? */
 		int j;
 		for (j = 0; j < 15; j++)
-			if (y < s->pic_priority_table[j])
+			if (y < s->pic_priority_table[j+1])
 				return j;
-		return 15; /* Maximum */
+		return 14; /* Maximum */
 	} else
-		return SCI0_VIEW_PRIORITY(y);
+	{
+		if (s->version >= SCI_VERSION_FTU_PRIORITY_14_ZONES)
+			return SCI0_VIEW_PRIORITY_14_ZONES(y);
+		else
+			return SCI0_VIEW_PRIORITY(y);
+		
+	}
 }
 
 inline int
@@ -183,9 +186,15 @@ _find_priority_band(state_t *s, int nr)
 	if (s->pic_priority_table) /* SCI01 priority table set? */
 		return s->pic_priority_table[nr];
 	else {
-		int retval = SCI0_PRIORITY_BAND_FIRST(nr);
-		if (s->version <= SCI_VERSION_LTU_PRIORITY_OB1)
-			--retval;
+		int retval;
+
+		if (s->version >= SCI_VERSION_FTU_PRIORITY_14_ZONES)
+			retval = SCI0_PRIORITY_BAND_FIRST_14_ZONES(nr);
+		else
+			retval = SCI0_PRIORITY_BAND_FIRST(nr);
+
+		/*		if (s->version <= SCI_VERSION_LTU_PRIORITY_OB1)
+				--retval; */
 		return retval;
 	}
 }
@@ -277,7 +286,7 @@ void
 kSetCursor(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
 
-	if (PARAM(1)) {
+	if (PARAM_OR_ALT(1,1)) {
 		s->mouse_pointer_nr = PARAM(0);
 	} else
 		s->mouse_pointer_nr = GFXOP_NO_POINTER;
@@ -302,9 +311,31 @@ _ascertain_port_contents(gfxw_port_t *port)
 void
 kShow(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
-	s->pic_visible_map = sci_ffs(UPARAM_OR_ALT(0, 1)) - 1;
+	int old_map = s->pic_visible_map;
+
+	s->pic_visible_map = UPARAM_OR_ALT(0, 1);
+
+	switch (s->pic_visible_map) {
+
+	case GFX_MASK_VISUAL:
+	case GFX_MASK_PRIORITY:
+	case GFX_MASK_CONTROL:
+		gfxop_set_visible_map(s->gfx_state, s->pic_visible_map);
+		if (old_map != s->pic_visible_map) {
+
+			if (s->pic_visible_map == GFX_MASK_VISUAL) /* Full widget redraw */
+				s->visual->draw(GFXW(s->visual), gfx_point(0,0));
+
+			gfxop_update(s->gfx_state);
+			sciprintf("Switching visible map to %x\n", s->pic_visible_map);
+		}
+		break;
+
+	default:
+		SCIkwarn(SCIkWARNING, "Show(%x) selects unknown map\n", s->pic_visible_map);
+
+	}
 
-	CHECK_THIS_KERNEL_FUNCTION;
 	s->pic_not_valid = 2;
 }
 
@@ -312,7 +343,6 @@ kShow(state_t *s, int funct_nr, int argc, heap_ptr argp)
 void
 kPicNotValid(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
-	CHECK_THIS_KERNEL_FUNCTION;
 	s->acc = s->pic_not_valid;
 	if (argc)
 		s->pic_not_valid = (byte)PARAM(0);
@@ -429,7 +459,6 @@ kGraph(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		port = s->port;
 
 		redraw_port = 1;
-		CHECK_THIS_KERNEL_FUNCTION;
 		break;
 
 	case K_GRAPH_FILL_BOX_FOREGROUND:
@@ -438,7 +467,6 @@ kGraph(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		port = s->port;
 
 		redraw_port = 1;
-		CHECK_THIS_KERNEL_FUNCTION;
 		break;
 
 	case K_GRAPH_FILL_BOX_ANY: {
@@ -452,7 +480,6 @@ kGraph(state_t *s, int funct_nr, int argc, heap_ptr argp)
 			  UPARAM(5));
 
 		ADD_TO_CURRENT_BG_WIDGETS(gfxw_new_box(s->gfx_state, area, color, color, GFX_BOX_SHADE_FLAT));
-		CHECK_THIS_KERNEL_FUNCTION;
 
 	}
 	break;
@@ -466,14 +493,12 @@ kGraph(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		area.y += s->port->zone.y;
 
 		gfxop_update_box(s->gfx_state, area);
-		CHECK_THIS_KERNEL_FUNCTION;
 
 	}
 	break;
 
 	case K_GRAPH_REDRAW_BOX: {
 
-		CHECK_THIS_KERNEL_FUNCTION;
 
 		SCIkdebug(SCIkGRAPHICS, "redraw_box(%d, %d, %d, %d)\n",
 			  PARAM(1), PARAM(2), PARAM(3), PARAM(4));
@@ -499,7 +524,6 @@ kGraph(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 	default:
 
-		CHECK_THIS_KERNEL_FUNCTION;
 		SCIkdebug(SCIkSTUB, "Unhandled Graph() operation %04x\n", PARAM(0));
 
 	}
@@ -554,7 +578,7 @@ kWait(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 	sci_get_current_time (&time);
 
-	s->acc = ((time.tv_usec - s->last_wait_time.tv_usec) * 60 / 1000000) +
+	s->acc = (guint16)((time.tv_usec - s->last_wait_time.tv_usec) * 60 / 1000000) +
 		(time.tv_sec - s->last_wait_time.tv_sec) * 60;
 
 	memcpy(&(s->last_wait_time), &time, sizeof(GTimeVal));
@@ -755,8 +779,8 @@ kCanBeHere(state_t *s, int funct_nr, int argc, heap_ptr argp)
                         if (widget->ID
 			    && (widget->signal & _K_VIEW_SIG_FLAG_FREESCI_STOPUPD)
 			    && (widget->ID != obj)
-                            && is_object(s, widget->ID))
-                                if (collides_with(s, abs_zone, widget->ID, 1, GASEOUS_VIEW_MASK_ACTIVE, funct_nr, argc, argp))
+                            && is_object(s, (heap_ptr)widget->ID))
+                                if (collides_with(s, abs_zone, (heap_ptr)widget->ID, 1, GASEOUS_VIEW_MASK_ACTIVE, funct_nr, argc, argp))
                                         return;
 
                         widget = (gfxw_dyn_view_t *) widget->next;
@@ -846,7 +870,6 @@ kNumLoops(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int view = GET_SELECTOR(obj, view);
 	int loops_nr = gfxop_lookup_view_get_loops(s->gfx_state, view);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (loops_nr < 0) {
 		SCIkwarn(SCIkERROR, "view.%d (0x%x) not found\n", view, view);
@@ -867,7 +890,6 @@ kNumCels(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int view = GET_SELECTOR(obj, view);
 	int cel = 0xffff;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (gfxop_check_cel(s->gfx_state, view, &loop, &cel)) { /* OK, this is a hack and there's a
 							       ** real function to calculate cel numbers... */
@@ -887,7 +909,6 @@ kOnControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int map, xstart, ystart;
 	int xlen = 1, ylen = 1;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (argc == 2 || argc == 4)
 		map = 4;
@@ -902,14 +923,9 @@ kOnControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	if (argc > 3) {
 		ylen = PARAM(arg+3) - ystart;
 		xlen = PARAM(arg+2) - xstart;
-		if (xlen > 1)
-			--xlen;
-		if (ylen > 1)
-			--ylen;
 	}
 
 	s->acc = gfxop_scan_bitmask(s->gfx_state, gfx_rect(xstart, ystart + 10, xlen, ylen), map);
-
 }
 
 void
@@ -925,7 +941,6 @@ kDrawPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int palette = PARAM_OR_ALT(3, 0);
 	gfx_color_t transparent = s->wm_port->bgcolor;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (s->version < SCI_VERSION_FTU_NEWER_DRAWPIC_PARAMETERS) {
 		if (!PARAM_OR_ALT(2, 0))
@@ -936,6 +951,11 @@ kDrawPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 
 	gfxop_disable_dirty_frames(s->gfx_state);
+
+	if (NULL != s->old_screen) {
+		GFX_ASSERT(gfxop_free_pixmap(s->gfx_state, s->old_screen));
+	}
+
 	s->old_screen = gfxop_grab_pixmap(s->gfx_state, gfx_rect(0, 10, 320, 190));
 
 	SCIkdebug(SCIkGRAPHICS,"Drawing pic.%03d\n", PARAM(0));
@@ -963,7 +983,9 @@ kDrawPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	s->wm_port->widfree(GFXW(s->wm_port));
 	s->picture_port->widfree(GFXW(s->picture_port));
 
-	s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
+	if (s->version < SCI_VERSION_FTU_NORMAL_PORTS)
+		s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 0, 320, 200), s->ega_colors[0], transparent); else
+		s->wm_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
 	s->picture_port = gfxw_new_port(s->visual, NULL, gfx_rect(0, 10, 320, 190), s->ega_colors[0], transparent);
 
 	s->visual->add(GFXWC(s->visual), GFXW(s->picture_port));
@@ -992,9 +1014,14 @@ kDrawPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	s->drop_views = NULL;
 
 	s->priority_first = 42;
-	s->priority_last = 190;
+
+	if (s->version < SCI_VERSION_FTU_PRIORITY_14_ZONES)
+		s->priority_last = 200;
+	else
+		s->priority_last = 190;
+
 	s->pic_not_valid = 1;
-	s->pic_is_new = 1;
+	s->pic_is_new = PIC_IS_NEW_PIC;
 
 }
 
@@ -1007,6 +1034,7 @@ set_base(state_t *s, heap_ptr object)
 	int x, y, original_y, z, ystep, xsize, ysize;
 	int xbase, ybase, xend, yend;
 	int view, loop, cel;
+	int oldloop, oldcel;
 	int xmod = 0, ymod = 0;
 	abs_rect_t retval;
 
@@ -1033,14 +1061,25 @@ set_base(state_t *s, heap_ptr object)
 
 
 	view = GET_SELECTOR(object, view);
-	loop = sign_extend_byte(GET_SELECTOR(object, loop));
-	cel = sign_extend_byte(GET_SELECTOR(object, cel));
+	oldloop = loop = sign_extend_byte(GET_SELECTOR(object, loop));
+	oldcel = cel = sign_extend_byte(GET_SELECTOR(object, cel));
 
 	if (gfxop_check_cel(s->gfx_state, view, &loop, &cel)) {
 		xsize = ysize = xmod = ymod = 0;
 	} else {
 		point_t offset = gfx_point(0, 0);
 
+		if (loop != oldloop) {
+			loop = 0;
+			PUT_SELECTOR(object, loop, 0);
+			fprintf(stderr, "Resetting loop for %04x!\n", object);
+		}
+
+		if (cel != oldcel) {
+			cel = 0;
+			PUT_SELECTOR(object, cel, 0);
+		}
+
 		gfxop_get_cel_parameters(s->gfx_state, view, loop, cel,
 					 &xsize, &ysize, &offset);
 
@@ -1048,6 +1087,7 @@ set_base(state_t *s, heap_ptr object)
 		ymod = offset.y;
 	}
 
+
 	xbase = x - xmod - (xsize >> 1);
 	xend = xbase + xsize;
 	yend = y /*- ymod*/ + 1;
@@ -1089,7 +1129,6 @@ kBaseSetter(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
 	heap_ptr object = PARAM(0);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	_k_base_setter(s, object);
 
@@ -1200,7 +1239,6 @@ kSetNowSeen(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 	heap_ptr object = PARAM(0);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 	_k_set_now_seen(s, object);
 
 } /* kSetNowSeen */
@@ -1217,7 +1255,6 @@ kDrawControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
 	heap_ptr obj = UPARAM(0);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	_k_draw_control(s, obj, 0);
 	FULL_REDRAW();
@@ -1229,7 +1266,6 @@ kHiliteControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
 	heap_ptr obj = UPARAM(0);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	_k_draw_control(s, obj, 1);
 }
@@ -1269,7 +1305,6 @@ kEditControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	heap_ptr obj = UPARAM(0);
 	heap_ptr event = UPARAM(1);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (obj) {
 		word ct_type = GET_SELECTOR(obj, type);
@@ -1302,7 +1337,7 @@ kEditControl(state_t *s, int funct_nr, int argc, heap_ptr argp)
 
 				if (modifiers & SCI_EVM_CTRL) {
 
-					switch (tolower(key)) {
+					switch (tolower((char)key)) {
 					case 'a': cursor = 0; break;
 					case 'e': cursor = textlen; break;
 					case 'f': if (cursor < textlen) ++cursor; break;
@@ -1447,8 +1482,8 @@ _k_draw_control(state_t *s, heap_ptr obj, int inverse)
 
 		SCIkdebug(SCIkGRAPHICS, "drawing button %04x to %d,%d\n", obj, x, y);
 		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_button_control(s->port, obj, area, text, font_nr,
-								  state & CONTROL_STATE_FRAMED,
-								  inverse, state & CONTROL_STATE_GRAY));
+								  (gint8)(state & CONTROL_STATE_FRAMED),
+								  (gint8)inverse, (gint8)(state & CONTROL_STATE_GRAY)));
 		break;
 
 	case K_CONTROL_TEXT:
@@ -1456,9 +1491,10 @@ _k_draw_control(state_t *s, heap_ptr obj, int inverse)
 
 		SCIkdebug(SCIkGRAPHICS, "drawing text %04x to %d,%d, mode=%d\n", obj, x, y, mode);
 
-		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_text_control(s->port, obj, area, text, font_nr, mode,
-								!!(state & CONTROL_STATE_DITHER_FRAMED),
-								inverse));
+		ADD_TO_CURRENT_BG_WIDGETS(
+			sciw_new_text_control(s->port, (int)obj, area, text, font_nr, mode,
+								(gint8)(!!(state & CONTROL_STATE_DITHER_FRAMED)),
+								(gint8)inverse));
 		break;
 
 	case K_CONTROL_EDIT:
@@ -1467,14 +1503,14 @@ _k_draw_control(state_t *s, heap_ptr obj, int inverse)
 		max = GET_SELECTOR(obj, max);
 		cursor = GET_SELECTOR(obj, cursor);
 
-		if (cursor > strlen(text))
+		if (cursor > (signed)strlen(text))
 			cursor = strlen(text);
 
 		if (text_offset == s->save_dir_copy)
 			update_cursor_limits(&s->save_dir_edit_offset, &cursor, max);
 
 		update_cursor_limits(&s->save_dir_edit_offset, &cursor, max);
-		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_edit_control(s->port, obj, area, text, font_nr, cursor, inverse));
+		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_edit_control(s->port, obj, area, text, font_nr, (unsigned)cursor, (gint8)inverse));
 		break;
 
 	case K_CONTROL_ICON:
@@ -1482,7 +1518,7 @@ _k_draw_control(state_t *s, heap_ptr obj, int inverse)
 		SCIkdebug(SCIkGRAPHICS, "drawing icon control %04x to %d,%d\n", obj, x, y -1);
 
 		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_icon_control(s->port, obj, area, view, loop, cel,
-							  state & CONTROL_STATE_FRAMED, inverse));
+							  (gint8)(state & CONTROL_STATE_FRAMED), (gint8)inverse));
 		break;
 
 	case K_CONTROL_CONTROL: {
@@ -1518,7 +1554,7 @@ _k_draw_control(state_t *s, heap_ptr obj, int inverse)
 		}
 
 		ADD_TO_CURRENT_BG_WIDGETS(sciw_new_list_control(s->port, obj, area, font_nr, entries_list, entries_nr,
-							  list_top, selection, inverse));
+							  list_top, selection, (gint8)inverse));
 		if (entries_nr)
 			free(entries_list);
 	}
@@ -1547,7 +1583,7 @@ draw_to_control_map(state_t *s, gfxw_dyn_view_t *view, int funct_nr, int argc, i
 /*	int has_nsrect = (view->ID <=0)? 0 : lookup_selector(s, view->ID, s->selector_map.nsBottom, NULL) == SELECTOR_VARIABLE;*/
 
 	if (view->ID > 0)
-		abs_zone = get_nsrect(s, view->ID, 1);
+		abs_zone = get_nsrect(s, (heap_ptr)view->ID, 1);
 	else {
 		abs_zone = nsrect_clip(s, view->pos.y,
 				       calculate_nsrect(s, view->pos.x, view->pos.y,
@@ -1585,9 +1621,14 @@ _k_view_list_do_postdraw(state_t *s, gfxw_list_t *list)
 
 	while (widget) {
 		int obj = widget->ID;
-
-		if ((widget->signal & (_K_VIEW_SIG_FLAG_FREESCI_PRIVATE | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_NO_UPDATE)) == _K_VIEW_SIG_FLAG_FREESCI_PRIVATE) {
-			int has_nsrect = lookup_selector(s, obj, s->selector_map.nsBottom, NULL) == SELECTOR_VARIABLE;
+		/*
+		 * this fixes a few problems, but doesn't match SSCI's logic.
+		 * The semantics of the private flag need to be verified before this can be uncommented.
+		 * Fixes bug #326 (CB1, ego falls down stairs)
+		 * if ((widget->signal & (_K_VIEW_SIG_FLAG_FREESCI_PRIVATE | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_NO_UPDATE)) == _K_VIEW_SIG_FLAG_FREESCI_PRIVATE) {
+		 */
+		if ((widget->signal & (_K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_NO_UPDATE)) == 0) {
+			int has_nsrect = lookup_selector(s, (heap_ptr)obj, s->selector_map.nsBottom, NULL) == SELECTOR_VARIABLE;
 
 			if (has_nsrect) {
 				int temp;
@@ -1615,11 +1656,10 @@ _k_view_list_do_postdraw(state_t *s, gfxw_list_t *list)
 				widget->signal |= _K_VIEW_SIG_FLAG_REMOVE;
 		}
 #ifdef DEBUG_LSRECT
-		fprintf(stderr, "obj %04x has pflags %x\n", obj, (widget->signal & (_K_VIEW_SIG_FLAG_FREESCI_PRIVATE | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_NO_UPDATE)));
+		fprintf(stderr, "obj %04x has pflags %x\n", obj, (widget->signal & (_K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_NO_UPDATE)));
 #endif
 
 		if (widget->signalp) {
-			widget->signal &= ~_K_VIEW_SIG_FLAG_FREESCI_PRIVATE;
 			PUT_HEAP(widget->signalp, widget->signal & 0xffff); /* Write back signal */
 		}
 
@@ -1653,7 +1693,7 @@ _k_view_list_dispose_loop(state_t *s, heap_ptr list_addr, gfxw_dyn_view_t *widge
 				int tempid = widget->ID;
 				heap_ptr under_bits = 0;
 
-				if (!is_object(s, tempid)) {
+				if (!is_object(s, (heap_ptr)tempid)) {
 					SCIkwarn(SCIkERROR, "Non-object %04x present"
 						  " in view list during delete time\n",
 						  tempid);
@@ -1769,8 +1809,9 @@ _k_make_dynview_obj(state_t *s, heap_ptr obj, int options, int nr, int funct_nr,
 	if (oldloop != loop)
 		PUT_SELECTOR(obj, loop, loop);
 
-	if (oldcel != cel)
+	if (oldcel != cel) {
 		PUT_SELECTOR(obj, cel, cel);
+	}
 
 	if (lookup_selector(s, obj, s->selector_map.underBits, &(under_bitsp))
 	    != SELECTOR_VARIABLE) {
@@ -1826,7 +1867,7 @@ _k_make_view_list(state_t *s, gfxw_list_t **widget_list, heap_ptr list, int opti
 
 	SCIkdebug(SCIkGRAPHICS, "Making list from %04x\n", list);
 
-	if (!listp(s, list - 2)) { /* heap size check */
+	if (!listp(s, (heap_ptr)(list - 2))) { /* heap size check */
 		SCIkwarn(SCIkWARNING, "Attempt to draw non-list at %04x\n", list);
 		return;
 	}
@@ -1844,20 +1885,13 @@ _k_make_view_list(state_t *s, gfxw_list_t **widget_list, heap_ptr list, int opti
 				SCIkdebug(SCIkGRAPHICS, "  invoking %04x::doit()\n", obj);
 				invoke_selector(INV_SEL(obj, doit, 1), 0); /* Call obj::doit() if neccessary */
 
-				if (s->pic_is_new) {
-					SCIkwarn(SCIkWARNING, "Warning: new pic"
-						 " or port used within Animate()!\n");
-					return;
-				}
 			}
 		}
 
 		widget = _k_make_dynview_obj(s, obj, options, sequence_nr--, funct_nr, argc, argp);
-		if (widget) {
+		if (widget)
 			GFX_ASSERT((*widget_list)->add(GFXWC(*widget_list), GFXW(widget)));
-
-			node = UGET_HEAP(node + LIST_NEXT_NODE); /* Next node */
-		}
+		node = UGET_HEAP(node + LIST_NEXT_NODE); /* Next node */
 	}
 
 
@@ -1865,8 +1899,7 @@ _k_make_view_list(state_t *s, gfxw_list_t **widget_list, heap_ptr list, int opti
 
 	while(widget) { /* Read back widget values */
 		if (widget->signalp)
-			widget->signal = widget->signal & _K_VIEW_SIG_FLAG_FREESCI_PRIVATE
-				| UGET_HEAP(widget->signalp);
+			widget->signal = UGET_HEAP(widget->signalp);
 
 		widget = (gfxw_dyn_view_t *) widget->next;
 	}
@@ -1881,16 +1914,16 @@ _k_prepare_view_list(state_t *s, gfxw_list_t *list, int options, int funct_nr, i
 	while (view) {
 		heap_ptr obj = view->ID;
 		int priority, _priority;
-		int has_nsrect = (view->ID <=0)? 0 : lookup_selector(s, view->ID, s->selector_map.nsBottom, NULL) == SELECTOR_VARIABLE;
+		int has_nsrect = (view->ID <=0)? 0 : lookup_selector(s, (heap_ptr)view->ID, s->selector_map.nsBottom, NULL) == SELECTOR_VARIABLE;
 		int oldsignal = view->signal;
 
-		_k_set_now_seen(s, view->ID);
+		_k_set_now_seen(s, (heap_ptr)view->ID);
 		_priority = /*GET_SELECTOR(obj, y); */((view->pos.y));/**/
 		_priority = VIEW_PRIORITY(_priority - 1);
 
 		if (options & _K_MAKE_VIEW_LIST_DRAW_TO_CONTROL_MAP) { /* Picview */
 			priority = GET_SELECTOR(obj, priority);
-			if (priority <= 0)
+			if (priority < 0)
 				priority = _priority; /* Always for picviews */
 		} else { /* Dynview */
 			if (has_nsrect
@@ -1924,35 +1957,42 @@ _k_prepare_view_list(state_t *s, gfxw_list_t *list, int options, int funct_nr, i
 			draw_to_control_map(s, view, funct_nr, argc, argp);
 
 
-		/* Extreme Pattern Matching ugliness ahead... */
-		if ((view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE
-		     && (  /* Brainwashed by LISP */
-			 (view->signal & (_K_VIEW_SIG_FLAG_UPDATED | _K_VIEW_SIG_FLAG_FORCE_UPDATE))
-			 || (((view->signal & (_K_VIEW_SIG_FLAG_UPDATED | _K_VIEW_SIG_FLAG_FORCE_UPDATE)) == 0)
-			     && (((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE)) == _K_VIEW_SIG_FLAG_REMOVE)
-				 || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)
-				 || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE))
-				     == (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE))
-				 )
-			     )
-			 )
-		     )
-		    ||
-		    (!(view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE)
-		     && (!!(view->signal & _K_VIEW_SIG_FLAG_STOP_UPDATE) ^ !!(view->signal & _K_VIEW_SIG_FLAG_ALWAYS_UPDATE))
-		     )
-		    ) {
-			view->signal |= _K_VIEW_SIG_FLAG_FREESCI_PRIVATE;
-			s->pic_not_valid++;
+		/* Extreme Pattern Matching ugliness ahead... step 9.1 */
+		if (view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE) {
+			if (((view->signal & (_K_VIEW_SIG_FLAG_UPDATED | _K_VIEW_SIG_FLAG_FORCE_UPDATE))) /* 9.1.1.1 */
+			    || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE)) == _K_VIEW_SIG_FLAG_HIDDEN)
+			    || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE)) == _K_VIEW_SIG_FLAG_REMOVE) /* 9.1.1.2 */
+			    || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == _K_VIEW_SIG_FLAG_ALWAYS_UPDATE) /* 9.1.1.3 */
+			    || ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE))) /* 9.1.1.4 */
+			{
+				s->pic_not_valid++;
+				view->signal &= ~_K_VIEW_SIG_FLAG_STOP_UPDATE;
+			}
+
+			else if (((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == 0)
+				|| ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_REMOVE))
+				|| ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE))
+				|| ((view->signal & (_K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == _K_VIEW_SIG_FLAG_HIDDEN))
+			{
+				view->signal &= ~_K_VIEW_SIG_FLAG_STOP_UPDATE;
+			}
+
 		}
+		else {
+			if (view->signal & _K_VIEW_SIG_FLAG_STOP_UPDATE) {
+				s->pic_not_valid++;
+				view->signal &= ~_K_VIEW_SIG_FLAG_FORCE_UPDATE;
+			}
+			else { /* if not STOP_UPDATE */
+				if (view->signal & _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)
+					s->pic_not_valid++;
+				view->signal &= ~_K_VIEW_SIG_FLAG_FORCE_UPDATE;
+			}
+		}
+		
 
 		SCIkdebug(SCIkGRAPHICS, "  dv[%04x]: signal %04x -> %04x\n", view->ID, oldsignal, view->signal);
 
-		if (view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE)
-			view->signal &= ~_K_VIEW_SIG_FLAG_STOP_UPDATE;
-		else
-			view->signal &= ~_K_VIEW_SIG_FLAG_FORCE_UPDATE;
-
 		/* Never happens
 		if (view->signal & 0) {
 			view->signal &= ~_K_VIEW_SIG_FLAG_FREESCI_STOPUPD;
@@ -1985,7 +2025,14 @@ _k_update_signals_in_view_list(gfxw_list_t *old_list, gfxw_list_t *new_list)
 			/* Transfer 'stopupd' flag */
 
 			if ((new_widget->pos.x != old_widget->pos.x)
-			    || (new_widget->pos.y != old_widget->pos.y))
+			    || (new_widget->pos.y != old_widget->pos.y)
+			    /* ** No idea why this is supposed to be bad **
+			    || (new_widget->z != old_widget->z)
+			    || (new_widget->view != old_widget->view)
+			    || (new_widget->loop != old_widget->loop)
+			    || (new_widget->cel != old_widget->cel)
+			    */
+			    )
 				carry = 0;
 
 			old_widget->signal = new_widget->signal |= carry;
@@ -2010,9 +2057,14 @@ _k_raise_topmost_in_view_list(state_t *s, gfxw_list_t *list, gfxw_dyn_view_t *vi
 	if (view) {
 		gfxw_dyn_view_t *next = (gfxw_dyn_view_t *) view->next;
 
+		/* step 11 */
 		if ((view->signal & (_K_VIEW_SIG_FLAG_NO_UPDATE | _K_VIEW_SIG_FLAG_HIDDEN | _K_VIEW_SIG_FLAG_ALWAYS_UPDATE)) == 0) {
 			SCIkdebug(SCIkGRAPHICS, "Forcing precedence 2 at [%04x] with %04x\n", view->ID, view->signal);
 			view->force_precedence = 2;
+
+			if ((view->signal & (_K_VIEW_SIG_FLAG_REMOVE | _K_VIEW_SIG_FLAG_HIDDEN)) == _K_VIEW_SIG_FLAG_REMOVE) {
+				view->signal &= ~_K_VIEW_SIG_FLAG_REMOVE;
+			}
 		}
 
 		gfxw_remove_widget_from_container(view->parent, GFXW(view));
@@ -2038,8 +2090,11 @@ _k_redraw_view_list(state_t *s, gfxw_list_t *list)
 
 		SCIkdebug(SCIkGRAPHICS, "  dv[%04x]: signal %04x\n", view->ID, view->signal);
 
+		/* step 1 of subalgorithm */
 		if (view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE) {
-			view->signal &= ~_K_VIEW_SIG_FLAG_FORCE_UPDATE;
+			if (view->signal & _K_VIEW_SIG_FLAG_FORCE_UPDATE)
+				view->signal &= ~_K_VIEW_SIG_FLAG_FORCE_UPDATE;
+
 			if (view->signal & _K_VIEW_SIG_FLAG_UPDATED)
 				view->signal &= ~(_K_VIEW_SIG_FLAG_UPDATED | _K_VIEW_SIG_FLAG_NO_UPDATE);
 		} else { /* NO_UPD is not set */
@@ -2057,7 +2112,7 @@ _k_redraw_view_list(state_t *s, gfxw_list_t *list)
 
 		SCIkdebug(SCIkGRAPHICS, "    at substep 11/14: signal %04x\n", view->signal);
 
-		if (!(view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE)) {
+		if (view->signal & _K_VIEW_SIG_FLAG_NO_UPDATE) {
 			if (view->signal & _K_VIEW_SIG_FLAG_HIDDEN)
 				view->signal |= _K_VIEW_SIG_FLAG_REMOVE;
 			else
@@ -2136,7 +2191,6 @@ kAddToPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 {
 	gfxw_list_t *pic_views;
 	heap_ptr list = PARAM(0);
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	assert_primary_widget_lists(s);
 
@@ -2159,7 +2213,8 @@ kAddToPic(state_t *s, int funct_nr, int argc, heap_ptr argp)
 			SCIkwarn(SCIkERROR, "Attempt to single-add invalid picview (%d/%d/%d)\n", view, loop, cel);
 		} else {
 			widget->ID = -1;
-			draw_to_control_map(s, (gfxw_dyn_view_t *) widget, funct_nr, argc, argp);
+			if (control >= 0)
+				draw_to_control_map(s, (gfxw_dyn_view_t *) widget, funct_nr, argc, argp);
 			ADD_TO_CURRENT_PICTURE_PORT(gfxw_picviewize_dynview((gfxw_dyn_view_t *) widget));
 		}
 
@@ -2198,7 +2253,6 @@ kSetPort(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	unsigned int port_nr = PARAM(0);
 	gfxw_port_t *new_port;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	new_port = gfxw_find_port(s->visual, port_nr);
 
@@ -2223,7 +2277,6 @@ kDrawCel(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int priority = PARAM_OR_ALT(5, -1);
 	gfxw_view_t *new_view;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	if (!view) {
 		SCIkwarn(SCIkERROR, "Attempt to draw non-existing view.%03n\n", view);
@@ -2254,7 +2307,6 @@ kDisposeWindow(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	gfxw_port_t *goner;
 	gfxw_port_t *pred;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	goner = gfxw_find_port(s->visual, goner_nr);
 	if ((goner_nr < 3) || (goner == NULL)) {
@@ -2291,14 +2343,13 @@ kNewWindow(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	gfx_color_t bgcolor	= {0};
 	int priority;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	y = PARAM(0);
 	x = PARAM(1);
 	yl = PARAM(2) - y;
 	xl = PARAM(3) - x;
 
-	s->pic_is_new = 1; /* Well, it's different, at any rate... */
+	s->pic_is_new = PIC_IS_NEW_PORT; /* Well, it's different, at any rate... */
 	/* This is for Animate(), to inform it if one of its doit()s is responsible
 	** for the window being opened- in this case, we restart Animate().  */
 
@@ -2769,7 +2820,6 @@ kAnimate(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	int old_pnv = s->pic_not_valid; /* Backup old value */
 	int open_animation = 0;
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	process_sound_events(s); /* Take care of incoming events (kAnimate is called semi-regularly) */
 	_k_animate_ran = 1; /* Used by some of the invoked functions to check for recursion, which may,
@@ -2802,7 +2852,7 @@ kAnimate(state_t *s, int funct_nr, int argc, heap_ptr argp)
 		/* End of doit() recovery code */
 
 
-		if (s->pic_is_new) { /* Happens if DrawPic() is executed by a dynview (yes, that happens) */
+		if (s->pic_is_new == PIC_IS_NEW_PIC) { /* Happens if DrawPic() is executed by a dynview (yes, that happens) */
 			kAnimate(s, funct_nr, argc, argp); /* Tail-recurse */
 			return;
 		}
@@ -2946,7 +2996,6 @@ kDisplay(state_t *s, int funct_nr, int argc, heap_ptr argp)
 	color0 = &(port->color);
 	bg_color = &(port->bgcolor);
 
-	CHECK_THIS_KERNEL_FUNCTION;
 
 	text = kernel_lookup_text(s, textp, index);
 
