@@ -1,6 +1,5 @@
 /***************************************************************************
- scriptdebug.c Copyright (C) 1999,2000,01,02 Christoph Reichenbach
-
+ scriptdebug.c Copyright (C) 1999 Christoph Reichenbach, TU Darmstadt
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -31,6 +30,11 @@
 #include <console.h>
 #include <kdebug.h>
 #include <vocabulary.h>
+#include <sci_widgets.h>
+
+#ifdef _MSC_VER
+#  include <sci_win32.h>
+#endif _MSC_VER
 
 int _debugstate_valid = 0; /* Set to 1 while script_debug is running */
 int _debug_step_running = 0; /* Set to >0 to allow multiple stepping */
@@ -39,19 +43,17 @@ int _debug_seeking = 0; /* Stepping forward until some special condition is met
 int _debug_seek_level = 0; /* Used for seekers that want to check their exec stack depth */
 int _debug_seek_special = 0; /* Used for special seeks */
 
-static char oldcommand[SCI_CONSOLE_MAX_INPUT + 1] = ""; /* Stores the last command executed */
-
 #define _DEBUG_SEEK_NOTHING 0
 #define _DEBUG_SEEK_CALLK 1 /* Step forward until callk is found */
 #define _DEBUG_SEEK_LEVEL_RET 2 /* Step forward until returned from this level */
 #define _DEBUG_SEEK_SPECIAL_CALLK 3 /* Step forward until a /special/ callk is found */
 #define _DEBUG_SEEK_SO 5 /* Step forward until specified PC (after the send command) and stack depth */
 
-heap_ptr *_pc;
-heap_ptr *_sp;
-heap_ptr *_pp;
-heap_ptr *_objp;
-int *_restadjust;
+heap_ptr *p_pc;
+heap_ptr *p_sp;
+heap_ptr *p_pp;
+heap_ptr *p_objp;
+int *p_restadjust;
 
 int _kdebug_cheap_event_hack = 0;
 int _kdebug_cheap_soundcue_hack = -1;
@@ -82,11 +84,11 @@ char *(*_debug_get_input)(void) = _debug_get_input_default;
 int
 c_debuginfo(state_t *s)
 {
-  sciprintf("pc=%04x acc=%04x o=%04x fp=%04x sp=%04x\n", *_pc & 0xffff, s->acc & 0xffff,
-	    *_objp & 0xffff, *_pp & 0xffff, *_sp & 0xffff);
+  sciprintf("pc=%04x acc=%04x o=%04x fp=%04x sp=%04x\n", *p_pc & 0xffff, s->acc & 0xffff,
+	    *p_objp & 0xffff, *p_pp & 0xffff, *p_sp & 0xffff);
   sciprintf("prev=%x sbase=%04x globls=%04x &restmod=%x\n",
 	    s->prev & 0xffff, s->stack_base & 0xffff, s->global_vars & 0xffff,
-	    *_restadjust & 0xffff);
+	    *p_restadjust & 0xffff);
   return 0;
 }
 
@@ -96,7 +98,7 @@ c_step(state_t *s)
 {
   _debugstate_valid = 0;
   if (cmd_paramlength && (cmd_params[0].val > 0))
-    _debug_step_running = cmd_params[0].val;
+    _debug_step_running = cmd_params[0].val - 1;
   return 0;
 }
 
@@ -112,7 +114,7 @@ c_stepover(state_t *s)
   }
 
   _debugstate_valid = 0;
-  opcode = s->heap [*_pc];
+  opcode = s->heap [*p_pc];
   opnumber = opcode >> 1;
   if (opnumber == 0x22 /* callb */ || opnumber == 0x23 /* calle */ ||
       opnumber == 0x25 /* send */ || opnumber == 0x2a /* self */ ||
@@ -124,21 +126,21 @@ c_stepover(state_t *s)
     switch (opcode)
     {
     case 0x46: /* calle W */
-      _debug_seek_special = *_pc + 5; break;
+      _debug_seek_special = *p_pc + 5; break;
 
     case 0x44: /* callb W */
     case 0x47: /* calle B */
     case 0x56: /* super W */
-      _debug_seek_special = *_pc + 4; break;
+      _debug_seek_special = *p_pc + 4; break;
 
     case 0x45: /* callb B */
     case 0x57: /* super B */
     case 0x4A: /* send W */
     case 0x54: /* self W */
-      _debug_seek_special = *_pc + 3; break;
+      _debug_seek_special = *p_pc + 3; break;
 
     default:
-      _debug_seek_special = *_pc + 2;
+      _debug_seek_special = *p_pc + 2;
     }
   }
 
@@ -265,12 +267,12 @@ c_viewinfo(state_t *s)
 {
 	int view = cmd_params[0].val;
 	int loops, i;
+	gfxr_view_t *view_pixmaps = NULL;
 
 	if (!s) {
 		sciprintf("Not in debug state\n");
 		return 1;
 	}
-
 	sciprintf("Resource view.%d ", view);
 
 	loops = gfxop_lookup_view_get_loops(s->gfx_state, view);
@@ -278,7 +280,6 @@ c_viewinfo(state_t *s)
 	if (loops < 0)
 		sciprintf("does not exist.\n");
 	else {
-
 		sciprintf("has %d loops:\n", loops);
 
 		for (i = 0; i < loops; i++) {
@@ -290,6 +291,13 @@ c_viewinfo(state_t *s)
 				int height;
 				point_t mod;
 
+				if (con_can_handle_pixmaps()) {
+					view_pixmaps = gfxr_get_view(s->gfx_state->resstate,
+								     view, &i, &j);
+					con_insert_pixmap(gfx_clone_pixmap(view_pixmaps->loops[i].cels[j],
+									   s->gfx_state->driver->mode));
+				}
+
 				gfxop_get_cel_parameters(s->gfx_state, view, i, j, &width, &height, &mod);
 
 				sciprintf("   cel %d: size %dx%d, adj+(%d,%d)\n",j, width, height, mod.x, mod.y);
@@ -360,7 +368,7 @@ _parse_getinp(int *i, int *nr)
 {
   char *token;
 
-  if (*i == cmd_paramlength)
+  if ((unsigned)*i == cmd_paramlength)
     return _parse_eoi;
 
   token = cmd_params[(*i)++].str;
@@ -500,7 +508,7 @@ c_save_game(state_t *s)
 		return 1;
 	}
 
-	s->amp_rest = *_restadjust;
+	s->amp_rest = *p_restadjust;
 
 	if (!omit_check) {
 		int result = 0;
@@ -595,7 +603,7 @@ c_stack(state_t *s)
   }
 
   for (i = cmd_params[0].val ; i > 0; i--)
-    sciprintf("[sp-%04x] = %04x\n", i * 2, 0xffff & getInt16(s->heap + *_sp - i*2));
+    sciprintf("[sp-%04x] = %04x\n", i * 2, 0xffff & getInt16(s->heap + *p_sp - i*2));
 
   return 0;
 }
@@ -690,187 +698,232 @@ print_objname(state_t *s, heap_ptr pos, int address)
 }
 
 heap_ptr
-disassemble(state_t *s, heap_ptr pos)
+disassemble(state_t *s, heap_ptr pos, int print_bw_tag, int print_bytecode)
 /* Disassembles one command from the heap, returns address of next command or 0 if a ret was
 ** encountered.
 */
 {
-  heap_ptr retval = pos + 1;
-  word param_value;
-  int opsize = s->heap[pos];
-  int opcode = opsize >> 1;
-  int i = 0;
+	heap_ptr retval = pos + 1;
+	word param_value;
+	int opsize = s->heap[pos];
+	int opcode = opsize >> 1;
+	int bytecount = 1;
+	int i = 0;
 
-  if (!_debugstate_valid) {
-    sciprintf("Not in debug state\n");
-    return 1;
-  }
-
-  opsize &= 1; /* byte if true, word if false */
-
-  sciprintf("%04x: [%c] %s", pos, opsize? 'B' : 'W', s->opcodes[opcode].name);
-
-  while (formats[opcode][i])
-
-    switch (formats[opcode][i++]) {
-
-    case Script_Invalid: sciprintf("-Invalid operation-"); break;
-
-    case Script_SByte:
-    case Script_Byte: sciprintf(" %02x", s->heap[retval++]); break;
-
-    case Script_Word:
-    case Script_SWord:
-      sciprintf(" %04x", 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8)));
-      retval += 2;
-      break;
-
-    case Script_SVariable:
-    case Script_Variable:
-    case Script_Property:
-    case Script_Global:
-    case Script_Local:
-    case Script_Temp:
-    case Script_Param:
-      if (opsize)
-	param_value = s->heap[retval++];
-      else {
-	param_value = 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8));
-	retval += 2;
-      }
-
-      if (opcode == op_callk)
-	sciprintf(" %s[%x]", (param_value < s->kernel_names_nr)
-		  ? s->kernel_names[param_value] : "<invalid>", param_value);
-      else sciprintf(opsize? " %02x" : " %04x", param_value);
-
-      break;
-
-    case Script_SRelative:
-      if (opsize)
-	param_value = s->heap[retval++];
-      else {
-	param_value = 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8));
-	retval += 2;
-      }
-      sciprintf (opsize? " %02x  [%04x]" : " %04x  [%04x]", param_value, retval+(short) param_value);
-      break;
-
-    case Script_End: retval = 0;
-      break;
-
-    default:
-      sciprintf("Internal assertion failed in 'disassemble', %s, L%d\n", __FILE__, __LINE__);
-
-    }
-
-  if (pos == *_pc) /* Extra information if debugging the current opcode */
-
-    if ((opcode == op_pTos)||(opcode == op_sTop)||
-        (opcode == op_pToa)||(opcode == op_aTop)||
-	(opcode == op_dpToa)||(opcode == op_ipToa)||
-	(opcode == op_dpTos)||(opcode == op_ipTos))
-    {
-      int prop_ofs = s->heap[retval - 1];
-      int prop_id = prop_ofs_to_id(s, prop_ofs, *_objp & 0xffff);
-
-      sciprintf("	(%s)", selector_name(s, prop_id));
-    }
-
-  sciprintf("\n");
-
-  if (pos == *_pc) { /* Extra information if debugging the current opcode */
-
-    if (opcode == op_callk) {
-      int stackframe = s->heap[retval - 1] + (*_restadjust * 2);
-      int argc = getInt16(s->heap + *_sp - stackframe - 2);
-      int i;
-
-      sciprintf(" Kernel params: (");
-
-      for (i = 0; i < argc; i++) {
-	sciprintf("%04x", getUInt16(s->heap + *_sp - stackframe + i*2));
-	if (i+1 < argc)
-	  sciprintf(", ");
-      }
-      sciprintf(")\n");
-
-    } else
-
-    if ((opcode == op_send) || (opcode == op_self)) {
-      int restmod = *_restadjust;
-      int stackframe = s->heap[retval - 1] + (restmod * 2);
-      word selector;
-      heap_ptr selector_addr;
-
-      while (stackframe > 0) {
-	int argc = getInt16(s->heap + *_sp - stackframe + 2);
-	heap_ptr nameaddress = 0;
-	heap_ptr called_obj_addr;
-	char *name;
-
-	if (opcode == op_send)
-	  called_obj_addr = s->acc;
-	else if (opcode == op_self)
-	  called_obj_addr = *_objp;
-	else { /* op_super */
-	  called_obj_addr = s->heap[retval - 2];
-
-	  called_obj_addr = (s->classtable[called_obj_addr].scriptposp)?
-		  0 :
-		  *(s->classtable[called_obj_addr].scriptposp)
-		  + s->classtable[called_obj_addr].class_offset;
+	if (!_debugstate_valid) {
+		sciprintf("Not in debug state\n");
+		return 1;
 	}
 
-	selector = getInt16(s->heap + *_sp - stackframe);
+	opsize &= 1; /* byte if true, word if false */
 
-	if (called_obj_addr > 100) /* If we are in valid heap space */
-	  if (getInt16(s->heap + called_obj_addr + SCRIPT_OBJECT_MAGIC_OFFSET)
-	      == SCRIPT_OBJECT_MAGIC_NUMBER)
-	    nameaddress = getUInt16(s->heap + called_obj_addr + SCRIPT_NAME_OFFSET);
+  
+	sciprintf("%04x: ", pos);
 
-	if (nameaddress)
-	  name = (char *) s->heap + nameaddress;
-	else
-	  name = "<invalid>";
+	if (print_bytecode) {
+		while (formats[opcode][i]) {
+			switch (formats[opcode][i++]) {
 
-	sciprintf("  %s::%s[", name, (selector > s->selector_names_nr)
-		  ? "<invalid>" : selector_name(s,selector));
+			case Script_SByte:
+			case Script_Byte: bytecount++;
+				break;
 
-	switch (lookup_selector(s, called_obj_addr, selector, &selector_addr)) {
-	case SELECTOR_METHOD:
-	  sciprintf("FUNCT");
-	  argc += restmod;
-	  restmod = 0;
-	  break;
-	case SELECTOR_VARIABLE:
-	  sciprintf("VAR");
-	  break;
-	case SELECTOR_NONE:
-	  sciprintf("INVALID");
-	  break;
-	}
+			case Script_Word:
+			case Script_SWord: bytecount += 2;
+				break;
 
-	sciprintf("](");
+			case Script_SVariable:
+			case Script_Variable:
+			case Script_Property:
+			case Script_Global:
+			case Script_Local:
+			case Script_Temp:
+			case Script_Param:
+			case Script_SRelative:
+				if (opsize)
+					bytecount ++;
+				else
+					bytecount += 2;
+				break;
 
-	while (argc--) {
+			default:
+				break;
+			}
+		}
 
-	  sciprintf("%04x", 0xffff & getUInt16(s->heap + *_sp - stackframe + 4));
-	  if (argc) sciprintf(", ");
-	  stackframe -= 2;
+		for (i = 0; i < bytecount; i++)
+			sciprintf("%02x ", s->heap[pos + i]);
+
+		for (i = bytecount; i < 5; i++)
+			sciprintf("   ");
 
 	}
 
-	sciprintf(")\n");
+	if (print_bw_tag)
+		sciprintf("[%c] ", opsize? 'B' : 'W');
+	sciprintf("%s", s->opcodes[opcode].name);
 
-	stackframe -= 4;
-      } /* while (stackframe > 0) */
+	i = 0;
+	while (formats[opcode][i])
 
-    } /* Send-like opcodes */
+		switch (formats[opcode][i++]) {
 
-  } /* (heappos == *_pc) */
+		case Script_Invalid: sciprintf("-Invalid operation-"); break;
 
-  return retval;
+		case Script_SByte:
+		case Script_Byte: sciprintf(" %02x", s->heap[retval++]); break;
+
+		case Script_Word:
+		case Script_SWord:
+			sciprintf(" %04x", 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8)));
+			retval += 2;
+			break;
+
+		case Script_SVariable:
+		case Script_Variable:
+		case Script_Property:
+		case Script_Global:
+		case Script_Local:
+		case Script_Temp:
+		case Script_Param:
+			if (opsize)
+				param_value = s->heap[retval++];
+			else {
+				param_value = 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8));
+				retval += 2;
+			}
+
+			if (opcode == op_callk)
+				sciprintf(" %s[%x]", (param_value < s->kernel_names_nr)
+					  ? s->kernel_names[param_value] : "<invalid>", param_value);
+			else sciprintf(opsize? " %02x" : " %04x", param_value);
+
+			break;
+
+		case Script_SRelative:
+			if (opsize)
+				param_value = s->heap[retval++];
+			else {
+				param_value = 0xffff & (s->heap[retval] | (s->heap[retval+1] << 8));
+				retval += 2;
+			}
+			sciprintf (opsize? " %02x  [%04x]" : " %04x  [%04x]", param_value, (0xffff) & (retval+(heap_ptr) param_value));
+			break;
+
+		case Script_End: retval = 0;
+			break;
+
+		default:
+			sciprintf("Internal assertion failed in 'disassemble', %s, L%d\n", __FILE__, __LINE__);
+
+		}
+
+	if (pos == *p_pc) /* Extra information if debugging the current opcode */
+
+		if ((opcode == op_pTos)||(opcode == op_sTop)||
+		    (opcode == op_pToa)||(opcode == op_aTop)||
+		    (opcode == op_dpToa)||(opcode == op_ipToa)||
+		    (opcode == op_dpTos)||(opcode == op_ipTos)) {
+			int prop_ofs = s->heap[retval - 1];
+			int prop_id = prop_ofs_to_id(s, prop_ofs, *p_objp & 0xffff);
+
+			sciprintf("	(%s)", selector_name(s, prop_id));
+		}
+
+	sciprintf("\n");
+
+	if (pos == *p_pc) { /* Extra information if debugging the current opcode */
+
+		if (opcode == op_callk) {
+			int stackframe = s->heap[retval - 1] + (*p_restadjust * 2);
+			int argc = getInt16(s->heap + *p_sp - stackframe - 2);
+			int i;
+
+			sciprintf(" Kernel params: (");
+
+			for (i = 0; i < argc; i++) {
+				sciprintf("%04x", getUInt16(s->heap + *p_sp - stackframe + i*2));
+				if (i+1 < argc)
+					sciprintf(", ");
+			}
+			sciprintf(")\n");
+
+		} else
+
+			if ((opcode == op_send) || (opcode == op_self)) {
+				int restmod = *p_restadjust;
+				int stackframe = s->heap[retval - 1] + (restmod * 2);
+				word selector;
+				heap_ptr selector_addr;
+
+				while (stackframe > 0) {
+					int argc = getInt16(s->heap + *p_sp - stackframe + 2);
+					heap_ptr nameaddress = 0;
+					heap_ptr called_obj_addr;
+					char *name;
+
+					if (opcode == op_send)
+						called_obj_addr = s->acc;
+					else if (opcode == op_self)
+						called_obj_addr = *p_objp;
+					else { /* op_super */
+						called_obj_addr = s->heap[retval - 2];
+
+						called_obj_addr = (s->classtable[called_obj_addr].scriptposp)?
+							0 :
+							*(s->classtable[called_obj_addr].scriptposp)
+							+ s->classtable[called_obj_addr].class_offset;
+					}
+
+					selector = getInt16(s->heap + *p_sp - stackframe);
+
+					if (called_obj_addr > 100) /* If we are in valid heap space */
+						if (getInt16(s->heap + called_obj_addr + SCRIPT_OBJECT_MAGIC_OFFSET)
+						    == SCRIPT_OBJECT_MAGIC_NUMBER)
+							nameaddress = getUInt16(s->heap + called_obj_addr + SCRIPT_NAME_OFFSET);
+
+					if (nameaddress)
+						name = (char *) s->heap + nameaddress;
+					else
+						name = "<invalid>";
+
+					sciprintf("  %s::%s[", name, (selector > s->selector_names_nr)
+						  ? "<invalid>" : selector_name(s,selector));
+
+					switch (lookup_selector(s, called_obj_addr, selector, &selector_addr)) {
+					case SELECTOR_METHOD:
+						sciprintf("FUNCT");
+						argc += restmod;
+						restmod = 0;
+						break;
+					case SELECTOR_VARIABLE:
+						sciprintf("VAR");
+						break;
+					case SELECTOR_NONE:
+						sciprintf("INVALID");
+						break;
+					}
+
+					sciprintf("](");
+
+					while (argc--) {
+
+						sciprintf("%04x", 0xffff & getUInt16(s->heap + *p_sp - stackframe + 4));
+						if (argc) sciprintf(", ");
+						stackframe -= 2;
+
+					}
+
+					sciprintf(")\n");
+
+					stackframe -= 4;
+				} /* while (stackframe > 0) */
+
+			} /* Send-like opcodes */
+
+	} /* (heappos == *p_pc) */
+
+	return retval;
 }
 
 int
@@ -1117,8 +1170,16 @@ c_gfx_priority(state_t *s)
 			zone = 0;
 		if (zone > 15) zone = 15;
 
-		sciprintf("Zone %x starts at y=%d\n", zone, PRIORITY_BAND_FIRST(zone));
+		if (s->pic_priority_table)
+			sciprintf("Zone %d starts at y=%d\n", zone, PRIORITY_BAND_FIRST(zone)); else
+			sciprintf("Zone %d starts at y=%d\n", zone, s->pic_priority_table[zone]);
 	} else {
+		if (s->pic_priority_table)
+		{
+			int j;
+			for (j=0;j<15;j++)
+				printf("Zone %d starts at y=%d\n", j, s->pic_priority_table[j]);
+		} else
 		sciprintf("Priority bands start at y=%d\nThey end at y=%d\n", s->priority_first, s->priority_last);
 	}
 
@@ -1157,6 +1218,22 @@ c_gfx_print_dynviews(state_t *s)
 	return 0;
 }
 
+int
+c_gfx_print_dropviews(state_t *s)
+{
+	if (!_debugstate_valid) {
+		sciprintf("Not in debug state\n");
+		return 1;
+	}
+
+	if (!s->drop_views)
+		sciprintf("No dropped dynview list active.\n");
+	else
+		s->drop_views->print(GFXW(s->drop_views), 0);
+
+	return 0;
+}
+
 int
 c_gfx_drawpic(state_t *s)
 {
@@ -1350,7 +1427,7 @@ bb = GET_SELECTOR(pos, bb);
 int
 c_gfx_draw_viewobj(state_t *s)
 {
-	heap_ptr pos = cmd_params[0].val;
+	heap_ptr pos = (heap_ptr) (cmd_params[0].val);
 	int is_view;
 	int x, y, priority;
 	int nsLeft, nsRight, nsBottom, nsTop;
@@ -1466,7 +1543,11 @@ int
 c_disasm(state_t *s)
 {
 	int vpc = cmd_params[0].val;
-	int op_count;
+	int op_count = 1;
+	int do_bwc = 0;
+	int do_bytes = 0;
+	unsigned int i;
+	int invalid = 0;
 
 	if (!_debugstate_valid) {
 		sciprintf("Not in debug state\n");
@@ -1474,19 +1555,26 @@ c_disasm(state_t *s)
 	}
 
 	if (vpc <= 0)
-		vpc = *_pc + vpc;
+		vpc = *p_pc + vpc;
 
-	if (cmd_paramlength > 1)
-		{
-			if (cmd_params[1].val < 0)
-				return 0;
-			op_count = cmd_params[1].val;
-		}
-	else
-		op_count = 1;
+	for (i = 1; i < cmd_paramlength; i++) {
+			if (!strcasecmp(cmd_params[i].str, "bwt"))
+				do_bwc = 1;
+			else if (!strcasecmp(cmd_params[i].str, "bc"))
+				do_bytes = 1;
+			else if (toupper(cmd_params[i].str[0]) == 'C')
+				op_count = atoi(cmd_params[i].str + 1);
+			else {
+				invalid = 1;
+				sciprintf("Invalid option '%s'\n", cmd_params[i].str);
+			}
+	}
+
+	if (invalid || op_count < 0)
+		return invalid;
 
 	do {
-		vpc = disassemble(s, vpc);
+		vpc = disassemble(s, (heap_ptr)vpc, do_bwc, do_bytes);
 	} while ((vpc > 0) && (vpc < 0xfff2) && (cmd_paramlength > 1) && (--op_count));
 	return 0;
 }
@@ -1565,7 +1653,7 @@ c_set_acc(state_t *s)
 		return 1;
 	}
 
-	s->acc = cmd_params[0].val;
+	s->acc = (gint16) (cmd_params[0].val);
 	return 0;
 }
 
@@ -1620,7 +1708,7 @@ c_listclones(state_t *s)
 void
 set_debug_mode (struct _state *s, int mode, char *areas)
 {
-	char modechars[] = "ulgcmfbadspMSFt"; /* Valid parameter chars */
+	char modechars[] = "ulgcmfbadspMSFtr"; /* Valid parameter chars */
 	char *parser;
 	int seeker;
 	char frob;
@@ -1893,7 +1981,7 @@ c_objs(state_t *s)
 
 				if (segment == sci_obj_object || segment == sci_obj_class)
 					print_objname_wrapper(print_objname
-							      (s, seeker - SCRIPT_OBJECT_MAGIC_OFFSET + 4
+							      (s, (heap_ptr)(seeker - SCRIPT_OBJECT_MAGIC_OFFSET + 4)
 							       /* to compensate for the header */, 1));
 
 				seeker += size;
@@ -2249,14 +2337,14 @@ int
 c_heapobj(state_t *s)
 {
   return
-    objinfo(s, cmd_params[0].val);
+    objinfo(s, (heap_ptr)cmd_params[0].val);
 }
 
 int
 c_obj(state_t *s)
 {
   return
-    objinfo(s, *_objp);
+    objinfo(s, *p_objp);
 }
 
 int
@@ -2418,6 +2506,27 @@ c_se(state_t *s)
 	return 0;
 }
 
+int
+c_statusbar(state_t *s)
+{
+	if (!s) {
+		sciprintf("Not in debug state\n");
+		return 1;
+	}
+
+	s->titlebar_port->color=s->ega_colors[cmd_params[0].val];
+	s->titlebar_port->bgcolor=s->ega_colors[cmd_params[1].val];
+
+	s->status_bar_foreground=cmd_params[0].val;
+	s->status_bar_background=cmd_params[1].val;
+
+	sciw_set_status_bar(s, s->titlebar_port, s->status_bar_text,
+			    s->status_bar_foreground,
+			    s->status_bar_background);
+	gfxop_update(s->gfx_state);
+	return 0;
+}
+
 int
 c_sci_version(state_t *s)
 {
@@ -2434,22 +2543,45 @@ c_sci_version(state_t *s)
 	return 0;
 }
 
+int
+c_sleep(state_t *s)
+{
+	sleep(cmd_params[0].val);
+	return 0;
+}
+
 void
 script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *objp,
 	     int *restadjust, int bp)
 {
+	int have_windowed = s->gfx_state->driver->capabilities & GFX_CAPABILITY_WINDOWED;
+	/* Do we support a separate console? */
+
+#ifndef WANT_CONSOLE
+	if (!have_windowed) {
+		script_debug_flag = sci_debug_flags = 0;
+		fprintf(stderr, "On-screen console disabled and driver claims not to support windowed mode.\n");
+		if (script_error_flag) {
+			fprintf(stderr, "Aborting...\n");
+			exit(1);
+		} else
+			fprintf(stderr, "Continuing...\n");
+		return;
+	}
+#endif
+
 
 	if (sci_debug_flags & _DEBUG_FLAG_LOGGING) {
 		int old_debugstate = _debugstate_valid;
 
-		_pc = pc;
-		_sp = sp;
-		_pp = pp;
-		_objp = objp;
-		_restadjust = restadjust;
-		sciprintf("acc=%04x  ", s->acc & 0xffff);
+		p_pc = pc;
+		p_sp = sp;
+		p_pp = pp;
+		p_objp = objp;
+		p_restadjust = restadjust;
+		sciprintf("%d: acc=%04x  ", script_step_counter, s->acc & 0xffff);
 		_debugstate_valid = 1;
-		disassemble(s, *pc);
+		disassemble(s, *pc, 0, 1);
 
 		_debugstate_valid = old_debugstate;
 
@@ -2491,15 +2623,15 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 	_debugstate_valid = (_debug_step_running == 0);
 
 	if (_debugstate_valid) {
-		_pc = pc;
-		_sp = sp;
-		_pp = pp;
-		_objp = objp;
-		_restadjust = restadjust;
+		p_pc = pc;
+		p_sp = sp;
+		p_pp = pp;
+		p_objp = objp;
+		p_restadjust = restadjust;
 
 		c_debuginfo(s);
 		sciprintf("Step #%d\n", script_step_counter);
-		disassemble(s, *pc);
+		disassemble(s, *pc, 1, 0);
 
 		if (_debug_commands_not_hooked) {
 
@@ -2515,8 +2647,14 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 			con_hook_command(c_heapdump, "heapdump", "ii", "Dumps data from the heap\n"
 					 "\nEXAMPLE\n\n    heapdump 0x1200 42\n\n  Dumps 42 bytes starting at heap address\n"
 					 "  0x1200");
-			con_hook_command(c_disasm, "disasm", "ii*", "Disassembles one or more commands\n\n"
-					 "USAGE\n\n  disasm [startaddr] <number of commands to disassemble>");
+			con_hook_command(c_disasm, "disasm", "is*", "Disassembles one or more commands\n\n"
+					 "USAGE\n\n  disasm [startaddr] <options>\n\n"
+					 "  Valid options are:\n"
+					 "  bwt  : Print byte/word tag\n"
+					 "  c<x> : Disassemble <x> bytes\n"
+					 "  bc   : Print bytecode\n\n"
+					 "  If the start address is specified to be zero, the\n"
+					 "  current program counter is used instead.\n\n");
 			con_hook_command(c_scriptinfo, "scripttable", "", "Displays information about all\n  loaded scripts");
 			con_hook_command(c_heapobj, "heapobj", "i", "Displays information about an\n  object or class on the\n"
 					 "specified heap address.\n\nSEE ALSO\n\n  obj, accobj");
@@ -2547,7 +2685,8 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 					 " for x:\n  u: Unimpl'd/stubbed stuff\n  l: Lists and nodes\n  g: Graphics\n"
 					 "  c: Character handling\n  m: Memory management\n  f: Function call checks\n"
 					 "  b: Bresenham details\n  a: Audio\n  d: System gfx management\n  s: Base setter"
-					 "\n  p: Parser\n  M: The menu system\n  S: Said specs\n  F: File I/O\n  t: GetTime\n"
+					 "\n  p: Parser\n  M: The menu system\n  S: Said specs\n  F: File I/O\n  t: GetTime"
+					 "\n  r: Room numbers\n"
 					 "  *: Everything\n\n"
 					 "  If invoked withour parameters,\n  it will list all activated\n  debug options.\n\n"
 					 "SEE ALSO\n"
@@ -2555,6 +2694,7 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 			con_hook_command(c_visible_map, "set_vismap", "i", "Sets the visible map.\n  Default is 0 (visual).\n"
 					 "  Other useful values are:\n  1: Priority\n  2: Control\n  3: Auxiliary\n");
 			con_hook_command(c_simkey, "simkey", "i", "Simulates a keypress with the\n  specified scancode.\n");
+			con_hook_command(c_statusbar, "statusbar", "ii", "Sets the colors of the status bar. Also controllable from the script.\n");
 			con_hook_command(c_bpx, "bpx", "s", "Sets a breakpoint on the execution of\n  the specified method.\n\n  EXAMPLE:\n"
 					 "  bpx ego::doit\n\n  May also be used to set a breakpoint\n  that applies whenever an object\n"
 					 "  of a specific type is touched:\n  bpx foo::\n");
@@ -2616,6 +2756,7 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 					 "    by songs run from the command line.\n"
 					 "SEE ALSO\n"
 					 "  sndmap\n");
+			con_hook_command(c_sleep, "sleep", "i", "Sleeps for the specified amount of seconds");
 			con_hook_command(c_sndmap, "sndmap", "sii*",
 					 "Changes MT32->GM mappings\n\n"
 					 "USAGE\n\n"
@@ -2675,6 +2816,7 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 					 " or the current port\n  if no port was specified.");
 			con_hook_command(c_gfx_print_visual, "gfx_print_visual", "", "Displays all information about the\n  current widget state");
 			con_hook_command(c_gfx_print_dynviews, "gfx_print_dynviews", "", "Shows the dynview list");
+			con_hook_command(c_gfx_print_dropviews, "gfx_print_dropviews", "", "Shows the list of dropped\n  dynviews");
 			con_hook_command(c_gfx_drawpic, "gfx_drawpic", "ii*", "Draws a pic resource\n\nUSAGE\n  gfx_drawpic <nr> [<pal> [<fl>]]\n"
 					 "  where <nr> is the number of the pic resource\n  to draw\n  <pal> is the optional default\n  palette for the pic (0 is"
 					 "\n  assumed if not specified)\n  <fl> are any pic draw flags (default\n  is 1)");
@@ -2700,6 +2842,10 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 			con_hook_command(c_gfx_priority, "gfx_priority", "i*", "Prints information about priority\n  bands\nUSAGE\n\n  gfx_priority\n\n"
 					 "  will print the min and max values\n  for the priority bands\n\n  gfx_priority <val>\n\n  Print start of the priority\n"
 					 "  band for the specified\n  priority\n");
+#if 0
+#warning "Adlib test: REMOVE ME"
+			con_hook_command(c_adlibtest, "adlibtest", "", "adlib + song iterator test");
+#endif
 
 
 			con_hook_int(&script_debug_flag, "script_debug_flag", "Set != 0 to enable debugger\n");
@@ -2721,57 +2867,27 @@ script_debug(state_t *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *obj
 	if (s->sound_server)
 		(s->sound_server->suspend)(s);
 
-WARNING(fixme!)
-#if 0
-		if (s->onscreen_console) {
-			s->osc_backup = con_backup_screen(s);
-			con_visible_rows = 20;
-
-			con_draw(s, s->osc_backup);
-
-			while (_debugstate_valid) {
-				sci_event_t event = (gfxop_get_event(s->gfx_state));
-				int redraw_console = 0;
-				char *commandbuf;
-
-				if ((event.type & SCI_EVT_KEYBOARD)
-				    || (s->pointer_x != s->last_pointer_x)
-				    || (s->pointer_y != s->last_pointer_y))
-					redraw_console = 1; /* Redraw on keypress or mouse movement */
-
-				if ((event.buckybits & SCI_EVM_CTRL) && (event.data == '`')) /* UnConsole command? */
-					_debugstate_valid = 0;
-				else
-					if ((commandbuf = con_input(&event))) {
-
-						sciprintf(" >%s\n", commandbuf);
-
-						if (strlen(commandbuf) == 0) /* Repeat old command? */
-							strcpy(commandbuf, oldcommand);
-						else /* No, a new valid command */
-							strcpy(oldcommand, commandbuf);
-
-						con_parse(s, commandbuf);
-
-						redraw_console = 1;
-					}
-
-				if (redraw_console)
-					con_draw(s, s->osc_backup);
-			}
-
-			con_restore_screen(s, s->osc_backup);
-
-		} else {
+	while (_debugstate_valid) {
+#ifdef WANT_CONSOLE
+#  ifndef FORCE_CONSOLE
+		if (!have_windowed) {
+#  endif
+			char *input;
+			con_gfx_show(s->gfx_state);
+			input = con_gfx_read(s->gfx_state);
+			con_gfx_hide(s->gfx_state);
+			if (input)
+				con_parse(s, input);
+#  ifndef FORCE_CONSOLE
+		} else
+#  endif
 #endif
-
-			while (_debugstate_valid) {
-				con_parse(s, _debug_get_input());
-				sciprintf("\n");
-			}
-#if 0
-		}
+#ifndef FORCE_CONSOLE
+			con_parse(s, _debug_get_input());
 #endif
+		sciprintf("\n");
+	}
+
 /* Resume music playing */
 	if (s->sound_server)
 		(s->sound_server->resume)(s);
