@@ -1,6 +1,5 @@
 /***************************************************************************
- vm.c Copyright (C) 1999, 2000, 2001, 02 Christoph Reichenbach
-
+ vm.c Copyright (C) 1999, 2000, 2001 Christoph Reichenbach
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -31,10 +30,15 @@
 #include <versions.h>
 #include <kdebug.h>
 
-#if !defined (_WIN32) && !defined (__BEOS__)
+#if !defined (_WIN32) && !defined (_DREAMCAST)
 #include <sys/resource.h>
 #endif
 
+#ifdef HAVE_SETJMP_H
+#include <setjmp.h>
+#endif
+
+
 /*#define VM_DEBUG_SEND*/
 #undef STRICT_SEND /* Disallows variable sends with more than one parameter */
 
@@ -208,9 +212,11 @@ execute_method(state_t *s, word script, word pubfunct, heap_ptr sp,
     magic_ofs=0;
 
   return
-    add_exec_stack_entry(s, scriptpos + GET_HEAP(tableaddress + (pubfunct * 2)) - magic_ofs, sp,
-			 calling_obj, argc, argp, -1, calling_obj, s->execution_stack_pos,
-			 s->scripttable[script].localvar_offset);
+//add_exec_stack_entry(struct _state *s, heap_ptr pc, heap_ptr sp, heap_ptr objp, int argc,
+//		     heap_ptr argp, int selector, heap_ptr sendp, int origin, int localvarp);
+    add_exec_stack_entry(s, (heap_ptr)(scriptpos + GET_HEAP(tableaddress + (pubfunct * 2)) - magic_ofs),
+		sp, calling_obj, (int)argc, argp, -1, calling_obj, s->execution_stack_pos,
+		s->scripttable[script].localvar_offset);
 }
 
 
@@ -493,6 +499,26 @@ add_exec_stack_entry(state_t *s, heap_ptr pc, heap_ptr sp, heap_ptr objp, int ar
 }
 
 
+static int jump_initialized = 0;
+#ifdef HAVE_SETJMP_H
+static jmp_buf vm_error_address;
+#endif
+
+void
+vm_handle_fatal_error(state_t *s, int line, char *file)
+{
+	fprintf(stderr, "Fatal VM error in %s, L%d; aborting...\n", file, line);
+#ifdef HAVE_SETJMP_H
+	if (jump_initialized)
+		longjmp(vm_error_address, 0);
+	else
+#endif
+		{
+			fprintf(stderr, "Could not recover, exitting...\n");
+			exit(1);
+		}
+}
+
 void
 run_vm(state_t *s, int restoring)
 {
@@ -513,6 +539,11 @@ run_vm(state_t *s, int restoring)
     return;
   }
 
+#ifdef HAVE_SETJMP_H
+  setjmp(vm_error_address);
+  jump_initialized = 1;
+#endif
+
 
   if (restoring) {
 
@@ -565,8 +596,7 @@ run_vm(state_t *s, int restoring)
     if (script_abort_flag)
       return; /* Emergency */
 
-    if (script_debug_flag || sci_debug_flags)
-    {
+    if (script_debug_flag || sci_debug_flags) {
       script_debug(s, &(xs->pc), &(xs->sp), &(xs->variables[VAR_TEMP]),
 		   &(xs->objp), &restadjust, bp_flag);
       bp_flag = 0;
@@ -870,8 +900,8 @@ run_vm(state_t *s, int restoring)
       temp = xs->sp;
       xs->sp -= (opparams[0] + (restadjust * 2)); /* Adjust stack */
 
-      xs_new = send_selector(s, s->acc, s->acc, temp, opparams[0],
-			     restadjust, xs->sp);
+      xs_new = send_selector(s, s->acc, s->acc, temp, (int)opparams[0],
+			     (word)restadjust, xs->sp);
 
       if (xs_new)
 	xs = xs_new;
@@ -889,7 +919,7 @@ run_vm(state_t *s, int restoring)
       xs->sp -= (opparams[0] + (restadjust * 2)); /* Adjust stack */
 
       xs_new = send_selector(s, xs->objp, xs->objp, temp,
-			     opparams[0], restadjust, xs->sp);
+			     (int)opparams[0], (word)restadjust, xs->sp);
 
       restadjust = 0;
 
@@ -906,7 +936,7 @@ run_vm(state_t *s, int restoring)
 	xs->sp -= (opparams[1] + (restadjust * 2)); /* Adjust stack */
         kludge = get_class_address(s, opparams[0]);
 	/* kludge necessary due to compiler bugs (egcs 2.91.66, at least) */
-	xs_new = send_selector(s, kludge, xs->objp, temp, opparams[1], restadjust, xs->sp);
+	xs_new = send_selector(s, (heap_ptr)kludge, xs->objp, temp, (int)opparams[1], (word)restadjust, xs->sp);
 	restadjust = 0;
 
 	if (xs_new) xs = xs_new;
@@ -979,14 +1009,16 @@ run_vm(state_t *s, int restoring)
 
     case 0x37: /* ipTos */
       temp2 = xs->objp + SCRIPT_SELECTOR_OFFSET + opparams[0];
-      temp = GET_HEAP(temp2);
-      PUT_HEAP(temp2, temp + 1);
+      temp = GET_HEAP(temp2)+1;
+      PUT_HEAP(temp2, temp);
+      PUSH(temp);
       break;
 
     case 0x38: /* dpTos */
       temp2 = xs->objp + SCRIPT_SELECTOR_OFFSET + opparams[0];
-      temp = GET_HEAP(temp2);
-      PUT_HEAP(temp2, temp - 1);
+      temp = GET_HEAP(temp2)-1;
+      PUT_HEAP(temp2, temp);
+      PUSH(temp);
       break;
 
     case 0x39: /* lofsa */
@@ -1274,7 +1306,7 @@ lookup_selector(state_t *s, heap_ptr obj, int selectorid, heap_ptr *address)
 			}
 
 	return
-		_lookup_selector_functions(s, obj, selectorid, address, speciespos);
+		_lookup_selector_functions(s, obj, selectorid, address, (heap_ptr)speciespos);
 	/* Call recursive function selector seeker */
 }
 
@@ -1376,7 +1408,6 @@ script_instantiate(state_t *s, int script_nr, int recursive)
 	*/
 
 	objlength = 0;
-
 	do {
 		pos += objlength; /* Step over the last checked object */
 
@@ -1388,7 +1419,11 @@ script_instantiate(state_t *s, int script_nr, int recursive)
 
 		else if (objtype == sci_obj_synonyms) {
 			s->scripttable[script_nr].synonyms_offset = pos + 4; /* +4 is to step over the header */
-			s->scripttable[script_nr].synonyms_nr = (objlength - 4) / 4;
+			s->scripttable[script_nr].synonyms_nr = (objlength) / 4;
+			if (GET_HEAP(s->scripttable[script_nr].synonyms_offset +
+				     ((s->scripttable[script_nr].synonyms_nr - 1) << 2)) < 0)
+				/* Adjust for "terminal" synonym entries */
+				--s->scripttable[script_nr].synonyms_nr;
 
 		} else if (objtype == sci_obj_class) {
 			heap_ptr classpos = pos - SCRIPT_OBJECT_MAGIC_OFFSET + 4/* Header */;
@@ -1487,7 +1522,7 @@ script_instantiate(state_t *s, int script_nr, int recursive)
 
 		pos -= 4; /* Step back on header */
 
-	} while ((objtype != 0) && ((pos - script_basepos) < script->size - 2));
+	} while ((objtype != 0) && (((unsigned)pos - script_basepos) < script->size - 2));
 
 	/*    if (script_nr == 0)   sci_hexdump(s->heap + script_basepos +2, script->size-2, script_basepos);*/
 	return s->scripttable[script_nr].heappos;
@@ -1599,10 +1634,10 @@ _game_run(state_t *s, int restoring)
 
 			game_exit(s);
 			game_init(s);
-			putInt16(s->heap + s->stack_base, s->selector_map.play); /* Call the play selector */
+			putInt16(s->heap + s->stack_base, (word)s->selector_map.play); /* Call the play selector */
 
 			putInt16(s->heap + s->stack_base + 2, 0);
-			send_selector(s, s->game_obj, s->game_obj, s->stack_base + 2, 4, 0, s->stack_base);
+			send_selector(s, s->game_obj, s->game_obj, (heap_ptr) (s->stack_base + 2), 4, 0, s->stack_base);
 
 			script_abort_flag = 0;
 			s->restarting_flags = SCI_GAME_WAS_RESTARTED | SCI_GAME_WAS_RESTARTED_AT_LEAST_ONCE;
@@ -1622,9 +1657,9 @@ _game_run(state_t *s, int restoring)
 				if (script_abort_flag == SCRIPT_ABORT_WITH_REPLAY) {
 					sciprintf("Restarting with replay()\n");
 					s->execution_stack_pos = -1; /* Resatart with replay */
-					putInt16(s->heap + s->stack_base, s->selector_map.replay); /* Call the replay selector */
+					putInt16(s->heap + s->stack_base, (word)s->selector_map.replay); /* Call the replay selector */
 					putInt16(s->heap + s->stack_base + 2, 0);
-					send_selector(s, s->game_obj, s->game_obj, s->stack_base + 2, 4, 0, s->stack_base);
+					send_selector(s, s->game_obj, s->game_obj, (heap_ptr) (s->stack_base + 2), 4, 0, s->stack_base);
 				}
 
 				script_abort_flag = 0;
@@ -1644,11 +1679,11 @@ game_run(state_t **_s)
 	state_t *s = *_s;
 
 	sciprintf(" Calling %s::play()\n", s->game_name);
-	putInt16(s->heap + s->stack_base, s->selector_map.play); /* Call the play selector... */
+	putInt16(s->heap + s->stack_base, (word)s->selector_map.play); /* Call the play selector... */
 	putInt16(s->heap + s->stack_base + 2, 0);                    /* ... with 0 arguments. */
 
 	/* Now: Register the first element on the execution stack- */
-	if (!send_selector(s, s->game_obj, s->game_obj, s->stack_base + 2, 4, 0, s->stack_base) || script_error_flag) {
+	if (!send_selector(s, s->game_obj, s->game_obj, (heap_ptr) (s->stack_base + 2), 4, 0, s->stack_base) || script_error_flag) {
 		sciprintf("Failed to run the game! Aborting...\n");
 		return 1;
 	}
@@ -1684,60 +1719,6 @@ game_restore(state_t **_s, char *game_name)
 }
 
 
-void
-version_require_earlier_than(state_t *s, sci_version_t version)
-{
-  if (s->version_lock_flag)
-    return;
-
-  if (version <= s->min_version) {
-    sciprintf("Version autodetect conflict: Less than %d.%03d.%03d was requested, but "
-	      "%d.%03d.%03d is the current minimum\n",
-	      SCI_VERSION_MAJOR(version), SCI_VERSION_MINOR(version), SCI_VERSION_PATCHLEVEL(version),
-	      SCI_VERSION_MAJOR(s->min_version), SCI_VERSION_MINOR(s->min_version),
-	      SCI_VERSION_PATCHLEVEL(s->min_version));
-    return;
-  }
-  else if (version < s->max_version) {
-    s->max_version = version -1;
-    if (s->max_version < s->version)
-      s->version = s->max_version;
-  }
-}
-
-
-void
-version_require_later_than(state_t *s, sci_version_t version)
-{
-  if (s->version_lock_flag)
-    return;
-
-  if (version > s->max_version) {
-    sciprintf("Version autodetect conflict: More than %d.%03d.%03d was requested, but less than"
-	      "%d.%03d.%03d is required ATM\n",
-	      SCI_VERSION_MAJOR(version), SCI_VERSION_MINOR(version), SCI_VERSION_PATCHLEVEL(version),
-	      SCI_VERSION_MAJOR(s->max_version), SCI_VERSION_MINOR(s->max_version),
-	      SCI_VERSION_PATCHLEVEL(s->max_version));
-    return;
-  }
-  else if (version > s->min_version) {
-    s->min_version = version;
-    if (s->min_version > s->version)
-      s->version = s->min_version;
-  }
-}
-
-sci_version_t
-version_parse(char *vn)
-{
-  int major = *vn - '0'; /* One version digit */
-  int minor = atoi(vn + 2);
-  int patchlevel = atoi(vn + 6);
-
-  return SCI_VERSION(major, minor, patchlevel);
-}
-
-
 void
 quit_vm()
 {
