@@ -1,5 +1,5 @@
 /***************************************************************************
- ggi_driver.c Copyright (C) 2000,01,02 Christoph Reichenbach
+ ggi_driver.c Copyright (C) 2000 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -587,12 +587,6 @@ ggi_draw_pixmap(gfx_driver_t *drv, gfx_pixmap_t *pxm, int priority,
 	assert(pri_map);
 	assert(vis);
 
-	if (src.xl != dest.xl
-	    || src.yl != dest.yl) {
-		GFXERROR("Attempt to draw scaled view- not supported by this driver!\n");
-		return GFX_FATAL; /* Scaling not supported! */
-	}
-
 	dbuf = ggiDBGetBuffer(vis, (vis == VISUAL)? buffer : 0);
 	if (!dbuf) {
 		GFXERROR("Could not acquire direct buffer!\n");
