@@ -1,5 +1,5 @@
 /***************************************************************************
- xlib_driver.h Copyright (C) 2000,01,02 Christoph Reichenbach
+ xlib_driver.c Copyright (C) 2000-2005 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -40,19 +40,23 @@
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <X11/extensions/XShm.h>
-#if defined(HAVE_X11_EXTENSIONS_XRENDER_H) && defined(HAVE_X11_XFT_XFT_H)
+#if defined(HAVE_X11_EXTENSIONS_XRENDER_H)
 #  define HAVE_RENDER
 #  include <X11/extensions/Xrender.h>
-#  include <X11/Xft/Xft.h>
 #endif
 #include <errno.h>
 #endif
 
+#ifdef HAVE_XM_MWMUTIL_H
+#include <Xm/MwmUtil.h>
+#endif
+
 #define SCI_XLIB_PIXMAP_HANDLE_NORMAL 0
 #define SCI_XLIB_PIXMAP_HANDLE_GRABBED 1
 
 #define SCI_XLIB_SWAP_CTRL_CAPS (1 << 0)
 #define SCI_XLIB_INSERT_MODE    (1 << 1)
+#define SCI_XLIB_NLS		(1 << 2) /* Non-US keyboard support, sacrificing shortcut keys */
 
 #define X_COLOR_EXT(c) ((c << 8) | c)
 
@@ -67,17 +71,17 @@ struct _xlib_state {
 	Pixmap visual[3];
 	gfx_pixmap_t *priority[2];
 #ifdef HAVE_MITSHM
-        XShmSegmentInfo *shm[4];
+	XShmSegmentInfo *shm[4];
 #endif
 	int use_render;
-#ifdef HAVE_RENDER
-	XftDraw *xftdraw;
-#endif
 	int buckystate;
 	XErrorHandler old_error_handler;
 	Cursor mouse_cursor;
 	byte *pointer_data[2];
-        int used_bytespp; /* bytes actually used to display stuff, rather than bytes occupied in data space */
+	int used_bytespp; /* bytes actually used to display stuff, rather than bytes occupied in data space */
+#ifdef HAVE_RENDER
+	Picture picture;
+#endif
 };
 
 #define S ((struct _xlib_state *)(drv->state))
@@ -100,17 +104,16 @@ int x11_error = 0;
 
 static int check_for_xshm(Display *display)
 {
-  int major, minor, ignore;
-  Bool pixmaps;
-  if (XQueryExtension(display, "MIT-SHM", &ignore, &ignore, &ignore)) {
-    if (XShmQueryVersion( display, &major, &minor, &pixmaps) == True) {
-      return (pixmaps == True) ? 2 : 1 ;
-    } else {
-      return 0;
-    }
-  }
-  return 0;
-
+	int major, minor, ignore;
+	Bool pixmaps;
+	if (XQueryExtension(display, "MIT-SHM", &ignore, &ignore, &ignore)) {
+		if (XShmQueryVersion( display, &major, &minor, &pixmaps) == True) {
+			return (pixmaps == True) ? 2 : 1 ;
+		} else {
+			return 0;
+		}
+	}
+	return 0;
 }
 #endif
 
@@ -200,6 +203,14 @@ xlib_error_handler(Display *display, XErrorEvent *error)
 	return 0;
 }
 
+#define UPDATE_NLS_CAPABILITY 							\
+		if (flags & SCI_XLIB_NLS)					\
+			drv->capabilities |= GFX_CAPABILITY_KEYTRANSLATE;	\
+		else								\
+			drv->capabilities &= ~GFX_CAPABILITY_KEYTRANSLATE
+
+
+
 static int
 xlib_set_parameter(struct _gfx_driver *drv, char *attribute, char *value)
 {
@@ -210,16 +221,30 @@ xlib_set_parameter(struct _gfx_driver *drv, char *attribute, char *value)
 		else
 			flags &= ~SCI_XLIB_SWAP_CTRL_CAPS;
 
+
 		return GFX_OK;
 	}
 
-#ifdef HAVE_MITSHM
+	if (!strncmp(attribute, "localised_keyboard", 18)
+	    || !strncmp(attribute, "localized_keyboard", 18)) {
+		if (string_truep(value))
+			flags |= SCI_XLIB_NLS;
+		else
+			flags &= ~SCI_XLIB_NLS;
+
+		UPDATE_NLS_CAPABILITY;
+
+		return GFX_OK;
+
+	}
+
 	if (!strncmp(attribute, "disable_shmem", 14)) {
+#ifdef HAVE_MITSHM
 	  if (string_truep(value))
 	    have_shmem = -1;
+#endif
 	  return GFX_OK;
 	}
-#endif
 
 	ERROR("Attempt to set xlib parameter \"%s\" to \"%s\"\n", attribute, value);
 	return GFX_ERROR;
@@ -247,6 +272,10 @@ static int
 xlib_draw_filled_rect(struct _gfx_driver *drv, rect_t rect,
                       gfx_color_t color1, gfx_color_t color2,
                       gfx_rectangle_fill_t shade_mode);
+static int
+xlib_draw_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm, int priority,
+		 rect_t src, rect_t dest, gfx_buffer_t buffer);
+
 
 
 static int
@@ -259,18 +288,27 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	int found_vistype = 0;
 	int red_shift, green_shift, blue_shift, alpha_shift;
 	int bytespp_physical;
+	int depth_mod; /* Number of bits to subtract from the depth while checking */
 	unsigned int alpha_mask;
 	int xsize, ysize;
 	XSizeHints *size_hints;
 	XClassHint *class_hint;
-        XImage *foo_image = NULL;
+	XImage *foo_image = NULL;
 	int reverse_endian = 0;
+#ifdef HAVE_XM_MWMUTIL_H
+	PropMotifWmHints motif_hints;
+	Atom prop, proptype;
+#endif
+
 	int i;
 
-	if (!S)
-		S = sci_malloc(sizeof(struct _xlib_state));
 
-	flags = SCI_XLIB_INSERT_MODE;
+	UPDATE_NLS_CAPABILITY;
+
+	if (!drv->state /* = S */)
+		drv->state = sci_malloc(sizeof(struct _xlib_state));
+
+	flags |= SCI_XLIB_INSERT_MODE;
 
 	S->display = XOpenDisplay(NULL);
 
@@ -311,10 +349,18 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	  S->shm[i] = NULL;
 #endif
 
-	while ((((bytespp > 1) && (vistype >= 4))
-		|| ((bytespp == 1) && (vistype == 3)))
-	       && !XMatchVisualInfo(S->display, default_screen, bytespp << 3, vistype, &xvisinfo))
-		vistype--;
+	depth_mod = 0;
+
+	do {
+		while ((((bytespp > 1) && (vistype >= 4))
+			|| ((bytespp == 1) && (vistype == 3)))
+		       && (!(found_vistype = XMatchVisualInfo(S->display, default_screen,
+							     (bytespp << 3) - depth_mod, vistype, &xvisinfo))))
+			vistype--;
+		++depth_mod;
+	} while (!found_vistype
+		 && (bytespp == 1
+		     || (depth_mod <= 7)));
 
 	if (vistype < 3 || ((vistype == 3) && (bytespp != 1))) {
 		if (bytespp == 1) {
@@ -344,6 +390,30 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	}
 
 	XSync(S->display, False);
+#ifdef HAVE_XM_MWMUTIL_H
+	motif_hints.flags = MWM_HINTS_DECORATIONS|MWM_HINTS_FUNCTIONS;
+	motif_hints.decorations = MWM_DECOR_BORDER|MWM_DECOR_TITLE|MWM_DECOR_MENU|MWM_DECOR_MINIMIZE;
+	motif_hints.functions=0
+#ifdef MWM_FUNC_MOVE
+		| MWM_FUNC_MOVE
+#endif
+#ifdef MWM_FUNC_MINIMIZE
+		| MWM_FUNC_MINIMIZE
+#endif
+#ifdef MWM_FUNC_CLOSE
+		| MWM_FUNC_CLOSE
+#endif
+#ifdef MWM_FUNC_QUIT_APP
+		| MWM_FUNC_QUIT_APP
+#endif
+		;
+
+	prop = XInternAtom(S->display, "_MOTIF_WM_HINTS", True );
+	if (prop) {
+		proptype = prop;
+		XChangeProperty(S->display, S->window, prop, proptype, 32, PropModeReplace, (unsigned char *) &motif_hints, PROP_MOTIF_WM_HINTS_ELEMENTS);
+	}
+#endif
 
 	XStoreName(S->display, S->window, "FreeSCI");
 	XDefineCursor(S->display, S->window, (S->mouse_cursor = x_empty_cursor(S->display, S->window)));
@@ -507,6 +577,7 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	      ERROR("ARGH!  Can't attach shared memory segment\n");
 	      have_shmem = 0;
 	    }
+	    XSync(S->display, False);
 	    shmctl(S->shm[i]->shmid, IPC_RMID, 0);
 	  }
 
@@ -543,19 +614,17 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	/** X RENDER handling **/
 #ifdef HAVE_RENDER
 	S->use_render = x_have_render(S->display);
+
 	if (S->use_render) {
-		S->xftdraw = XftDrawCreate(S->display, (Drawable) S->visual[1],
-					   DefaultVisual(S->display,
-							 DefaultScreen(S->display)),
-					   S->colormap);
-
-		if (!S->xftdraw) {
-			S->use_render = 0;
-			ERROR("XftDrawCreate() failed!\n");
-		}
-	}
-
-	if (!S->use_render)
+		XRenderPictFormat * format
+			= XRenderFindVisualFormat (S->display,
+						   DefaultVisual(S->display,
+								 DefaultScreen(S->display)));
+		S->picture =  XRenderCreatePicture (S->display,
+						    (Drawable) S->visual[1],
+						    format,
+						    0, 0);
+	} else /* No Xrender */
 		drv->draw_filled_rect = xlib_draw_filled_rect;
 #else
 	S->use_render = 0;
@@ -578,7 +647,7 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 	}
 #endif
 
-        S->used_bytespp = bytespp;
+	S->used_bytespp = bytespp;
 	S->old_error_handler = (XErrorHandler) XSetErrorHandler(xlib_error_handler);
 	S->pointer_data[0] = NULL;
 	S->pointer_data[1] = NULL;
@@ -591,7 +660,7 @@ xlib_init_specific(struct _gfx_driver *drv, int xfact, int yfact, int bytespp)
 static void
 xlib_xdpy_info()
 {
-        int i;
+	int i;
 	XVisualInfo foo;
 	XVisualInfo *visuals;
 	int visuals_nr;
@@ -669,12 +738,15 @@ xlib_exit(struct _gfx_driver *drv)
 		    XFreePixmap(S->display, S->visual[i]);
 		}
 
+#ifdef HAVE_RENDER
+		XRenderFreePicture(S->display, S->picture);
+#endif
 		XFreeGC(S->display, S->gc);
 		XDestroyWindow(S->display, S->window);
 		XCloseDisplay(S->display);
 		XSetErrorHandler((XErrorHandler) (S->old_error_handler));
 		sci_free(S);
-		S = NULL;
+		drv->state /* = S */ = NULL;
 		gfx_free_mode(drv->mode);
 	}
 }
@@ -684,7 +756,7 @@ xlib_exit(struct _gfx_driver *drv)
 
 static int
 xlib_draw_line(struct _gfx_driver *drv, rect_t line, gfx_color_t color,
-               gfx_line_mode_t line_mode, gfx_line_style_t line_style)
+	       gfx_line_mode_t line_mode, gfx_line_style_t line_style)
 {
 	int linewidth = (line_mode == GFX_LINE_MODE_FINE)? 1:
 		(drv->mode->xfact + drv->mode->yfact) >> 1;
@@ -730,8 +802,8 @@ xlib_draw_line(struct _gfx_driver *drv, rect_t line, gfx_color_t color,
 
 static int
 xlib_draw_filled_rect(struct _gfx_driver *drv, rect_t rect,
-                      gfx_color_t color1, gfx_color_t color2,
-                      gfx_rectangle_fill_t shade_mode)
+		      gfx_color_t color1, gfx_color_t color2,
+		      gfx_rectangle_fill_t shade_mode)
 {
 
 	if (color1.mask & GFX_MASK_VISUAL) {
@@ -754,31 +826,19 @@ xlib_draw_filled_rect_RENDER(struct _gfx_driver *drv, rect_t rect,
 			     gfx_rectangle_fill_t shade_mode)
 {
 	if (color1.mask & GFX_MASK_VISUAL) {
-		XftColor fg;
+		XRenderColor fg;
 
-		fg.color.red = X_COLOR_EXT(color1.visual.r);
-		fg.color.green = X_COLOR_EXT(color1.visual.g);
-		fg.color.blue = X_COLOR_EXT(color1.visual.b);
-		fg.color.alpha = 0xffff - X_COLOR_EXT(color1.alpha);
+		fg.red = X_COLOR_EXT(color1.visual.r);
+		fg.green = X_COLOR_EXT(color1.visual.g);
+		fg.blue = X_COLOR_EXT(color1.visual.b);
+		fg.alpha = 0xffff - X_COLOR_EXT(color1.alpha);
 
-		fg.pixel = xlib_map_color(drv, color1);
-
-#if SUFFICIENTLY_NEW_X11 /* FIXME */
-		XftDrawRect(S->xftdraw, &fg, rect.x, rect.y,
-			    rect.xl, rect.yl);
-#else
-		{ /* Ugly hack: Based on Keith Packard's HelloX patch's hack */
-			XRenderPictFormat *format;
-			Picture            pict;
-
-			format = XRenderFindVisualFormat (S->display, DefaultVisual(S->display, DefaultScreen(S->display)));
-
-			pict =  XRenderCreatePicture (S->display,
-						      (Drawable) S->visual[1], format, 0, 0);
-			XRenderFillRectangle(S->display,PictOpOver,pict,&(&fg)->color,
-					     rect.x, rect.y, rect.xl, rect.yl);
-		}
-#endif
+		XRenderFillRectangle(S->display,
+				     PictOpOver,
+				     S->picture,
+				     &fg,
+				     rect.x, rect.y,
+				     rect.xl, rect.yl);
 	}
 
 	if (color1.mask & GFX_MASK_PRIORITY)
@@ -825,7 +885,7 @@ xlib_unregister_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm)
 
 static int
 xlib_draw_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm, int priority,
-                 rect_t src, rect_t dest, gfx_buffer_t buffer)
+		 rect_t src, rect_t dest, gfx_buffer_t buffer)
 {
 	int bufnr = (buffer == GFX_BUFFER_STATIC)? 2:1;
 	int pribufnr = bufnr -1;
@@ -836,13 +896,13 @@ xlib_draw_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm, int priority,
 		      src.xl, src.yl, dest.xl, dest.yl);
 		return GFX_ERROR;
 	}
-	fflush(stdout);
+
 	if (pxm->internal.handle == SCI_XLIB_PIXMAP_HANDLE_GRABBED) {
 		XPutImage(S->display, S->visual[bufnr], S->gc, (XImage *) pxm->internal.info,
 			  src.x, src.y, dest.x, dest.y, dest.xl, dest.yl);
 		return GFX_OK;
 	}
-	fflush(stdout);
+
 	tempimg = XGetImage(S->display, S->visual[bufnr], dest.x, dest.y,
 			    dest.xl, dest.yl, 0xffffffff, ZPixmap);
 
@@ -855,7 +915,7 @@ xlib_draw_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm, int priority,
 			     (byte *) tempimg->data, tempimg->bytes_per_line,
 			     S->priority[pribufnr]->index_data,
 			     S->priority[pribufnr]->index_xl, 1,
-                             GFX_CROSSBLIT_FLAG_DATA_IS_HOMED);
+			     GFX_CROSSBLIT_FLAG_DATA_IS_HOMED);
 
 	XPutImage(S->display, S->visual[bufnr], S->gc, tempimg,
 		  0, 0, dest.x, dest.y, dest.xl, dest.yl);
@@ -864,9 +924,10 @@ xlib_draw_pixmap(struct _gfx_driver *drv, gfx_pixmap_t *pxm, int priority,
 	return GFX_OK;
 }
 
+
 static int
 xlib_grab_pixmap(struct _gfx_driver *drv, rect_t src, gfx_pixmap_t *pxm,
-                 gfx_map_mask_t map)
+		 gfx_map_mask_t map)
 {
 
 	if (src.x < 0 || src.y < 0) {
@@ -1080,22 +1141,15 @@ x_unmap_key(gfx_driver_t *drv, int keycode)
 {
 	KeySym xkey = XKeycodeToKeysym(S->display, keycode, 0);
 
-
 	return 0;
 }
 
-
 int
-x_map_key(gfx_driver_t *drv, int keycode)
+x_map_key(gfx_driver_t *drv, XEvent *key_event, char *character)
 {
-	KeySym xkey = XKeycodeToKeysym(S->display, keycode, 0);
+        KeySym xkey = XKeycodeToKeysym(S->display, key_event->xkey.keycode, 0);
 
-	if ((xkey >= 'A') && (xkey <= 'Z'))
-		return xkey;
-	if ((xkey >= 'a') && (xkey <= 'z'))
-		return xkey;
-	if ((xkey >= '0') && (xkey <= '9'))
-		return xkey;
+	*character = 0;
 
 	if (flags & SCI_XLIB_SWAP_CTRL_CAPS) {
 		switch (xkey) {
@@ -1104,14 +1158,16 @@ x_map_key(gfx_driver_t *drv, int keycode)
 		}
 	}
 
-	switch (xkey) {
+	switch(xkey) {
+
 	case XK_BackSpace: return SCI_K_BACKSPACE;
 	case XK_Tab: return 9;
 	case XK_Escape: return SCI_K_ESC;
 	case XK_Return:
 	case XK_KP_Enter: return SCI_K_ENTER;
 
-	case XK_KP_Decimal: return SCI_K_DELETE;
+	case XK_KP_Decimal:
+	case XK_KP_Delete: return SCI_K_DELETE;
 	case XK_KP_0:
 	case XK_KP_Insert: return SCI_K_INSERT;
 	case XK_End:
@@ -1164,29 +1220,29 @@ x_map_key(gfx_driver_t *drv, int keycode)
 	case XK_Shift_L:/* S->buckystate |= SCI_EVM_LSHIFT; return 0; */
 	case XK_Shift_R:/* S->buckystate |= SCI_EVM_RSHIFT; return 0; */
 		return 0;
+	default:
+		break;
+	}
 
 
+	if (flags & SCI_XLIB_NLS) {
+		/* Localised key lookup */
+		XLookupString(&(key_event->xkey), character, 1, &xkey, NULL);
+	}
+
+	if ((xkey >= ' ') && (xkey <= '~'))
+		return xkey; /* All printable ASCII characters */
+
+	switch (xkey) {
 	case XK_KP_Add: return '+';
 	case XK_KP_Divide: return '/';
 	case XK_KP_Subtract: return '-';
 	case XK_KP_Multiply: return '*';
-
-	case ',':
-	case '.':
-	case '/':
-	case '\\':
-	case ';':
-	case '\'':
-	case '[':
-	case ']':
-	case '`':
-	case '-':
-	case '=':
-	case '<':
-	case ' ':
-		return xkey;
 	}
 
+	if (*character)
+		return xkey; /* Should suffice for all practical purposes */
+
 	sciprintf("Unknown X keysym: %04x\n", xkey);
 	return 0;
 }
@@ -1202,6 +1258,8 @@ x_get_event(gfx_driver_t *drv, int eventmask, long wait_usec, sci_event_t *sci_e
 	struct timeval ctime, timeout_time, sleep_time;
 	long usecs_to_sleep;
 
+	eventmask |= ExposureMask; /* Always check for this */
+
 	gettimeofday(&timeout_time, NULL);
 	timeout_time.tv_usec += wait_usec;
 
@@ -1210,80 +1268,97 @@ x_get_event(gfx_driver_t *drv, int eventmask, long wait_usec, sci_event_t *sci_e
 	timeout_time.tv_usec %= 1000000;
 
 	do {
+		int hasnext_event = 1;
 		int redraw_pointer_request = 0;
 
-		while (XCheckWindowEvent(display, window, eventmask | ExposureMask, &event)) {
-			switch (event.type) {
+		while (hasnext_event) {
+			if (sci_event) { /* Capable of handling any event? */
+				hasnext_event = XPending(display);
+				if (hasnext_event)
+					XNextEvent(display, &event);
+			} else 
+				hasnext_event = XCheckWindowEvent(display, window, eventmask, &event);
 
-			case ReparentNotify:
-			case ConfigureNotify:
-			case MapNotify:
-			case UnmapNotify:
-				break;
+			
+			if (hasnext_event) 
+				switch (event.type) {
 
-			case KeyPress: {
-				int modifiers = event.xkey.state;
-				sci_event->type = SCI_EVT_KEYBOARD;
+				    case ReparentNotify:
+				    case ConfigureNotify:
+				    case MapNotify:
+				    case UnmapNotify:
+					    break;
 
-				S->buckystate = ((flags & SCI_XLIB_INSERT_MODE)? SCI_EVM_INSERT : 0)
-					| (((modifiers & LockMask)? SCI_EVM_LSHIFT | SCI_EVM_RSHIFT : 0)
-					   | ((modifiers & ControlMask)? SCI_EVM_CTRL : 0)
-					   | ((modifiers & (Mod1Mask | Mod4Mask))? SCI_EVM_ALT : 0)
-					   | ((modifiers & Mod2Mask)? SCI_EVM_NUMLOCK : 0)
-					   | ((modifiers & Mod5Mask)? SCI_EVM_SCRLOCK : 0))
-					^ ((modifiers & ShiftMask)? SCI_EVM_LSHIFT | SCI_EVM_RSHIFT : 0);
+				    case KeyPress: {
+					    int modifiers = event.xkey.state;
+					    char ch = 0;
+					    sci_event->type = SCI_EVT_KEYBOARD;
 
-				sci_event->buckybits = S->buckystate;
-				sci_event->data = x_map_key(drv, event.xkey.keycode);
+					    S->buckystate = ((flags & SCI_XLIB_INSERT_MODE)? SCI_EVM_INSERT : 0)
+						    | (((modifiers & LockMask)? SCI_EVM_LSHIFT | SCI_EVM_RSHIFT : 0)
+						       | ((modifiers & ControlMask)? SCI_EVM_CTRL : 0)
+						       | ((modifiers & (Mod1Mask | Mod4Mask))? SCI_EVM_ALT : 0)
+						       | ((modifiers & Mod2Mask)? SCI_EVM_NUMLOCK : 0)
+						       | ((modifiers & Mod5Mask)? SCI_EVM_SCRLOCK : 0))
+						    ^ ((modifiers & ShiftMask)? SCI_EVM_LSHIFT | SCI_EVM_RSHIFT : 0);
 
-				if (sci_event->data == SCI_K_INSERT)
-					flags ^= SCI_XLIB_INSERT_MODE;
+					    sci_event->buckybits = S->buckystate;
+					    sci_event->data =
+						    x_map_key(drv, &event, &ch);
 
-				if (sci_event->data)
-					return;
+					    if (ch)
+						    sci_event->character = ch;
+					    else
+						    sci_event->character = sci_event->data;
 
-				break;
-			}
+					    if (sci_event->data == SCI_K_INSERT)
+						    flags ^= SCI_XLIB_INSERT_MODE;
 
-			case KeyRelease:
-				/*x_unmap_key(drv, event.xkey.keycode);*/
-				break;
+					    if (sci_event->data)
+						    return;
 
-			case ButtonPress: {
-				sci_event->type = SCI_EVT_MOUSE_PRESS;
-				sci_event->buckybits = S->buckystate;
-				sci_event->data = x_button_xlate[event.xbutton.button];
-				return;
-			}
+					    break;
+				    }
 
-			case ButtonRelease: {
-				sci_event->type = SCI_EVT_MOUSE_RELEASE;
-				sci_event->buckybits = S->buckystate;
-				sci_event->data = x_button_xlate[event.xbutton.button];
-				return;
-			}
+				    case KeyRelease:
+					    /*x_unmap_key(drv, event.xkey.keycode);*/
+					    break;
 
-			case MotionNotify: {
+				    case ButtonPress: {
+					    sci_event->type = SCI_EVT_MOUSE_PRESS;
+					    sci_event->buckybits = S->buckystate;
+					    sci_event->data = x_button_xlate[event.xbutton.button];
+					    return;
+				    }
 
-				drv->pointer_x = event.xmotion.x;
-				drv->pointer_y = event.xmotion.y;
-			}
-			break;
+				    case ButtonRelease: {
+					    sci_event->type = SCI_EVT_MOUSE_RELEASE;
+					    sci_event->buckybits = S->buckystate;
+					    sci_event->data = x_button_xlate[event.xbutton.button];
+					    return;
+				    }
 
-			case GraphicsExpose:
-			case Expose: {
-				XCopyArea(S->display, S->visual[0], S->window, S->gc,
-					  event.xexpose.x, event.xexpose.y, event.xexpose.width, event.xexpose.height,
-					  event.xexpose.x, event.xexpose.y);
-			}
-			break;
+				    case MotionNotify: {
 
-			case NoExpose:
-				break;
+					    drv->pointer_x = event.xmotion.x;
+					    drv->pointer_y = event.xmotion.y;
+				    }
+					    break;
 
-			default:
-				ERROR("Received unhandled X event %04x\n", event.type);
-			}
+				    case GraphicsExpose:
+				    case Expose: {
+					    XCopyArea(S->display, S->visual[0], S->window, S->gc,
+						      event.xexpose.x, event.xexpose.y, event.xexpose.width, event.xexpose.height,
+						      event.xexpose.x, event.xexpose.y);
+				    }
+					    break;
+
+				    case NoExpose:
+					    break;
+
+				    default:
+					    ERROR("Received unhandled X event %04x\n", event.type);
+				}
 		}
 
 		gettimeofday(&ctime, NULL);
@@ -1334,14 +1409,15 @@ xlib_usec_sleep(struct _gfx_driver *drv, long usecs)
 gfx_driver_t
 gfx_driver_xlib = {
 	"xlib",
-	"0.2",
+	"0.6",
 	SCI_GFX_DRIVER_MAGIC,
 	SCI_GFX_DRIVER_VERSION,
 	NULL,
 	0, 0,
 	GFX_CAPABILITY_STIPPLED_LINES | GFX_CAPABILITY_MOUSE_SUPPORT
 	| GFX_CAPABILITY_MOUSE_POINTER | GFX_CAPABILITY_PIXMAP_REGISTRY
-	| GFX_CAPABILITY_PIXMAP_GRABBING | GFX_CAPABILITY_FINE_LINES,
+	| GFX_CAPABILITY_PIXMAP_GRABBING | GFX_CAPABILITY_FINE_LINES
+	| GFX_CAPABILITY_WINDOWED,
 	0/*GFX_DEBUG_POINTER | GFX_DEBUG_UPDATES | GFX_DEBUG_PIXMAPS | GFX_DEBUG_BASIC*/,
 	xlib_set_parameter,
 	xlib_init_specific,
