@@ -1,5 +1,5 @@
 /***************************************************************************
- gfx_test.c Copyright (C) 2000,01,02 Christoph Reichenbach
+ gfx_test.c Copyright (C) 2000 Christoph Reichenbach
 
 
  This program may be modified and copied freely according to the terms of
@@ -44,11 +44,24 @@
 #include <gfx_tools.h>
 #include <gfx_operations.h>
 
-int
-sci0_palette;
+
+int sci0_palette;
+
+gfx_pixmap_color_t gfx_sci0_image_colors[1][16];
+
+gfx_pixmap_color_t *
+gfxr_interpreter_get_static_palette(gfx_resstate_t *state, int version, int *colors_nr, void *internal)
+{
+	return NULL;
+}
+
+
+void sci_sched_yield()
+{
+}
 
 int
-sciprintf(char *fmt, ...)
+sciprintf(const char *fmt, ...)
 {
 	va_list argp;
 	va_start(argp, fmt);
