@@ -1,6 +1,6 @@
-/***************************************************************************
- gfx_operations Copyright (C) 2000,01,02 Christoph Reichenbach
 
+/***************************************************************************
+ gfx_operations Copyright (C) 2000 Christoph Reichenbach
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -29,6 +29,9 @@
 #include <sci_memory.h>
 #include <gfx_operations.h>
 
+
+#define PRECISE_PRIORITY_MAP /* Duplicate all operations on the local priority map as appropriate */
+
 #undef GFXW_DEBUG_DIRTY /* Enable to debug stuff relevant for dirty rects
 			 ** in widget management  */
 
@@ -38,6 +41,19 @@
 #  define DDIRTY if (0) fprintf
 #endif
 
+/* Default color maps */
+#define DEFAULT_COLORS_NR 16
+gfx_pixmap_color_t default_colors[DEFAULT_COLORS_NR] =
+	{{GFX_COLOR_SYSTEM, 0x00, 0x00, 0x00}, {GFX_COLOR_SYSTEM, 0x00, 0x00, 0xaa},
+	 {GFX_COLOR_SYSTEM, 0x00, 0xaa, 0x00}, {GFX_COLOR_SYSTEM, 0x00, 0xaa, 0xaa},
+	 {GFX_COLOR_SYSTEM, 0xaa, 0x00, 0x00}, {GFX_COLOR_SYSTEM, 0xaa, 0x00, 0xaa},
+	 {GFX_COLOR_SYSTEM, 0xaa, 0x55, 0x00}, {GFX_COLOR_SYSTEM, 0xaa, 0xaa, 0xaa},
+	 {GFX_COLOR_SYSTEM, 0x55, 0x55, 0x55}, {GFX_COLOR_SYSTEM, 0x55, 0x55, 0xff},
+	 {GFX_COLOR_SYSTEM, 0x55, 0xff, 0x55}, {GFX_COLOR_SYSTEM, 0x55, 0xff, 0xff},
+	 {GFX_COLOR_SYSTEM, 0xff, 0x55, 0x55}, {GFX_COLOR_SYSTEM, 0xff, 0x55, 0xff},
+	 {GFX_COLOR_SYSTEM, 0xff, 0xff, 0x55}, {GFX_COLOR_SYSTEM, 0xff, 0xff, 0xff}}; /* "Normal" EGA */
+
+
 #define POINTER_VISIBLE_BUT_CLIPPED 2
 
 /* Performs basic checks that apply to most functions */
@@ -158,35 +174,51 @@ _gfxop_grab_pixmap(gfx_state_t *state, gfx_pixmap_t **pxmp, int x, int y,
 					  priority? GFX_MASK_PRIORITY : GFX_MASK_VISUAL);
 }
 
+#define DRAW_LOOP(condition)										\
+{													\
+	rect_t drawrect = gfx_rect(pos.x, pos.y, pxm->index_xl, pxm->index_yl);				\
+	int offset, base_offset;									\
+	int read_offset, base_read_offset;								\
+	int x,y;											\
+													\
+	if (!pxm->index_data) {										\
+		GFXERROR("Attempt to draw control color %d on pixmap %d/%d/%d without index data!\n",	\
+			 color, pxm->ID, pxm->loop, pxm->cel);						\
+		return;											\
+	}												\
+													\
+	if (_gfxop_clip(&drawrect, gfx_rect(0, 0, 320, 200)))						\
+		return;											\
+													\
+	offset = base_offset = drawrect.x + drawrect.y * 320;						\
+	read_offset = base_read_offset = (drawrect.x - pos.x) + ((drawrect.y - pos.y) * pxm->index_xl);	\
+													\
+	for (y = 0; y < drawrect.yl; y++) {								\
+		for (x = 0; x < drawrect.xl; x++)							\
+			if (pxm->index_data[read_offset++] != GFX_COLOR_INDEX_TRANSPARENT) {		\
+				if (condition)								\
+					map->index_data[offset++] = color;				\
+				else									\
+					++offset;							\
+			} else ++offset;								\
+													\
+		offset = base_offset += 320;								\
+		read_offset = base_read_offset += pxm->index_xl;					\
+	}												\
+}
+
 static void
 _gfxop_draw_control(gfx_pixmap_t *map, gfx_pixmap_t *pxm, int color, point_t pos)
-{
-	rect_t drawrect = gfx_rect(pos.x, pos.y, pxm->index_xl, pxm->index_yl);
-	int offset, base_offset;
-	int read_offset, base_read_offset;
-	int x,y;
+DRAW_LOOP(1) /* Always draw */
 
-	if (!pxm->index_data) {
-		GFXERROR("Attempt to draw control color %d on pixmap %d/%d/%d without index data!\n",
-			 color, pxm->ID, pxm->loop, pxm->cel);
-		return;
-	}
+#ifdef PRECISE_PRIORITY_MAP
+static void
+_gfxop_draw_priority(gfx_pixmap_t *map, gfx_pixmap_t *pxm, int color, point_t pos)
+DRAW_LOOP(map->index_data[offset] < color) /* Draw only lower priority */
+#endif
 
-	if (_gfxop_clip(&drawrect, gfx_rect(0, 0, 320, 200)))
-		return;
+#undef DRAW_LOOP
 
-	offset = base_offset = drawrect.x + drawrect.y * 320;
-	read_offset = base_read_offset = (drawrect.x - pos.x) + ((drawrect.y - pos.y) * pxm->index_xl);
-
-	for (y = 0; y < drawrect.yl; y++) {
-		for (x = 0; x < drawrect.xl; x++)
-			if (pxm->index_data[read_offset++] < GFX_COLOR_INDEX_TRANSPARENT)
-				map->index_data[offset++] = color;
-
-		offset = base_offset += 320;
-		read_offset = base_read_offset += pxm->index_xl;
-	}
-}
 
 static int
 _gfxop_install_pixmap(gfx_driver_t *driver, gfx_pixmap_t *pxm)
@@ -233,14 +265,24 @@ _gfxop_install_pixmap(gfx_driver_t *driver, gfx_pixmap_t *pxm)
 
 static int
 _gfxop_draw_pixmap(gfx_driver_t *driver, gfx_pixmap_t *pxm, int priority, int control,
-		   rect_t src, rect_t dest, rect_t clip, int static_buf, gfx_pixmap_t *control_map)
+		   rect_t src, rect_t dest, rect_t clip, int static_buf, gfx_pixmap_t *control_map,
+		   gfx_pixmap_t *priority_map)
 {
 	int error;
 	rect_t clipped_dest = gfx_rect(dest.x, dest.y, dest.xl, dest.yl);
-	if (control >= 0)
-		_gfxop_draw_control(control_map, pxm, control,
-				    gfx_point(dest.x / driver->mode->xfact,
-					      dest.y / driver->mode->yfact));
+
+	if (control >= 0 || priority >= 0) {
+		point_t original_pos = gfx_point(dest.x / driver->mode->xfact,
+						     dest.y / driver->mode->yfact);
+
+		if (control >= 0)
+			_gfxop_draw_control(control_map, pxm, control, original_pos);
+
+#ifdef PRECISE_PRIORITY_MAP
+		if (priority >= 0)
+			_gfxop_draw_priority(priority_map, pxm, priority, original_pos);
+#endif
+	}
 
 	if (_gfxop_clip(&clipped_dest, clip))
 		return GFX_OK;
@@ -253,7 +295,7 @@ _gfxop_draw_pixmap(gfx_driver_t *driver, gfx_pixmap_t *pxm, int priority, int co
 	error = _gfxop_install_pixmap(driver, pxm);
 	if (error) return error;
 
-	DDIRTY(stderr, "\\-> Drawing to actual %d %d %d %d\n", 
+	DDIRTY(stderr, "\\-> Drawing to actual %d %d %d %d\n",
 	       clipped_dest.x / driver->mode->xfact,
 	       clipped_dest.y / driver->mode->yfact,
 	       clipped_dest.xl / driver->mode->xfact,
@@ -358,7 +400,7 @@ _gfxop_draw_pointer(gfx_state_t *state)
 					   gfx_rect(0, 0, ppxm->xl, ppxm->yl),
 					   gfx_rect(x, y, ppxm->xl, ppxm->yl),
 					   gfx_rect(0, 0, xfact * 320 , yfact * 200),
-					   0, state->control_map);
+					   0, state->control_map, state->priority_map);
 
 		if (error)
 			return error;
@@ -515,7 +557,11 @@ _gfxop_clear_dirty_rec(gfx_state_t *state, struct _dirty_rect *rect)
 	fprintf(stderr, "\tClearing dirty (%d %d %d %d)\n",
 		GFX_PRINT_RECT(rect->rect));
 #endif
-	retval = _gfxop_update_box(state, rect->rect);
+	if (!state->fullscreen_override)
+		retval = _gfxop_update_box(state, rect->rect);
+	else
+		retval = GFX_OK;
+
 	retval |= _gfxop_clear_dirty_rec(state, rect->next);
 
 	free(rect);
@@ -525,6 +571,14 @@ _gfxop_clear_dirty_rec(gfx_state_t *state, struct _dirty_rect *rect)
 
 /*** Exported operations ***/
 
+static void
+init_aux_pixmap(gfx_pixmap_t **pixmap)
+{
+	*pixmap = gfx_pixmap_alloc_index_data(gfx_new_pixmap(320, 200, GFX_RESID_NONE, 0, 0));
+	(*pixmap)->flags |= GFX_PIXMAP_FLAG_EXTERNAL_PALETTE;
+	(*pixmap)->colors_nr = DEFAULT_COLORS_NR;
+	(*pixmap)->colors = default_colors;
+}
 
 static int
 _gfxop_init_common(gfx_state_t *state, gfx_options_t *options, void *misc_payload)
@@ -563,8 +617,11 @@ _gfxop_init_common(gfx_state_t *state, gfx_options_t *options, void *misc_payloa
 
 	state->mouse_pointer = state->mouse_pointer_bg = NULL;
 	state->mouse_pointer_visible = 0;
-	state->control_map = gfx_pixmap_alloc_index_data(gfx_new_pixmap(320, 200, GFX_RESID_NONE, 0, 0));
-	state->control_map->flags |= GFX_PIXMAP_FLAG_EXTERNAL_PALETTE;
+
+	init_aux_pixmap(&(state->control_map));
+	init_aux_pixmap(&(state->priority_map));
+	init_aux_pixmap(&(state->static_priority_map));
+
 	state->options = options;
 	state->mouse_pointer_in_hw = 0;
 	state->disable_dirty = 0;
@@ -635,6 +692,16 @@ gfxop_exit(gfx_state_t *state)
 		state->control_map = NULL;
 	}
 
+	if (state->priority_map) {
+		gfx_free_pixmap(state->driver, state->priority_map);
+		state->priority_map = NULL;
+	}
+
+	if (state->static_priority_map) {
+		gfx_free_pixmap(state->driver, state->static_priority_map);
+		state->static_priority_map = NULL;
+	}
+
 	if (state->mouse_pointer_bg) {
 		gfx_free_pixmap(state->driver, state->mouse_pointer_bg);
 		state->mouse_pointer_bg = NULL;
@@ -645,6 +712,13 @@ gfxop_exit(gfx_state_t *state)
 }
 
 
+int
+gfxop_have_mouse(gfx_state_t *state)
+{
+	return state->driver->capabilities & GFX_CAPABILITY_MOUSE_SUPPORT;
+}
+
+
 static int
 _gfxop_scan_one_bitmask(gfx_pixmap_t *pixmap, rect_t zone)
 {
@@ -677,13 +751,17 @@ gfxop_scan_bitmask(gfx_state_t *state, rect_t area, gfx_map_mask_t map)
 	gfxr_pic_t *pic = (state->pic_unscaled)? state->pic_unscaled : state->pic;
 	int retval = 0;
 
-	_gfxop_clip(&area, gfx_rect(0, 0, 320, 200));
+	_gfxop_clip(&area, gfx_rect(0, 10, 320, 200));
+
+	if (area.xl <= 0
+	    || area.yl <= 0)
+		return 0;
 
 	if (map & GFX_MASK_VISUAL)
 		retval |= _gfxop_scan_one_bitmask(pic->visual_map, area);
 
 	if (map & GFX_MASK_PRIORITY)
-		retval |= _gfxop_scan_one_bitmask(pic->priority_map, area);
+		retval |= _gfxop_scan_one_bitmask(state->priority_map, area);
 	if (map & GFX_MASK_CONTROL)
 		retval |= _gfxop_scan_one_bitmask(state->control_map, area);
 
@@ -692,22 +770,36 @@ gfxop_scan_bitmask(gfx_state_t *state, rect_t area, gfx_map_mask_t map)
 
 
 int
-gfxop_set_visible_map(gfx_state_t *state, gfx_map_mask_t map)
+gfxop_set_visible_map(gfx_state_t *state, gfx_map_mask_t visible_map)
 {
-	BASIC_CHECKS(GFX_ERROR);
+	switch (visible_map) {
 
-	if (map != GFX_MASK_VISUAL
-	    && map != GFX_MASK_PRIORITY
-	    && map != GFX_MASK_CONTROL) {
-		GFXWARN("Attempt to set invalid visible map #%d\n", map);
+	case GFX_MASK_VISUAL:
+		state->fullscreen_override = NULL;
+		if (visible_map != state->visible_map) {
+			rect_t rect = gfx_rect(0, 0, 320, 200);
+			gfxop_clear_box(state, rect);
+			gfxop_update_box(state, rect);
+		}
+		break;
+
+	case GFX_MASK_PRIORITY:
+		state->fullscreen_override = state->priority_map;
+		break;
+
+	case GFX_MASK_CONTROL:
+		state->fullscreen_override = state->control_map;
+		break;
+
+	default:
+		fprintf(stderr, "Invalid display map %d selected!\n", visible_map);
 		return GFX_ERROR;
 	}
 
-	state->visible_map = map;
+	state->visible_map = visible_map;
 	return GFX_OK;
 }
 
-
 #define MIN_X 0
 #define MIN_Y 0
 #define MAX_X 319
@@ -786,7 +878,7 @@ gfxop_set_color(gfx_state_t *state, gfx_color_t *color, int r, int g, int b, int
 				if (error_code < 0) {
 					GFXWARN("Could not get color entry for %02x/%02x/%02x\n", r, g, b);
 					return error_code;
-				} else if ((error_code = state->driver->set_palette(state->driver, pixmap_color.global_index, r, g, b))) {
+				} else if ((error_code = state->driver->set_palette(state->driver, pixmap_color.global_index, (byte) r, (byte) g, (byte) b))) {
 					GFXWARN("Graphics driver failed to set color index %d to (%02x/%02x/%02x)\n",
 						pixmap_color.global_index, r, g, b);
 					return error_code;
@@ -909,24 +1001,24 @@ line_clip(rect_t *line, rect_t clip, int xfact, int yfact)
 
 	} else { /* "normal" line */
 		float start = 0.0, end = 1.0;
-		float xv = 1.0 * line->xl;
-		float yv = 1.0 * line->yl;
+		float xv = (float) line->xl;
+		float yv = (float) line->yl;
 
 		if (line->xl < 0)
-			clip_line_partial(&start, &end, 1.0 / xv, 1.0 * line->x, 1.0 * (clip.x + clip.xl), 1.0 * clip.x);
+			clip_line_partial(&start, &end, (float) (1.0 / xv), (float) line->x, (float) (clip.x + clip.xl), (float) clip.x);
 		else
-			clip_line_partial(&start, &end, 1.0 / xv, 1.0 * line->x, 1.0 * clip.x, 1.0 * (clip.x + clip.xl));
+			clip_line_partial(&start, &end, (float) (1.0 / xv), (float) line->x, (float) clip.x, (float) (clip.x + clip.xl));
 
 		if (line->yl < 0)
-			clip_line_partial(&start, &end, 1.0 / yv, 1.0 * line->y, 1.0 * (clip.y + clip.yl), 1.0 * clip.y);
+			clip_line_partial(&start, &end, (float) (1.0 / yv), (float) line->y, (float) (clip.y + clip.yl), (float) clip.y);
 		else
-			clip_line_partial(&start, &end, 1.0 / yv, 1.0 * line->y, 1.0 * clip.y, 1.0 * (clip.y + clip.yl));
+			clip_line_partial(&start, &end, (float) (1.0 / yv), (float) line->y, (float) clip.y, (float) (clip.y + clip.yl));
 
-		line->x += (int) xv * start;
-		line->y += (int) yv * start;
+		line->x += (int) (xv * start);
+		line->y += (int) (yv * start);
 
-		line->xl = (int) xv * (end-start);
-		line->yl = (int) yv * (end-start);
+		line->xl = (int) (xv * (end-start));
+		line->yl = (int) (yv * (end-start));
 
 		return (start > 1.0 || end < 0.0);
 	}
@@ -1207,15 +1299,15 @@ gfxop_draw_box(gfx_state_t *state, rect_t box, gfx_color_t color1, gfx_color_t c
 	case GFX_BOX_SHADE_LEFT: reverse = 1;
 	case GFX_BOX_SHADE_RIGHT:
 		driver_shade_type = GFX_SHADE_HORIZONTALLY;
-		mod_offset = ((new_box.x - box.x) * 1.0) / (box.xl * 1.0);
-		mod_breadth = (new_box.xl * 1.0) / (box.xl * 1.0);
+		mod_offset = (float) (((new_box.x - box.x) * 1.0) / (box.xl * 1.0));
+		mod_breadth = (float) ((new_box.xl * 1.0) / (box.xl * 1.0));
 		break;
 
 	case GFX_BOX_SHADE_UP: reverse = 1;
 	case GFX_BOX_SHADE_DOWN:
 		driver_shade_type = GFX_SHADE_VERTICALLY;
-		mod_offset = ((new_box.y - box.y) * 1.0) / (box.yl * 1.0);
-		mod_breadth = (new_box.yl * 1.0) / (box.yl * 1.0);
+		mod_offset = (float) (((new_box.y - box.y) * 1.0) / (box.yl * 1.0));
+		mod_breadth = (float) ((new_box.yl * 1.0) / (box.yl * 1.0));
 		break;
 
 	default:
@@ -1225,7 +1317,7 @@ gfxop_draw_box(gfx_state_t *state, rect_t box, gfx_color_t color1, gfx_color_t c
 
 
 	if (reverse)
-		mod_offset = 1.0 - (mod_offset + mod_breadth);
+		mod_offset = (float) (1.0 - (mod_offset + mod_breadth));
 	/* Reverse offset if we have to interpret colors inversely */
 
 	if (shade_type == GFX_BOX_SHADE_FLAT)
@@ -1240,17 +1332,17 @@ gfxop_draw_box(gfx_state_t *state, rect_t box, gfx_color_t color1, gfx_color_t c
 		draw_color1.priority = draw_color2.priority = color1.priority;
 
 		if (draw_color1.mask & GFX_MASK_VISUAL) {
-			draw_color1.visual.r = COLOR_MIX(visual.r, mod_offset);
-			draw_color1.visual.g = COLOR_MIX(visual.g, mod_offset);
-			draw_color1.visual.b = COLOR_MIX(visual.b, mod_offset);
-			draw_color1.alpha = COLOR_MIX(alpha, mod_offset);
+			draw_color1.visual.r = (guint8) COLOR_MIX(visual.r, mod_offset);
+			draw_color1.visual.g = (guint8) COLOR_MIX(visual.g, mod_offset);
+			draw_color1.visual.b = (guint8) COLOR_MIX(visual.b, mod_offset);
+			draw_color1.alpha = (guint8) COLOR_MIX(alpha, mod_offset);
 
 			mod_offset += mod_breadth;
 
-			draw_color2.visual.r = COLOR_MIX(visual.r, mod_offset);
-			draw_color2.visual.g = COLOR_MIX(visual.g, mod_offset);
-			draw_color2.visual.b = COLOR_MIX(visual.b, mod_offset);
-			draw_color2.alpha = COLOR_MIX(alpha, mod_offset);
+			draw_color2.visual.r = (guint8) COLOR_MIX(visual.r, mod_offset);
+			draw_color2.visual.g = (guint8) COLOR_MIX(visual.g, mod_offset);
+			draw_color2.visual.b = (guint8) COLOR_MIX(visual.b, mod_offset);
+			draw_color2.alpha = (guint8) COLOR_MIX(alpha, mod_offset);
 		}
 		if (reverse)
 			return drv->draw_filled_rect(drv, new_box, draw_color2, draw_color1, driver_shade_type);
@@ -1298,6 +1390,13 @@ gfxop_clear_box(gfx_state_t *state, rect_t box)
 	    && (sci0_palette == 1)) {
 		BREAKPOINT();
 	}
+
+	_gfxop_clip(&box, gfx_rect(0, 0, 320, 200));
+#ifdef PRECISE_PRIORITY_MAP
+	if (state->pic_unscaled)
+		gfx_copy_pixmap_box_i(state->priority_map, state->static_priority_map, box);
+#endif
+
 	_gfxop_scale_rect(&box, state->driver->mode);
 
 	return _gfxop_buffer_propagate_box(state, box, GFX_BUFFER_BACK);
@@ -1316,6 +1415,14 @@ gfxop_update(gfx_state_t *state)
 
 	state->dirty_rects = NULL;
 
+	if (state->fullscreen_override) {
+		/* We've been asked to re-draw the active full-screen image, essentially. */
+		rect_t rect = gfx_rect(0, 0, 320, 200);
+		gfx_xlate_pixmap(state->fullscreen_override, state->driver->mode, GFX_XLATE_FILTER_NONE);
+		gfxop_draw_pixmap(state, state->fullscreen_override, rect, gfx_point(0,0));
+		retval |= _gfxop_update_box(state, rect);
+	}
+
 	if (retval) {
 		GFXERROR("Clearing the dirty rectangles failed!\n");
 	}
@@ -1541,7 +1648,7 @@ gfxop_set_pointer_view(gfx_state_t *state, int nr, int loop, int cel)
 	new_pointer = _gfxr_get_cel(state, nr, &real_loop, &real_cel);
 
 
-	if (!state->mouse_pointer) {
+	if (!state->mouse_pointer || (real_loop == -1)) {
 		GFXWARN("Attempt to set invalid pointer #%d\n", nr);
 	} else if (real_loop != loop || real_cel != cel) {
 		GFXDEBUG("Changed loop/cel from %d/%d to %d/%d in view %d\n",
@@ -1564,6 +1671,107 @@ gfxop_set_pointer_position(gfx_state_t *state, point_t pos)
 	return 0;
 }
 
+#define SCANCODE_ROWS_NR 3
+
+struct {
+	int offset;
+	char *keys;
+} scancode_rows[SCANCODE_ROWS_NR] = {
+	{0x10, "QWERTYUIOP[]"},
+	{0x1e, "ASDFGHJKL;'\\"},
+	{0x2c, "ZXCVBNM,./"}
+};
+
+static int
+_gfxop_scancode(int ch)
+     /* Calculates a PC keyboard scancode from a character */
+{
+	int row;
+	int c = toupper((char)ch);
+
+	for (row = 0; row < SCANCODE_ROWS_NR; row++) {
+		char *keys = scancode_rows[row].keys;
+		int offset = scancode_rows[row].offset;
+
+		while (*keys) {
+			if (*keys == c)
+				return offset << 8;
+
+			offset++;
+			keys++;
+		}
+	}
+
+	return ch; /* not found */
+}
+
+static int
+_gfxop_shiftify(int c)
+{
+	char shifted_numbers[] = ")!@#$%^&*(";
+	
+	if (c < 256)
+	{
+		c = toupper((char)c);
+
+		if (c >= 'A' && c <= 'Z')
+			return c;
+
+		if (c >= '0' && c <= '9')
+			return shifted_numbers[c-'0'];
+	
+		switch (c) {
+		case SCI_K_TAB: return SCI_K_SHIFT_TAB;
+		case ']': return '}';
+		case '[': return '{';
+		case '`': return '~';
+		case '-': return '_';
+		case '=': return '+';
+		case ';': return ':';
+		case '\'': return '"';
+		case '\\': return '|';
+		case ',': return '<';
+		case '.': return '>';
+		case '/': return '?';
+		default: return c; /* No match */
+		}
+	}
+
+	switch (c)
+	{
+	case SCI_K_F1 : return SCI_K_SHIFT_F1;
+	case SCI_K_F2 : return SCI_K_SHIFT_F2;
+	case SCI_K_F3 : return SCI_K_SHIFT_F3;
+	case SCI_K_F4 : return SCI_K_SHIFT_F4;
+	case SCI_K_F5 : return SCI_K_SHIFT_F5;
+	case SCI_K_F6 : return SCI_K_SHIFT_F6;
+	case SCI_K_F7 : return SCI_K_SHIFT_F7;
+	case SCI_K_F8 : return SCI_K_SHIFT_F8;
+	case SCI_K_F9 : return SCI_K_SHIFT_F9;
+	case SCI_K_F10 : return SCI_K_SHIFT_F10;
+	}
+}
+
+static int
+_gfxop_numlockify(int c)
+{
+	switch (c) {
+	case SCI_K_DELETE: return '.';
+	case SCI_K_INSERT: return '0';
+	case SCI_K_END: return '1';
+	case SCI_K_DOWN: return '2';
+	case SCI_K_PGDOWN: return '3';
+	case SCI_K_LEFT: return '4';
+	case SCI_K_CENTER: return '5';
+	case SCI_K_RIGHT: return '6';
+	case SCI_K_HOME: return '7';
+	case SCI_K_UP: return '8';
+	case SCI_K_PGUP: return '9';
+	default: return c; /* Unchanged */
+	}
+}
+
+
 sci_event_t
 gfxop_get_event(gfx_state_t *state, unsigned int mask)
 {
@@ -1591,6 +1799,10 @@ gfxop_get_event(gfx_state_t *state, unsigned int mask)
 			if (event.type) {
 				*seekerp = sci_malloc(sizeof(gfx_input_event_t));
 				(*seekerp)->next = NULL;
+
+				event.data = (char) (event.data);
+				/* Clip illegal bits */
+
 				(*seekerp)->event = event;
 				seekerp = &((*seekerp)->next);
 			}
@@ -1603,6 +1815,30 @@ gfxop_get_event(gfx_state_t *state, unsigned int mask)
 		GFXERROR("Failed to update the mouse pointer!\n");
 		return error_event;
 	}
+
+	if (event.type == SCI_EVT_KEYBOARD
+	    && !(state->driver->capabilities & GFX_CAPABILITY_KEYTRANSLATE)) {
+		/* Do we still have to translate the key? */
+
+		event.character = event.data;
+
+		/* Scancodify if appropriate */
+		if (event.buckybits & SCI_EVM_ALT)
+			event.character = _gfxop_scancode(event.character);
+
+		/* Shift if appropriate */
+		else if (((event.buckybits & (SCI_EVM_RSHIFT | SCI_EVM_LSHIFT))
+			  && !(event.buckybits & SCI_EVM_CAPSLOCK))
+			 ||
+			 (!(event.buckybits & (SCI_EVM_RSHIFT | SCI_EVM_LSHIFT))
+			  && (event.buckybits & SCI_EVM_CAPSLOCK)))
+			event.character = _gfxop_shiftify(event.character);
+
+		/* Numlockify if appropriate */
+		else if (event.buckybits & SCI_EVM_NUMLOCK)
+			event.data = _gfxop_numlockify(event.data);
+	}
+
 	return event;
 }
 
@@ -1647,7 +1883,10 @@ gfxop_lookup_view_get_cels(gfx_state_t *state, int nr, int loop)
 		GFXWARN("Loop number was corrected from %d to %d in view %d\n", loop, real_loop, nr);
 	}
 
-	return view->loops[real_loop].cels_nr;
+	if (loop == -1)
+		return 0;
+	else
+		return view->loops[real_loop].cels_nr;
 }
 
 
@@ -1661,7 +1900,10 @@ gfxop_check_cel(gfx_state_t *state, int nr, int *loop, int *cel)
 		return GFX_ERROR;
 	}
 
-	return GFX_OK;
+	if (*loop == -1)
+		return GFX_ERROR;
+	else
+		return GFX_OK;
 }
 
 int
@@ -1700,6 +1942,13 @@ gfxop_get_cel_parameters(gfx_state_t *state, int nr, int loop, int cel,
 		return GFX_ERROR;
 	}
 
+	if (loop == -1) {
+		*width = 0;
+		*height = 0;
+		offset-> x = offset->y = 0;
+		return GFX_OK;
+	}
+
 	pxm = view->loops[loop].cels[cel];
 	*width = pxm->index_xl;
 	*height = pxm->index_yl;
@@ -1725,6 +1974,10 @@ _gfxop_draw_cel_buffer(gfx_state_t *state, int nr, int loop, int cel,
 		GFXWARN("Attempt to draw loop/cel %d/%d in invalid view %d\n", loop, cel, nr);
 		return GFX_ERROR;
 	}
+
+	if (loop == -1)
+		return GFX_OK;
+
 	pxm = view->loops[loop].cels[cel];
 
 	old_x = pos.x -= pxm->xoffset;
@@ -1740,7 +1993,10 @@ _gfxop_draw_cel_buffer(gfx_state_t *state, int nr, int loop, int cel,
 				  gfx_rect(0, 0, pxm->xl, pxm->yl),
 				  gfx_rect(pos.x, pos.y, pxm->xl, pxm->yl),
 				  state->clip_zone,
-				  static_buf , state->control_map);
+				  static_buf , state->control_map,
+				  static_buf
+				  ? state->static_priority_map
+				  : state->priority_map);
 }
 
 
@@ -1784,35 +2040,13 @@ gfxop_draw_cel_static_clipped(gfx_state_t *state, int nr, int loop, int cel,
 static int
 _gfxop_set_pic(gfx_state_t *state)
 {
-	gfx_pixmap_t *pxm = NULL;
 	byte unscaled = (state->options->pic0_unscaled);
 	gfx_copy_pixmap_box_i(state->control_map, state->pic->control_map, gfx_rect(0, 0, 320, 200));
-	state->control_map->colors_nr = state->pic->control_map->colors_nr;
-	state->control_map->colors = state->pic->control_map->colors;
+	gfx_copy_pixmap_box_i(state->priority_map, state->pic_unscaled->priority_map, gfx_rect(0, 0, 320, 200));
+	gfx_copy_pixmap_box_i(state->static_priority_map, state->pic_unscaled->priority_map, gfx_rect(0, 0, 320, 200));
 
-
-	switch (state->visible_map) {
-
-	case GFX_MASK_VISUAL:
-		pxm = unscaled? state->pic_unscaled->visual_map : state->pic->visual_map;
-		break;
-
-	case GFX_MASK_PRIORITY:
-		pxm = state->pic->priority_map;
-		break;
-
-	case GFX_MASK_CONTROL:
-		pxm = state->pic->control_map;
-		break;
-
-
-	default:
-		GFXERROR("Attempt to draw invalid map #%d!\n", state->visible_map);
-		return GFX_ERROR;
-	}
-
-	_gfxop_install_pixmap(state->driver, pxm);
-	return state->driver->set_static_buffer(state->driver, pxm, state->pic->priority_map);
+	_gfxop_install_pixmap(state->driver, state->pic->visual_map);
+	return state->driver->set_static_buffer(state->driver, state->pic->visual_map, state->pic->priority_map);
 }
 
 
@@ -1832,12 +2066,12 @@ gfxop_new_pic(gfx_state_t *state, int nr, int flags, int default_palette)
 	state->tag_mode = 1;
 	state->palette_nr = default_palette;
 
-	state->pic = gfxr_get_pic(state->resstate, nr, state->visible_map, flags, default_palette, 1);
+	state->pic = gfxr_get_pic(state->resstate, nr, GFX_MASK_VISUAL, flags, default_palette, 1);
 
 	if (state->driver->mode->xfact == 1 && state->driver->mode->yfact == 1)
 		state->pic_unscaled = state->pic;
 	else
-		state->pic_unscaled = gfxr_get_pic(state->resstate, nr, state->visible_map, flags, default_palette, 0);
+		state->pic_unscaled = gfxr_get_pic(state->resstate, nr, GFX_MASK_VISUAL, flags, default_palette, 0);
 
 	if (!state->pic || !state->pic_unscaled) {
 		GFXERROR("Could not retreive background pic %d!\n", nr);
@@ -1870,12 +2104,12 @@ gfxop_add_to_pic(gfx_state_t *state, int nr, int flags, int default_palette)
 	}
 
 	if (!(state->pic = gfxr_add_to_pic(state->resstate, state->pic_nr, nr,
-					   state->visible_map, flags, state->palette_nr, default_palette, 1))) {
+					   GFX_MASK_VISUAL, flags, state->palette_nr, default_palette, 1))) {
 		GFXERROR("Could not add pic #%d to pic #%d!\n", state->pic_nr, nr);
 		return GFX_ERROR;
 	}
 	state->pic_unscaled = gfxr_add_to_pic(state->resstate, state->pic_nr, nr,
-					      state->visible_map, flags,
+					      GFX_MASK_VISUAL, flags,
 					      state->palette_nr,
 					      default_palette, 1);
 
@@ -2005,11 +2239,11 @@ gfxop_new_text(gfx_state_t *state, int font_nr, char *text, int maxwidth,
 
 			for (j = 0; j < i; j++)
 				gfx_free_pixmap(state->driver, handle->text_pixmaps[j]);
-			free(handle->text_pixmaps);
-			free(handle->text);
-			free(handle->lines);
-			free(handle);
+			sci_free(handle->text_pixmaps);
+			sci_free(handle->text);
+			sci_free(handle->lines);
 			GFXERROR("Failed to draw text pixmap for line %d/%d\n", i, handle->lines_nr);
+			sci_free(handle);
 			return NULL;
 		}
 	}
@@ -2126,7 +2360,7 @@ gfxop_draw_text(gfx_state_t *state, gfx_text_handle_t *handle, rect_t zone)
 
 		_gfxop_draw_pixmap(state->driver, pxm, handle->priority, handle->control,
 				   gfx_rect(0, 0, pxm->xl, pxm->yl), pos, state->clip_zone, 0,
-				   state->control_map);
+				   state->control_map, state->priority_map);
 
 		pos.y += line_height;
 	}
@@ -2182,7 +2416,7 @@ gfxop_draw_pixmap(gfx_state_t *state, gfx_pixmap_t *pxm, rect_t zone, point_t po
 
 	return _gfxop_draw_pixmap(state->driver, pxm, -1, -1, zone, target,
 				  gfx_rect(0, 0, 320*state->driver->mode->xfact,
-					   200*state->driver->mode->yfact), 0, NULL);
+					   200*state->driver->mode->yfact), 0, NULL, NULL);
 }
 
 int
