@@ -1,6 +1,5 @@
 /***************************************************************************
- pic_0.c Copyright (C) 2000,01,02 Christoph Reichenbach
-
+ sci_pic_0.c Copyright (C) 2000 Christoph Reichenbach
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -33,14 +32,11 @@
 #include <gfx_tools.h>
 
 #undef GFXR_DEBUG_PIC0 /* Enable to debug pic0 messages */
+#undef FILL_RECURSIVE_DEBUG /* Enable for verbose fill debugging */
 
 #define GFXR_PIC0_PALETTE_SIZE 40
 #define GFXR_PIC0_NUM_PALETTES 4
 
-#undef FILL_RECURSIVE_DEBUG /* Enable for verbose fill debugging */
-
-#define INTERCOL(a, b) ((int) sqrt((((3.3 * (a))*(a)) + ((1.7 * (b))*(b))) / 5.0))
-/* Macro for color interpolation */
 
 #define SCI_PIC0_MAX_FILL 30 /* Number of times to fill before yielding to scheduler */
 
@@ -49,6 +45,8 @@
 int sci0_palette = 0;
 
 
+#define INTERCOL(a, b) ((int) sqrt((((3.3 * (a))*(a)) + ((1.7 * (b))*(b))) / 5.0))
+/* Macro for color interpolation */
 /* Default color maps */
 gfx_pixmap_color_t gfx_sci0_image_colors[SCI0_MAX_PALETTE+1][GFX_SCI0_IMAGE_COLORS_NR] = {
 	{{GFX_COLOR_SYSTEM, 0x00, 0x00, 0x00}, {GFX_COLOR_SYSTEM, 0x00, 0x00, 0xaa},
@@ -104,15 +102,12 @@ gfx_pixmap_color_t gfx_sci0_pic_colors[GFX_SCI0_PIC_COLORS_NR]; /* Initialized d
 
 static int _gfxr_pic0_colors_initialized = 0;
 
-/************************************/
-/************************************/
+#ifdef FILL_RECURSIVE_DEBUG
 /************************************/
 int fillc = 100000000;
-int fillrc = 3000000;
 int fillmagc = 30000000;
 /************************************/
-/************************************/
-/************************************/
+#endif
 
 void
 gfxr_init_static_palette()
@@ -319,6 +314,7 @@ _gfxr_auxbuf_line_clear(gfxr_pic_t *pic, rect_t line, int color)
 #undef LINEMACRO
 
 
+#ifdef WITH_PIC_SCALING
 static void
 _gfxr_auxbuf_propagate_changes(gfxr_pic_t *pic, int bitmask)
 {
@@ -344,159 +340,7 @@ _gfxr_auxbuf_propagate_changes(gfxr_pic_t *pic, int bitmask)
 		++data;
 	}
 }
-
-
-static void
-_gfxr_auxbuf_fill_helper(gfxr_pic_t *pic, int old_xl, int old_xr, int y, int dy,
-			 int clipmask, int control)
-{
-	int xl, xr;
-	int oldytotal = y * 320;
-	unsigned char fillmask = 0x78;
-
-	do {
-		int ytotal = oldytotal + (320 * dy);
-		int xcont;
-		int state;
-
-		y += dy;
-
-		if (y < SCI_TITLEBAR_SIZE || y > 199)
-			return;
-
-		xl = old_xl;
-		if (!(pic->aux_map[ytotal + xl] & clipmask)) { /* go left */
-			while (xl && !(pic->aux_map[ytotal + xl - 1] & clipmask))
-				--xl;
-		} else /* go right and look for the first valid spot */
-			while ((xl <= old_xr) && (pic->aux_map[ytotal + xl] & clipmask))
-				++xl;
-
-		if (xl > old_xr) /* No fillable strip above the last one */
-			return;
-
-		if ((ytotal + xl) < 0) { fprintf(stderr,"AARGH-%d\n", __LINE__); BREAKPOINT(); }
-
-		xr = xl;
-		while (xr < 320 && !(pic->aux_map[ytotal + xr] & clipmask)) {
-			pic->aux_map[ytotal + xr] |= fillmask;
-			++xr;
-		}
-
-		if ((ytotal + xr) > 64000) { fprintf(stderr,"AARGH-%d\n", __LINE__);
-		BREAKPOINT();
-		}
-
-		--xr;
-
-		if (xr < xl)
-			return;
-
-		/* Check whether we need to recurse on branches in the same direction */
-		if ((y > SCI_TITLEBAR_SIZE && dy < 0)
-		    || (y < 199 && dy > 0)) {
-
-			state = 0;
-			xcont = xr + 1;
-			while (xcont <= old_xr) {
-				if (pic->aux_map[ytotal + xcont] & clipmask)
-					state = 0;
-				else if (!state) { /* recurse */
-					state = 1;
-					_gfxr_auxbuf_fill_helper(pic, xcont, old_xr,
-								 y - dy, dy, clipmask, control);
-				}
-				++xcont;
-			}
-		}
-
-		/* Check whether we need to recurse on backward branches: */
-		/* left */
-		if (xl < old_xl - 1) {
-			state = 0;
-			for (xcont = old_xl - 1; xcont >= xl; xcont--) {
-				if (pic->aux_map[oldytotal + xcont] & clipmask)
-					state = xcont;
-				else if (state) { /* recurse */
-					_gfxr_auxbuf_fill_helper(pic, xcont, state,
-								 y, -dy, clipmask, control);
-					state = 0;
-				}
-			}
-		}
-
-		/* right */
-		if (xr > old_xr + 1) {
-			state = 0;
-			for (xcont = old_xr + 1; xcont <= xr; xcont++) {
-				if (pic->aux_map[oldytotal + xcont] & clipmask)
-					state = xcont;
-				else if (state) { /* recurse */
-					_gfxr_auxbuf_fill_helper(pic, state, xcont,
-								 y, -dy, clipmask, control);
-					state = 0;
-				}
-			}
-		}
-
-		if ((ytotal + xl) < 0) { fprintf(stderr,"AARGH-%d\n", __LINE__); BREAKPOINT() }
-		if ((ytotal + xr+1) > 64000) { fprintf(stderr,"AARGH-%d\n", __LINE__); BREAKPOINT(); }
-
-		if (control)
-			memset(pic->control_map->index_data + ytotal + xl, control, xr-xl+1);
-
-		oldytotal = ytotal;
-		old_xr = xr;
-		old_xl = xl;
-
-	} while (1);
-}
-
-
-static void
-_gfxr_auxbuf_fill(gfxr_pic_t *pic, int x, int y, int clipmask, int control)
-{
-	/* Fills the aux buffer and the control map (if control != 0) */
-	int xl, xr;
-	unsigned char fillmask = 0x78;
-	int ytotal = y * 320;
-
-	if (clipmask & 1)
-		clipmask = 1; /* vis */
-	else if (clipmask & 2)
-		clipmask = 2; /* pri */
-	else if (clipmask & 4)
-		clipmask = 4; /* ctl */
-	else return;
-
-	clipmask |= fillmask; /* Bits 3-5 */
-
-	if (pic->aux_map[ytotal + x] & clipmask)
-		return;
-
-	pic->aux_map[ytotal + x] |= fillmask;
-
-	xl = x;
-	while (xl && !(pic->aux_map[ytotal + xl - 1] & clipmask)) {
-		--xl;
-		pic->aux_map[ytotal + xl] |= fillmask;
-	}
-
-	xr = x;
-	while ((xr < 319) && !(pic->aux_map[ytotal + xr + 1] & clipmask)) {
-		++xr;
-		pic->aux_map[ytotal + xr] |= fillmask;
-	}
-
-	if (control) /* Draw the same strip on the control map */
-		memset(pic->control_map->index_data + ytotal + xl, control, xr - xl + 1);
-
-	if (y > SCI_TITLEBAR_SIZE)
-		_gfxr_auxbuf_fill_helper(pic, xl, xr, y, -1, clipmask, control);
-
-	if (y < 199)
-		_gfxr_auxbuf_fill_helper(pic, xl, xr, y, +1, clipmask, control);
-}
+#endif
 
 
 static inline void
@@ -525,6 +369,7 @@ _gfxr_auxbuf_spread(gfxr_pic_t *pic, int *min_x, int *min_y, int *max_x, int *ma
 	*max_x = *max_y = -1;
 	*min_x = *min_y = 320;
 
+#ifdef FILL_RECURSIVE_DEBUG
 	if (!fillmagc) {
 		fprintf(stderr,"------------------------------------------------\n");
 		fprintf(stderr,"LineID:   ");
@@ -532,6 +377,7 @@ _gfxr_auxbuf_spread(gfxr_pic_t *pic, int *min_x, int *min_y, int *max_x, int *ma
 			fprintf(stderr,"  %d       ", i);
 		fprintf(stderr,"\n");
 	}
+#endif
 
 	for (y = 10; y < 200; y++) {
 		int ivi = y & 1; /* InterVal Index: Current intervals; !ivi is the list of old ones */
@@ -627,20 +473,17 @@ _gfxr_auxbuf_spread(gfxr_pic_t *pic, int *min_x, int *min_y, int *max_x, int *ma
 					}
 					i++;
 				}
-#if 0
-				if (!found_interval && y) /* No 'parent' interval? */
-					_gfxr_auxbuf_tag_line(pic, pos + xl - 320, xr - xl + 1);
-#endif
-
 				width = 0;
 			}
 
+#ifdef FILL_RECURSIVE_DEBUG
 		if (!fillmagc && intervals_nr) {
 			fprintf(stderr,"AI L#%03d:", y);
 			for (j = 0; j < intervals_nr; j++)
 				fprintf(stderr, "%c[%03d,%03d]", intervals[ivi][j].tag? ' ':'-', intervals[ivi][j].xl, intervals[ivi][j].xr);
 			fprintf(stderr,"\n");
 		}
+#endif
 
 		if (intervals_nr) {
 			if (y < *min_y)
@@ -821,12 +664,12 @@ _gfxr_auxplot_brush(gfxr_pic_t *pic, byte *buffer, int yoffset, int offset, int
 
 			srand(randseed);
 
-			x -= (sizex * rand()*1.0)/(RAND_MAX + 1.0);
-			x += (sizex * rand()*1.0)/(RAND_MAX + 1.0);
-			y -= (sizey * rand()*1.0)/(RAND_MAX + 1.0);
-			y += (sizey * rand()*1.0)/(RAND_MAX + 1.0);
-			sizex = (sizex * rand()*1.0)/(RAND_MAX + 1.0);
-			sizey = (sizey * rand()*1.0)/(RAND_MAX + 1.0);
+			x -= (int) ((sizex * rand()*1.0)/(RAND_MAX + 1.0));
+			x += (int) ((sizex * rand()*1.0)/(RAND_MAX + 1.0));
+			y -= (int) ((sizey * rand()*1.0)/(RAND_MAX + 1.0));
+			y += (int) ((sizey * rand()*1.0)/(RAND_MAX + 1.0));
+			sizex = (int) ((sizex * rand()*1.0)/(RAND_MAX + 1.0));
+			sizey = (int) ((sizey * rand()*1.0)/(RAND_MAX + 1.0));
 
 			_gfxr_fill_ellipse(pic, buffer, line_width, x, y, pic->mode->xfact >> 1, pic->mode->yfact >> 1,
 					   color, ELLIPSE_SOLID);
@@ -1246,227 +1089,8 @@ _gfxr_draw_line(gfxr_pic_t *pic, int x, int y, int ex, int ey, int color,
 
 #define IS_FILL_BOUNDARY(x) (((x) & legalmask) != legalcolor)
 
-static void
-_gfxr_fill_recursive(gfxr_pic_t *pic, int old_xl, int old_xr, int y, int dy, byte *bounds,
-		     int legalcolor, int legalmask, int color, int priority, int drawenable)
-{
-	int linewidth = pic->mode->xfact * 320;
-	int miny = pic->mode->yfact * SCI_TITLEBAR_SIZE;
-	int maxy = pic->mode->yfact * 200;
-	int xl, xr;
-	int oldytotal = y * linewidth;
-	int old_proj_y = -42;
-	int proj_y;
-	int proj_ytotal;
-	int proj_x;
-	int proj_xl_bound = 0;
-	int proj_xr_bound = 0;
-
-	do {
-		int ytotal = oldytotal + (linewidth * dy);
-		int xcont;
-		int state;
-
-		y += dy;
-		proj_y = y / pic->mode->yfact;
-
-		if (!fillc)
-			return;
-		else if (!fillmagc) { --fillc;
-		}
-
-		if (y < miny || y >= maxy) {
-#ifdef FILL_RECURSIVE_DEBUG
-			if (!fillmagc)
-				fprintf(stderr,"ABRT on failed initial assertion!\n");
-#endif
-			return;
-		}
-
-		if (proj_y != old_proj_y) {
-			/* First, find the projected coordinates, unless known already: */
-			proj_ytotal = proj_y * 320;
-			proj_x = old_xl / pic->mode->xfact;
-
-			proj_xl_bound = proj_x;
-			if (pic->aux_map[proj_ytotal + proj_xl_bound] & FRESH_PAINT) {
-				while (proj_xl_bound &&
-				       (pic->aux_map[proj_ytotal + proj_xl_bound - 1] & FRESH_PAINT))
-					--proj_xl_bound;
-			} else {
-				while (proj_xl_bound < 319 &&
-				       !(pic->aux_map[proj_ytotal + proj_xl_bound + 1] & FRESH_PAINT))
-					++proj_xl_bound;
-
-				if (proj_xl_bound < 319)
-					++proj_xl_bound;
-			}
-
-			if (proj_xl_bound == 319 && !(pic->aux_map[proj_ytotal + proj_xl_bound] & FRESH_PAINT)) {
-#ifdef FILL_RECURSIVE_DEBUG
-				if (!fillmagc)
-					fprintf(stderr,"ABRT because proj_xl_bound couldn't be found\n");
-#endif
-				return;
-			}
-
-			proj_xr_bound = (proj_xl_bound > proj_x)? proj_xl_bound : proj_x;
-			while ((proj_xr_bound < 319) &&
-			       (pic->aux_map[proj_ytotal + proj_xr_bound + 1] & FRESH_PAINT))
-				++proj_xr_bound;
-
-			if (!fillmagc) {
-				fprintf(stderr,"l%d: {%d,%d} | ", proj_y, proj_xl_bound, proj_xr_bound);
-				pic->aux_map[proj_y*320 + proj_xl_bound] |= 0x2;
-				pic->aux_map[proj_y*320 + proj_xr_bound] |= 0x2;
-			}
-
-			proj_xl_bound *= pic->mode->xfact;
-			if (proj_xl_bound)
-				proj_xl_bound -= pic->mode->xfact - 1;
-
-			if (proj_xr_bound < 319)
-				++proj_xr_bound;
-			proj_xr_bound *= pic->mode->xfact;
-			proj_xr_bound += pic->mode->xfact -1;
-
-			old_proj_y = proj_y;
-		}
-
-		/* Now we have the projected limits, get the real ones: */
-
-		xl = (old_xl > proj_xl_bound)? old_xl : proj_xl_bound;
-		if (!IS_FILL_BOUNDARY(bounds[ytotal + xl])) { /* go left as far as possible */
-			while (xl > proj_xl_bound && (!IS_FILL_BOUNDARY(bounds[ytotal + xl - 1])))
-				--xl;
-		} else /* go right until the fillable area starts */
-			while (xl < proj_xr_bound && (IS_FILL_BOUNDARY(bounds[ytotal + xl])))
-				++xl;
-
-
-#ifdef FILL_RECURSIVE_DEBUG
-		if (!fillmagc)
-			fprintf(stderr,"<%d,", xl);
-#endif
-
-		if ((xl > proj_xr_bound)
-		    || (xl > old_xr)) {
-#ifdef FILL_RECURSIVE_DEBUG
-			if (!fillmagc)
-				fprintf(stderr,"ABRT because xl > xr_bound\n");
-#endif
-			return;
-		}
-
-		xr = (xl > old_xl)? xl : old_xl;
-		while (xr < proj_xr_bound && (!IS_FILL_BOUNDARY(bounds[ytotal + xr + 1])))
-			++xr;
-
-#ifdef FILL_RECURSIVE_DEBUG
-		if (!fillmagc)
-			fprintf(stderr,"%d> -> ", xr);
-#endif
-
-		if (IS_FILL_BOUNDARY(bounds[ytotal + xl])) {
-#ifdef FILL_RECURSIVE_DEBUG
-			if (!fillmagc)
-				fprintf(stderr,"ABRT because xl illegal\n");
-#endif
-			return;
-		}
-
-		if (!fillmagc)
-			fprintf(stderr,"[%d[%d,%d]%d]\n", proj_xl_bound, xl, xr, proj_xr_bound);
-
-		if (xl < proj_xl_bound && xr - 3*pic->mode->xfact < proj_xl_bound) {
-#ifdef FILL_RECURSIVE_DEBUG
-			if (!fillmagc)
-				fprintf(stderr,"ABRT interval left of zone\n");
-#endif
-			return;
-		}
-
-		if (xr > proj_xr_bound && xl + 3*pic->mode->xfact > proj_xr_bound) {
-#ifdef FILL_RECURSIVE_DEBUG
-			if (!fillmagc)
-				fprintf(stderr,"ABRT because interval right of zone\n");
-#endif
-			return;
-		}
-
-		if (drawenable & GFX_MASK_VISUAL)
-			memset(pic->visual_map->index_data + ytotal + xl, color, xr - xl + 1);
-
-		if (drawenable & GFX_MASK_PRIORITY)
-			memset(pic->priority_map->index_data + ytotal + xl, priority, xr - xl + 1);
-
-
-		/* Check whether we need to recurse on branches in the same direction */
-		state = 0;
-		xcont = xr + 1;
-		while (xcont <= old_xr) {
-			if (IS_FILL_BOUNDARY(bounds[ytotal + xcont]))
-				state = xcont;
-			else if (state) { /* recurse */
-#ifdef FILL_RECURSIVE_DEBUG
-				if (!fillmagc)
-					fprintf(stderr, "[%d[%d,%d],%d]: rec BRANCH %d [%d,%d] l%d\n", old_xl, xl, xr, old_xr, dy,
-						state, xcont, y -dy);
-#endif
-				_gfxr_fill_recursive(pic, state, xcont, y - dy, dy, bounds, legalcolor,
-						     legalmask, color, priority, drawenable);
-				state = 0;
-			}
-			++xcont;
-		}
-
-		/* Check whether we need to recurse on backward branches: */
-		/* left */
-		if (xl < old_xl - 1) {
-			state = 0;
-			for (xcont = old_xl-1; xcont >= xl; xcont--) {
-				if (IS_FILL_BOUNDARY(bounds[oldytotal + xcont]))
-					state = xcont;
-				else if (state) { /* recurse */
-#ifdef FILL_RECURSIVE_DEBUG
-					if (!fillmagc)
-						fprintf(stderr, "[%d[%d,%d],%d]: rec BACK-LEFT %d [%d,%d] l%d\n", old_xl, xl, xr, old_xr, -dy,
-							xcont, state, y);
-#endif
-					_gfxr_fill_recursive(pic, xcont, state, y, -dy, bounds,
-							     legalcolor, legalmask, color, priority, drawenable);
-					state = 0;
-				}
-			}
-		}
-
-		/* right */
-		if (xr > old_xr + 1) {
-			state = 0;
-			for (xcont = old_xr + 1; xcont <= xr; xcont++) {
-				if (IS_FILL_BOUNDARY(bounds[oldytotal + xcont]))
-					state = xcont;
-				else if (state) { /* recurse */
-#ifdef FILL_RECURSIVE_DEBUG
-					if (!fillmagc)
-						fprintf(stderr, "[%d[%d,%d],%d]: rec BACK-RIGHT %d [%d,%d] l%d\n", old_xl, xl, xr, old_xr, -dy,
-							state, xcont, y);
-#endif
-					_gfxr_fill_recursive(pic, state, xcont, y, -dy, bounds,
-							     legalcolor, legalmask, color, priority, drawenable);
-					state = 0;
-				}
-			}
-		}
-
-		oldytotal = ytotal;
-		old_xl = xl;
-		old_xr = xr;
-
-	} while (1);
-}
-
 
+#ifdef WITH_PIC_SCALING
 
 #define TEST_POINT(xx, yy) \
   if (pic->aux_map[(yy)*320 + (xx)] & FRESH_PAINT) { \
@@ -1550,129 +1174,33 @@ _gfxr_find_fill_point(gfxr_pic_t *pic, int min_x, int min_y, int max_x, int max_
 
 #undef TEST_POINT
 
-static void
-_gfxr_fill(gfxr_pic_t *pic, int x_320, int y_200, int color, int priority, int control, int drawenable)
-{
-	int linewidth = pic->mode->xfact * 320;
-	int x, y;
-	int xl, xr;
-	int ytotal;
-	int bitmask;
-	byte *bounds = NULL;
-	int legalcolor, legalmask;
-	int min_x, min_y, max_x, max_y;
+/* Now include the actual filling code (with scaling support) */
+#define FILL_FUNCTION _gfxr_fill_any
+#define FILL_FUNCTION_RECURSIVE _gfxr_fill_any_recursive
+#define AUXBUF_FILL_HELPER _gfxr_auxbuf_fill_any_recursive
+#define AUXBUF_FILL _gfxr_auxbuf_fill_any
+#define DRAW_SCALED
+#  include "sci_picfill_aux.c"
+#  include "sci_picfill.c"
+#undef DRAW_SCALED
+#undef AUXBUF_FILL
+#undef AUXBUF_FILL_HELPER
+#undef FILL_FUNCTION_RECURSIVE
+#undef FILL_FUNCTION
 
-	if (pic->control_map->index_data[y_200 * 320 + x_320] != 0)
-		drawenable &= ~GFX_MASK_CONTROL;
+#endif /* defined(WITH_PIC_SCALING) */
 
-	if (color == 0xff)
-		drawenable &= ~GFX_MASK_VISUAL;
-
-	if (priority == 0)
-		drawenable &= ~GFX_MASK_PRIORITY;
-
-	_gfxr_auxbuf_fill(pic, x_320, y_200, drawenable, (drawenable & GFX_MASK_CONTROL)? control : 0);
-	_gfxr_auxbuf_spread(pic, &min_x, &min_y, &max_x, &max_y);
-
-	if (_gfxr_find_fill_point(pic, min_x, min_y, max_x, max_y, x_320, y_200, color, drawenable, &x, &y)) {
-		/* GFXWARN("Could not find scaled fill point, but unscaled fill point was available!\n"); */
-		drawenable &= GFX_MASK_PRIORITY;
-		if (!drawenable)
-			_gfxr_auxbuf_propagate_changes(pic, 0);
-	}
-
-	ytotal = y * linewidth;
-
-	if (!drawenable)
-		return;
-
-	if (drawenable & GFX_MASK_VISUAL) {
-		bounds = pic->visual_map->index_data;
-		if ((color & 0xf) == 0xf /* When dithering with white, do more
-					 ** conservative checks  */
-		    || (color & 0xf0) == 0xf0)
-			legalcolor = 0xff;
-		else
-			legalcolor = 0xf0; /* Only check the second color */
-
-		legalmask = legalcolor;
-	} else if (drawenable & GFX_MASK_PRIORITY) {
-		bounds = pic->priority_map->index_data;
-		legalcolor = 0;
-		legalmask = 0xf;
-	} else {
-		legalcolor = 0;
-		legalmask = 0xf;
-	}
-
-	if (!bounds || IS_FILL_BOUNDARY(bounds[ytotal + x]))
-		return;
-
-	if (bounds) {
-		int proj_y = y_200;
-		int proj_ytotal = proj_y * 320;
-		int proj_x = x_320;
-		int proj_xl_bound;
-		int proj_xr_bound;
-		int proj_xl, proj_xr;
-
-		ytotal = y * linewidth;
-
-		proj_xl_bound = proj_x;
-		if (pic->aux_map[proj_ytotal + proj_xl_bound] & FRESH_PAINT) {
-			while (proj_xl_bound &&
-			       (pic->aux_map[proj_ytotal + proj_xl_bound - 1] & FRESH_PAINT))
-				--proj_xl_bound;
-		} else
-			while (proj_xl_bound < 319 &&
-			       !(pic->aux_map[proj_ytotal + proj_xl_bound + 1] & FRESH_PAINT))
-				++proj_xl_bound;
-
-		proj_xr_bound = (proj_xl_bound > proj_x)? proj_xl_bound : proj_x;
-		while ((proj_xr_bound < 319) &&
-		       (pic->aux_map[proj_ytotal + proj_xr_bound + 1] & FRESH_PAINT))
-			++proj_xr_bound;
-
-		proj_xl = proj_xl_bound;
-		proj_xr = proj_xr_bound;
-
-		proj_xl_bound *= pic->mode->xfact;
-		if (proj_xl_bound)
-			proj_xl_bound -= pic->mode->xfact -1;
-
-		if (proj_xr_bound < 319)
-			++proj_xr_bound;
-		proj_xr_bound *= pic->mode->xfact;
-		proj_xr_bound += pic->mode->xfact -1;
-
-		xl = x;
-		while (xl > proj_xl_bound && (!IS_FILL_BOUNDARY(bounds[ytotal + xl -1])))
-			--xl;
-
-		while (x < proj_xr_bound && (!IS_FILL_BOUNDARY(bounds[ytotal + x + 1])))
-			++x;
-		xr = x;
-
-		if (drawenable & GFX_MASK_VISUAL)
-			memset(pic->visual_map->index_data + ytotal + xl, color, xr - xl + 1);
-
-		if (drawenable & GFX_MASK_PRIORITY)
-			memset(pic->priority_map->index_data + ytotal + xl, priority, xr - xl + 1);
-
-		_gfxr_fill_recursive(pic, xl, xr, y, -1, bounds, legalcolor, legalmask, color, priority, drawenable);
-		_gfxr_fill_recursive(pic, xl, xr, y, +1, bounds, legalcolor, legalmask, color, priority, drawenable);
-	}
-
-	/* Now finish the aux buffer */
-	bitmask = drawenable &
-		(
-			((color != 0xff)? 1 : 0)
-			| ((priority)? 2 : 0)
-			| ((control)? 4 : 0)
-			);
-	if (fillmagc)
-		_gfxr_auxbuf_propagate_changes(pic, bitmask);
-}
+/* Include again, but this time without support for scaling */
+#define FILL_FUNCTION _gfxr_fill_1
+#define FILL_FUNCTION_RECURSIVE _gfxr_fill_1_recursive
+#define AUXBUF_FILL_HELPER _gfxr_auxbuf_fill_1_recursive
+#define AUXBUF_FILL _gfxr_auxbuf_fill_1
+#  include "sci_picfill_aux.c"
+#  include "sci_picfill.c"
+#undef AUXBUF_FILL
+#undef AUXBUF_FILL_HELPER
+#undef FILL_FUNCTION_RECURSIVE
+#undef FILL_FUNCTION
 
 
 #define GET_ABS_COORDS(x, y) \
@@ -1841,11 +1369,11 @@ gfxr_remove_artifacts_pic0(gfxr_pic_t *dest, gfxr_pic_t *src)
 			for (sub_y = 0; sub_y < bound_y; sub_y++) {
 				for (sub_x = 0; sub_x < bound_x; sub_x++) {
 					check_and_remove_artifact(dest->visual_map->index_data + write_offset,
-								  src_visualp, 0xff,
-								  x_320, x_320<319, y_200>10, y_200<199);
+								  src_visualp, (int)0xff,
+								  (byte)x_320, (byte)(x_320 < 319), (byte)(y_200 > 10), (byte)(y_200 < 199));
 					check_and_remove_artifact(dest->priority_map->index_data + write_offset,
 								  src_priorityp, 0,
-								  x_320, x_320<319, y_200>10, y_200<199);
+								  (byte)x_320, (byte)(x_320 < 319), (byte)(y_200 > 10), (byte)(y_200 < 199));
 					++write_offset;
 				}
 				write_offset += scaled_line_size - bound_x;
@@ -1895,7 +1423,7 @@ gfxr_draw_pic0(gfxr_pic_t *pic, int fill_normally, int default_palette, int size
 	int fill_count = 0;
 	byte op, opx;
 
-#ifdef GFXR_DEBUG_PIC0
+#ifdef FILL_RECURSIVE_DEBUG
 	fillmagc = atoi(getenv("FOO"));
 	fillc = atoi(getenv("FOO2"));
 #endif /* GFXR_DEBUG_PIC0 */
@@ -1927,11 +1455,6 @@ gfxr_draw_pic0(gfxr_pic_t *pic, int fill_normally, int default_palette, int size
 			}
 
 			color = palette[pal][index];
-#if 0
-#ifdef GFXR_DEBUG_PIC0
-			color &= 0x77;
-#endif
-#endif
 			p0printf("  color <- %02x [%d/%d]\n", color, pal, index);
 			drawenable |= GFX_MASK_VISUAL;
 			break;
@@ -2037,14 +1560,20 @@ gfxr_draw_pic0(gfxr_pic_t *pic, int fill_normally, int default_palette, int size
 				GET_ABS_COORDS(x, y);
 				p0printf("Abs coords %d,%d\n", x, y);
 				/*fprintf(stderr,"C=(%d,%d)\n", x, y + SCI_TITLEBAR_SIZE);*/
-				_gfxr_fill(pic, x, y + SCI_TITLEBAR_SIZE, (fill_normally)? color : 0, priority, control, drawenable);
+#ifdef WITH_PIC_SCALING
+				if (pic->mode->xfact > 1
+				    || pic->mode->yfact > 1)
+					_gfxr_fill_any(pic, x, y + SCI_TITLEBAR_SIZE, (fill_normally)? color : 0, priority, control, drawenable);
+				else
+#endif
+				_gfxr_fill_1(pic, x, y + SCI_TITLEBAR_SIZE, (fill_normally)? color : 0, priority, control, drawenable);
 
 				if (fill_count++ > SCI_PIC0_MAX_FILL) {
 					sci_sched_yield();
 					fill_count = 0;
 				}
 
-#ifdef GFXR_DEBUG_PIC0
+#ifdef  FILL_RECURSIVE_DEBUG
 				if (!fillmagc) {
 					int x,y;
 					if (getenv("FOO1"))
@@ -2227,7 +1756,10 @@ gfxr_draw_pic0(gfxr_pic_t *pic, int fill_normally, int default_palette, int size
 
 				pri_table = pic->internal;
 
-				for (i = 0; i < 16; i++)
+				pri_table[0] = 0;
+				pri_table[15] = 190;
+
+				for (i = 1; i < 15; i++)
 					pri_table[i] = resource[pos++];
 			}
 			break;
