@@ -1,7 +1,6 @@
 /***************************************************************************
  gfx_driver.h Copyright (C) 2000 Christoph Reichenbach
 
-
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
  notice and the licensing information contained herein are preserved.
@@ -50,6 +49,8 @@ typedef enum {
 #define GFX_CAPABILITY_MOUSE_SUPPORT (1<<7)
 #define GFX_CAPABILITY_POINTER_PIXMAP_REGISTRY (1<<8)
 #define GFX_CAPABILITY_FINE_LINES (1<<9)
+#define GFX_CAPABILITY_WINDOWED (1<<10)
+#define GFX_CAPABILITY_KEYTRANSLATE (1<<11)
 
 #define GFX_DEBUG_POINTER (1<<0)
 #define GFX_DEBUG_UPDATES (1<<1)
@@ -140,6 +141,8 @@ typedef struct _gfx_driver { /* Graphics driver */
 	**    available.
 	** GFX_CAPABILITY_FINE_LINES: Should be set IFF the driver supports drawing
 	**    fine (width 1) lines
+	** GFX_CAPABILITY_WINDOWED: Driver runs in a window and supports a debug
+	**    console running on stdin/stdout
 	*/
 
 	int debug_flags; /* Driver debug flags */
@@ -175,8 +178,8 @@ typedef struct _gfx_driver { /* Graphics driver */
 	** and is used for internal representation of graphical data. The physical
 	** resolution set by the graphics driver may be different for practical
 	** reasons.
-	** Must also set drv->mode, preferrably with the gfx_new_mode() function
-	** specified in gfx_utils.h.
+	** Must also set drv->mode, preferably with the gfx_new_mode() function
+	** specified in gfx_tools.h.
 	*/
 
 	int (*init) (struct _gfx_driver *drv);
@@ -187,8 +190,8 @@ typedef struct _gfx_driver { /* Graphics driver */
 	** most 'natural' to the graphics target. Target implementors have relatively
 	** free reign in choosing the heuristics used to determine the resulting
 	** mode.
-	** Must also set drv->mode, preferrably with the gfx_new_mode() function
-	** specified in gfx_utils.h.
+	** Must also set drv->mode, preferably with the gfx_new_mode() function
+	** specified in gfx_tools.h.
 	*/
 
 	void (*exit) (struct _gfx_driver *drv);
@@ -300,7 +303,7 @@ typedef struct _gfx_driver { /* Graphics driver */
 	** Parameters: (gfx_driver_t *) drv: The affected driver
 	**             (rect_t) src: Source rectangle
 	**             (point_t) dest: Destination point
-	**             (int) buffer: One of GFX_BUFFER_FRONT and GFX_BUFFER_BACK
+	**             (int) buffer: One of GFX_BUFFER_FRONT or GFX_BUFFER_BACK
 	** Returns   : (int) GFX_OK, GFX_ERROR or GFX_FATAL
 	** This function updates either the visual front buffer, or the two back
 	** buffers, by copying the specified source region to the destination
