@@ -1,6 +1,5 @@
 /***************************************************************************
- gfx_operations Copyright (C) 2000,01 Christoph Reichenbach
-
+ gfx_operations.h Copyright (C) 2000,01 Christoph Reichenbach
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -109,6 +108,8 @@ typedef struct {
 
 	gfx_resstate_t *resstate; /* Resource state */
 
+	gfx_pixmap_t *priority_map; /* back buffer priority map (unscaled) */
+	gfx_pixmap_t *static_priority_map; /* static buffer priority map (unscaled) */
 	gfx_pixmap_t *control_map; /* back buffer control map */
 
 
@@ -134,6 +135,10 @@ typedef struct {
 
 	gfx_input_event_t *events;
 
+	gfx_pixmap_t *fullscreen_override; /* An optional override picture which must have unscaled
+					   ** full-screen size, which overrides all other visibility, and
+					   ** which is generally slow */
+
 	gfxr_pic_t *pic, *pic_unscaled; /* The background picture and its unscaled equivalent */
 
 	struct _dirty_rect *dirty_rects; /* Dirty rectangles */
@@ -221,6 +226,13 @@ gfxop_set_clip_zone(gfx_state_t *state, rect_t zone);
 ** Returns   : (int) GFX_OK
 */
 
+int
+gfxop_have_mouse(gfx_state_t *state);
+/* Determines whether a pointing device is attached
+** Parameters: (gfx_state_t *) state: The state to inspect
+** Returns   : (int) zero iff no pointing device is attached
+*/
+
 
 /******************************/
 /* Generic drawing operations */
