@@ -1,7 +1,6 @@
 /***************************************************************************
  kernel.h Copyright (C) 1999 Christoph Reichenbach, TU Darmstadt
 
-
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
  notice and the licensing information contained herein are preserved.
@@ -142,7 +141,7 @@ getHeapUInt16(unsigned char *base, int address)
 
 #define INV_SEL(_object_, _selector_, _noinvalid_) \
   s, _object_,  s->selector_map._selector_, _noinvalid_, funct_nr, argp, argc, __FILE__, __LINE__
-/* Kludge for use with incoke_selector(). Used for compatibility with compilers that can't
+/* Kludge for use with invoke_selector(). Used for compatibility with compilers that can't
 ** handle vararg macros.
 */
 
@@ -221,9 +220,9 @@ is_object(struct _state *s, heap_ptr offset);
 
 /* Functions for internal macro use */
 void
-_SCIkvprintf(FILE *file, char *format, va_list args);
+_SCIkvprintf(FILE *file, const char *format, va_list args);
 void
-_SCIkprintf(FILE *file, char *format, ...);
+_SCIkprintf(FILE *file, const char *format, ...);
 
 
 
@@ -287,12 +286,18 @@ _find_view_priority(struct _state *s, int y);
 ** Returns   : (int) The priority band y belongs to
 */
 
-#define SCI0_VIEW_PRIORITY(y) (((y) < s->priority_first)? 0 : (((y) > s->priority_last)? 14 : 1\
+#define SCI0_VIEW_PRIORITY_14_ZONES(y) (((y) < s->priority_first)? 0 : (((y) >= s->priority_last)? 14 : 1\
 	+ ((((y) - s->priority_first) * 14) / (s->priority_last - s->priority_first))))
 
-#define SCI0_PRIORITY_BAND_FIRST(nr) ((((nr) == 0)? 0 :  \
+#define SCI0_PRIORITY_BAND_FIRST_14_ZONES(nr) ((((nr) == 0)? 0 :  \
         ((s->priority_first) + (((nr)-1) * (s->priority_last - s->priority_first)) / 14)))
 
+#define SCI0_VIEW_PRIORITY(y) (((y) < s->priority_first)? 0 : (((y) >= s->priority_last)? 14 : 1\
+	+ ((((y) - s->priority_first) * 15) / (s->priority_last - s->priority_first))))
+
+#define SCI0_PRIORITY_BAND_FIRST(nr) ((((nr) == 0)? 0 :  \
+        ((s->priority_first) + (((nr)-1) * (s->priority_last - s->priority_first)) / 15)))
+
 #define VIEW_PRIORITY(y) _find_view_priority(s, y)
 #define PRIORITY_BAND_FIRST(nr) _find_priority_band(s, nr)
 
@@ -360,7 +365,6 @@ process_sound_events(struct _state *s); /* Get all sound events, apply their cha
 #define _K_VIEW_SIG_FLAG_IGNORE_ACTOR   0x4000
 #define _K_VIEW_SIG_FLAG_DISPOSE_ME     0x8000
 
-#define _K_VIEW_SIG_FLAG_FREESCI_PRIVATE 0x10000000
 #define _K_VIEW_SIG_FLAG_FREESCI_STOPUPD 0x20000000 /* View has been stop-updated */
 
 
@@ -525,6 +529,9 @@ void kSetNowSeen(struct _state *s, int funct_nr, int argc, heap_ptr argp);
 void kDoAvoider(struct _state *s, int funct_nr, int argc, heap_ptr argp);
 void kFileIO(struct _state *s, int funct_nr, int argc, heap_ptr argp);
 void kMemory(struct _state *s, int funct_nr, int argc, heap_ptr argp);
+void kSort(struct _state *s, int funct_nr, int argc, heap_ptr argp);
+void kAvoidPath(struct _state *s, int funct_nr, int argc, heap_ptr argp);
+void kLock(struct _state *s, int funct_nr, int argc, heap_ptr argp);
 void k_Unknown(struct _state *s, int funct_nr, int argc, heap_ptr argp);
 /* The Unknown/Unnamed kernel function */
 void kstub(struct _state *s, int funct_nr, int argc, heap_ptr argp);
