@@ -1,7 +1,6 @@
 /***************************************************************************
  soundserver.h (C) 1999, 2000 Christoph Reichenbach, TU Darmstadt
 
-
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
  notice and the licensing information contained herein are preserved.
@@ -34,6 +33,7 @@
 #include <sound.h>
 #include <scitypes.h>
 
+/* #define DEBUG_SOUND_SERVER 1 */
 
 #define SOUND_SERVER_TIMEOUT 100000
 /* microseconds until SOUND_COMMAND_TEST fails */
@@ -96,7 +96,7 @@ typedef struct _song_iterator {
 	byte resetflag; /* for 0x4C -- on DoSound StopSound, do we return to start? */
 
 	int (*read_next_command) (struct _song_iterator *self,
-				  byte *buf, int *buf_size);
+				  byte *buf, unsigned int *buf_size);
 	/* Reads the next MIDI operation _or_ delta time
 	** Parameters: (song_iterator_t *) self
 	**             (byte *) buf: The buffer to write to (needs to be able to
@@ -114,7 +114,7 @@ typedef struct _song_iterator {
 	*/
 
 	byte * (*check_pcm) (struct _song_iterator *self,
-			     int *size, int *sample_rate);
+			     unsigned int *size, unsigned int *sample_rate);
 	/* Checks for the presence of a pcm sample
 	** Parameters: (song_iterator_t *) self
 	**             (int *) size: Return variable for the sample size
@@ -146,7 +146,7 @@ typedef struct _song {
 	int pitch[MIDI_CHANNELS]; /* Pitch wheel */
 	int channel_map[MIDI_CHANNELS]; /* Number of HW channels to use */
 
-	int size; /* Song size */
+	unsigned int size; /* Song size */
 	int pos;  /* Current position in song */
 	int loopmark; /* loop position */
 	int fading;   /* Ticks left until faded out, or -1 if not fading */
@@ -166,8 +166,9 @@ typedef struct _song {
 	int file_nr;	/* used for saving */
 	song_iterator_t *it;
 
-	struct _song *next; /* Next song or NULL if this is the last one */
+	int shared;   /* zero if *data is not shared with anyone else */
 
+	struct _song *next; /* Next song or NULL if this is the last one */
 } song_t;
 
 
@@ -327,7 +328,7 @@ song_lib_find_active(songlib_t songlib, song_t *last_played_song);
 */
 
 int
-song_lib_remove(songlib_t songlib, word handle);
+song_lib_remove(songlib_t songlib, word handle, resource_mgr_t *resmgr);
 /* Removes a song from the library
 ** Parameters: (songlib_t) songlib: An existing sound library
 **             (word) handle: Handle of the song to remove
@@ -408,20 +409,26 @@ sound_eq_peek_event(sound_eq_t *queue);
 */
 
 void sci_midi_command(FILE *debugstream, song_t *song, guint8 command, guint8 param,
-		      guint8 param2, int *ccc);
+		      guint8 param2, guint32 other_data, int *ccc, int master_volume);
 /* Performs a regular midi event in the song.
 ** Parameters: (FILE *) debugstream: The stream to write all debug information to
 **             (song_t *) song: The song to play the event from
 **             (guint8) command: MIDI command
 **             (guint8) param: MIDI command parameter 1
 **             (guint8) param2: MIDI command parameter 2
+**             (guint32) other_data: Driver-specific data
 **             (int *) ccc: Cumulative cue counter
+**             (int) master_volume: (duh)
 */
 
 
 /* Sound server has its own cwd (relative to the main program) */
 #define SOUNDSERVER_FLAG_SEPARATE_CWD (1 << 0)
 
+/* Soundserver operates within the same memory context as themain program.
+   Meaning we access the song data off the heap directly */
+#define SOUNDSERVER_FLAG_SHAREDMEM (1 << 1)
+
 #define SOUNDSERVER_INIT_FLAG_REVERSE_STEREO (1 << 0) /* Reverse pan control changes */
 
 #define SOUNDSERVER_FLAG_PIPED (1 << 0)	/* driver is piped (UNIX) */
@@ -555,6 +562,7 @@ extern sound_server_t *global_sound_server; /* current soundserver */
 typedef struct {
 	int suspended;	/* if sound server is suspended */
 	sound_server_t *ss_driver;	/* driver currently being used for sound server */
+	resource_mgr_t *resmgr;
 	int reverse_stereo;	/* reverse stereo setting */
 	byte mute_channel[MIDI_CHANNELS];	/* which channels are muted */
 
@@ -562,6 +570,14 @@ typedef struct {
 
 } sound_server_state_t;
 
+/*
+void
+sci0_event_ss(sound_server_state_t *ss_state);
+*/
+
+void
+sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state);
+
 
 int
 soundsrv_save_state(FILE *debugstream, char *dir, sound_server_state_t *sss);
@@ -651,7 +667,7 @@ snd_srv_process(sound_server_state_t *s, sound_event_t cmd);
 **             (sound_event_t) cmd: The sound command to process
 ** Returns   : (void)
 ** This function must be invoked with 'cmd' filled out whenever a command was
-** received. 
+** received.
 ** s->queue_event() is called from here on occasion, to send back events.
 ** snd_srv_play() should be called immediately after snd_srv_process().
 */
@@ -822,6 +838,13 @@ restore_sound_state(sound_server_state_t *ss_state);
 ** Returns   : (void)
 */
 
+int do_sound(sound_server_state_t *sss, int buff_ss);
+/* Executes main sound loop
+** Paramters: (sound_server_state_t *) sss: Global soundserver state
+**            (int) buff_ss: Indicates if notes fetched should be buffered or
+**                           be played immediately
+*/
+
 #ifdef DEBUG_SOUND_SERVER
 extern int channel_instrument[16];
 extern void print_channels_any(int mapped, sound_server_state_t *ss_state);
