@@ -187,7 +187,7 @@ typedef struct {
 
 	int looper, mover, isBlocked, heading; /* Used in DoAvoider */
 
-	int caller; /* Used for DoBresen */
+	int caller, moveDone, moveSpeed; /* Used for DoBresen */
 
 #ifndef __cplusplus
 /* Compiling as C++ not really supported; let's hope you're only using this on an insignificant sub-part */
@@ -196,6 +196,16 @@ typedef struct {
 /* Same size */
 	int _0_;
 #endif
+
+	int vol;
+	int pri;
+	
+	int min;	/* SMPTE time format */
+	int sec;
+	int frame;
+
+	int dataInc;
+	int size;
 } selector_map_t; /* Contains selector IDs for a few selected selectors */
 
 typedef struct {
@@ -223,7 +233,7 @@ typedef struct {
 
 typedef struct {
 	heap_ptr objp;
-	heap_ptr sendp;
+	heap_ptr sendp; /* Pointer to the object containing the invoked method */
 	heap_ptr pc; /* Not accurate for the TOS element */
 	heap_ptr sp; /* Not accurate for the TOS element */
 	int argc;
@@ -365,6 +375,14 @@ run_vm(struct _state *s, int restoring);
 ** correctly.
 */
 
+void
+vm_handle_fatal_error(struct _state *s, int line, char *file);
+/* Handles a fatal error condition
+** Parameters: (state_t *) s: The state to recover from
+**             (int) line: Source code line number the error occured in
+**             (char *) file: File the error occured in
+*/
+
 
 void
 script_debug(struct _state *s, heap_ptr *pc, heap_ptr *sp, heap_ptr *pp, heap_ptr *objp,
