@@ -34,6 +34,7 @@
 #include <sci_conf.h>
 #include <kdebug.h>
 #include <sys/types.h>
+#include <game_select.h>
 #ifdef HAVE_UNISTD_H
 #  include <unistd.h>
 #endif
@@ -41,10 +42,14 @@
 #  include <sys/wait.h>
 #endif
 
+#if defined(HAVE_SDL) && defined(MACOSX)
+#  include <SDL.h>
+/* On OS X, SDL must #define main to something else in order to function */
+#endif
+
 #ifdef _MSC_VER
 #define extern __declspec(dllimport) extern
 #include <win32/getopt.h>
-#include <crtdbg.h>
 #endif
 
 #ifdef HAVE_READLINE_READLINE_H
@@ -72,14 +77,22 @@
 #ifdef _WIN32
 #  ifdef _MSC_VER
 #    include <direct.h>
-#    define PATH_MAX 255
-#    define sleep(x) timeBeginPeriod(0); Sleep(x); timeEndPeriod(0);
 #    define strcasecmp stricmp
 #  endif
 #  define WIN32_LEAN_AND_MEAN
 #  include <windows.h>
 #endif
 
+#ifdef _DREAMCAST
+#  include <dc.h>
+#  include <selectgame.h>
+#endif
+
+#ifdef ARM_WINCE
+#  include <win32/getopt.h>
+#  include <win32/getopt.c>
+#endif
+
 #ifdef _MSC_VER
 #  define MSVC_FUNCTYPECAST_KLUDGE (void *)
 #else
@@ -102,7 +115,7 @@ check_features()
 
 	helper = 0x100;
 #ifdef __DECC
-	helper = asm("amask %0, %v0", helper);
+	axp_have_mvi = asm("amask %0, %v0", helper);
 #else
 	__asm__ ("amask %1, %0"
 		 : "=r"(axp_have_mvi)
@@ -125,6 +138,7 @@ static gfx_options_t static_gfx_options; /* see below */
 static state_t *gamestate; /* The main game state */
 static gfx_state_t *gfx_state = &static_gfx_state; /* The graphics state */
 static gfx_options_t *gfx_options = &static_gfx_options; /* Graphics options */
+static char *commandline_config_file = NULL;
 
 int
 c_quit(state_t *s)
@@ -239,6 +253,11 @@ init_directories(char *work_dir, char *game_id)
 			chdir(game_id);
 	}
 
+#ifdef _DREAMCAST
+	/* Set dreamcast work dir to ramdisk */
+	fs_chdir("/ram");
+#endif
+
 	getcwd(work_dir, PATH_MAX);
 
 	return 0;
@@ -252,6 +271,7 @@ get_gets_input(void)
 
 	putchar('>');
 
+	fflush(NULL);
 	while (!strchr(input, '\n'))
 		fgets(input, 1024, stdin);
 
@@ -279,7 +299,7 @@ get_gets_input(void)
 
 
 
-static void
+void
 list_graphics_drivers()
 {
 	int i = 0;
@@ -294,6 +314,18 @@ list_graphics_drivers()
 	printf("\n");
 }
 
+static void
+list_pcmout_drivers()
+{
+	int i = 0;
+	while (pcmout_drivers[i]) {
+		if (i != 0)
+			printf(", ");
+		printf(pcmout_drivers[i]->name);
+		i++;
+	}
+	printf("\n");
+}
 
 static void
 list_midiout_drivers()
@@ -340,22 +372,12 @@ list_sound_servers()
 /* Startup and config management                          */
 /**********************************************************/
 
-typedef struct {
-	int script_debug_flag;
-	int scale_x, scale_y, color_depth;
-	int mouse;
-	sci_version_t version;
-	char *gfx_driver_name;
-	char *gamedir;
-        char *midiout_driver_name;
-        char *midi_device_name;
-	char *sound_server_name;
-} cl_options_t;
-
 #define ON 1
 #define OFF 0
 #define DONTCARE -1
 
+static int use_readline = 1;
+
 static char *
 parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame_name)
 {
@@ -370,15 +392,19 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 		{"sci-version", required_argument, 0, 'V'},
 		{"graphics", required_argument, 0, 'g'},
 		{"midiout", required_argument, 0, 'O'},
+		{"pcmout", required_argument, 0, 'P'},
 		{"sound-server", required_argument, 0, 'S'},
 		{"mididevice", required_argument, 0, 'M'},
 		{"version", no_argument, 0, 'v'},
 		{"help", no_argument, 0, 'h'},
+		{"disable-readline", no_argument, &use_readline, 0},
 		{"scale-x", required_argument, 0, 'x'},
 		{"scale-y", required_argument, 0, 'y'},
 		{"color-depth", required_argument, 0, 'c'},
 		{"disable-mouse", no_argument, 0, 'm'},
 		{"list-savegames", no_argument, 0, 'l'},
+		{"show-rooms", no_argument, 0, 's'},
+		{"config-file", required_argument, 0, 'f'},
 		{0,0,0,0}
 	};
 
@@ -392,14 +418,29 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 	cl_options->gfx_driver_name = NULL;
 	cl_options->gamedir = NULL;
 	cl_options->midiout_driver_name = NULL;
+	cl_options->pcmout_driver_name = NULL;
 	cl_options->midi_device_name = NULL;
 	cl_options->sound_server_name = NULL;
 	cl_options->mouse = ON;
+	cl_options->show_rooms = 0;
+#ifdef _DREAMCAST
 
+	/* On the Dreamcast there is no command line, so we don't try to read
+	** any options and just set the savegame_name to NULL and return a
+	** game name of NULL. We set the path to the config file to
+	** "/ram/config", this file is created by the Dreamcast game
+	** selection menu.
+	*/
+
+	savegame_name = NULL;
+	commandline_config_file = sci_strdup("/ram/config");
+	return NULL;
+
+#else /* !_DREAMCAST */
 #ifdef HAVE_GETOPT_LONG
-	while ((c = getopt_long(argc, argv, "lvrhmDd:V:g:x:y:c:M:O:S:", options, &optindex)) > -1) {
+	while ((c = getopt_long(argc, argv, "lvrhmsDd:V:g:x:y:c:M:O:S:P:f:", options, &optindex)) > -1) {
 #else /* !HAVE_GETOPT_LONG */
-	while ((c = getopt(argc, argv, "lvrhmDd:V:g:x:y:c:M:O:S:")) > -1) {
+	while ((c = getopt(argc, argv, "lvrhmsDd:V:g:x:y:c:M:O:S:P:f:")) > -1) {
 #endif /* !HAVE_GETOPT_LONG */
 		switch (c) {
 
@@ -407,6 +448,10 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			cl_options->script_debug_flag = 0;
 			break;
 
+		case 's':
+			cl_options->show_rooms = 1;
+			break;
+
 		case 'D':
 			cl_options->script_debug_flag = 1;
 			break;
@@ -418,6 +463,10 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			cl_options->gamedir = sci_strdup(optarg);
 			break;
 
+		case 'f':
+			commandline_config_file = sci_strdup(optarg);
+			break;
+
 		case 'V': {
 			int major = *optarg - '0'; /* One version digit */
 			int minor = atoi(optarg + 2);
@@ -433,20 +482,25 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			cl_options->gfx_driver_name = sci_strdup(optarg);
 			break;
 		case 'O':
-		        if (cl_options->midiout_driver_name)
-		            free(cl_options->midiout_driver_name);
-		        cl_options->midiout_driver_name = sci_strdup(optarg);
-		        break;
+			if (cl_options->midiout_driver_name)
+				free(cl_options->midiout_driver_name);
+			cl_options->midiout_driver_name = sci_strdup(optarg);
+			break;
+		case 'P':
+			if (cl_options->pcmout_driver_name)
+				free(cl_options->pcmout_driver_name);
+			cl_options->pcmout_driver_name = sci_strdup(optarg);
+			break;
 		case 'M':
-		        if (cl_options->midi_device_name)
-		            free(cl_options->midi_device_name);
-		        cl_options->midi_device_name = sci_strdup(optarg);
-		        break;
+			if (cl_options->midi_device_name)
+				free(cl_options->midi_device_name);
+			cl_options->midi_device_name = sci_strdup(optarg);
+			break;
 		case 'S':
-		        if (cl_options->sound_server_name)
-		            free(cl_options->sound_server_name);
-		        cl_options->sound_server_name = sci_strdup(optarg);
-		        break;
+			if (cl_options->sound_server_name)
+				free(cl_options->sound_server_name);
+			cl_options->sound_server_name = sci_strdup(optarg);
+			break;
 		case '?':
 			/* getopt_long already printed an error message. */
 			exit(1);
@@ -460,7 +514,7 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			break;
 
 		case 'c':
-			cl_options->color_depth = (atoi(optarg) +4) >> 3;
+			cl_options->color_depth = atoi(optarg);
 			break;
 
 		case 'm':
@@ -485,6 +539,9 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			printf("Supported midi 'devices': ");
 			list_midi_devices();
 
+			printf("Supported pcmout drivers: ");
+			list_pcmout_drivers();
+
 			printf("\n");
 			exit(0);
 
@@ -505,8 +562,10 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 			       EXPLAIN_OPTION("--disable-mouse", "-m", "Disable support for pointing device")
 			       EXPLAIN_OPTION("--midiout drv\t", "-Odrv", "use the 'drv' midiout driver")
 			       EXPLAIN_OPTION("--mididevice drv", "-Mdrv", "use the 'drv' midi device (eg mt32 or adlib)")
+			       EXPLAIN_OPTION("--pcmout drv\t", "-Pdrv", "use the 'drv' pcmout driver")
 			       EXPLAIN_OPTION("--sound-server srv", "-Ssrv", "Specifies the asynchronous sound server to use")
 			       EXPLAIN_OPTION("--list-savegames", "-l", "Lists all savegame IDs")
+			       EXPLAIN_OPTION("--show-rooms\t", "-s","Displays room numbers on the game console")
 			       "\n"
 			       "The game name, if provided, must be equal to a game name as specified in the\n"
 			       "FreeSCI config file.\n"
@@ -537,6 +596,7 @@ parse_arguments(int argc, char **argv, cl_options_t *cl_options, char **savegame
 
 	return
 		argv[optind];
+#endif /* !_DREAMCAST */
 }
 
 
@@ -545,11 +605,18 @@ read_config(char *game_name, config_entry_t **conf, int *conf_entries,
 	    sci_version_t *version)
 {
 	int i, conf_nr = 0;
+	char cur_dir[PATH_MAX+1];
 
-	*conf_entries = config_init(conf, NULL);
+	/* config_init() changes cwd, so we restore it after the call */
+	getcwd(cur_dir, PATH_MAX);
+
+	*conf_entries = config_init(conf, commandline_config_file);
+
+	chdir(cur_dir);
 
 	for (i = 1; i < *conf_entries; i++)
-		if (!strcasecmp((*conf)[i].name, game_name)) {
+		if (game_name &&
+		    !strcasecmp((*conf)[i].name, game_name)) {
 			conf_nr = i;
 			*version = (*conf)[i].version;
 		}
@@ -560,10 +627,8 @@ read_config(char *game_name, config_entry_t **conf, int *conf_entries,
 static void
 init_console()
 {
-#if 0
-	con_init();
-	con_init_gfx();
-	con_visible_rows = 1; /* Fool the functions into believing that we *have* a display */
+#ifdef WANT_CONSOLE
+	con_gfx_init();
 #endif
 	con_hook_command(&c_quit, "quit", "", "console: Quits gracefully");
 	con_hook_command(&c_die, "die", "", "console: Quits ungracefully");
@@ -586,16 +651,21 @@ init_console()
 		     "SCI01 priority table debugging flags: 1:Disable, 2:Print on change\n");
 
 	con_passthrough = 1; /* enables all sciprintf data to be sent to stdout */
+}
 
+static void
+init_console_input()
+{
 #ifdef HAVE_READLINE_HISTORY_H
 	using_history(); /* Activate history for readline */
 #endif /* HAVE_READLINE_HISTORY_H */
 
 #ifdef HAVE_READLINE_READLINE_H
-	_debug_get_input = get_readline_input; /* Use readline for debugging input */
-#else /* !HAVE_READLINE_READLINE_H */
-	_debug_get_input = get_gets_input; /* Use gets for debug input */
+	if (use_readline)
+		_debug_get_input = get_readline_input; /* Use readline for debugging input */
+	else
 #endif /* !HAVE_READLINE_READLINE_H */
+		_debug_get_input = get_gets_input; /* Use gets for debug input */
 }
 
 
@@ -665,7 +735,7 @@ init_gfx(config_entry_t *conf, cl_options_t *cl_options, gfx_driver_t *driver, r
 			scale_y = conf->y_scale;
 
 		if (conf->color_depth)
-			color_depth = conf->color_depth >> 3; /* In there it's bpp */
+			color_depth = conf->color_depth; /* In there it's bpp */
 	}
 
 	gfx_state->driver = driver;
@@ -689,6 +759,16 @@ init_gfx(config_entry_t *conf, cl_options_t *cl_options, gfx_driver_t *driver, r
 	if (cl_options->color_depth > 0)
 		color_depth = cl_options->color_depth;
 
+	if (color_depth > 0 && scale_x == 0)
+		scale_x = scale_y = 2; /* Some default setting */
+
+	fprintf(stderr, "cd-conf=%d, cd-cl=%d, cd=%d\n",
+		conf->color_depth, cl_options->color_depth, color_depth);
+
+	/* Convert to bytespp */
+	color_depth = (color_depth + 7) >> 3;
+	fprintf(stderr, "Checking byte depth %d\n", color_depth);
+
 	if (scale_x > 0) {
 
 		if (color_depth > 0) {
@@ -721,10 +801,8 @@ init_gfx(config_entry_t *conf, cl_options_t *cl_options, gfx_driver_t *driver, r
 
 typedef void *old_lookup_funct_t(char *name);
 
-typedef void *lookup_funct_t(char *path, char *name);
 
-
-static void *
+void *
 lookup_driver(lookup_funct_t lookup_func, void explain_func(void),
 	      char *driver_class, char *driver_name, char *path)
 {
@@ -793,7 +871,7 @@ list_savegames(state_t *s)
 int
 main(int argc, char** argv)
 {
-	config_entry_t *active_conf;	/* Active configuration used */
+	config_entry_t *active_conf = NULL;	/* Active configuration used */
 	config_entry_t *confs = {0};	/* Configuration read from config file (if it exists) */
 	cl_options_t cl_options;		/* Command line options */
 	int conf_entries	= -1;		/* Number of config entries */
@@ -805,6 +883,7 @@ main(int argc, char** argv)
 	char *gfx_driver_name			= NULL;
 	char *midiout_driver_name		= NULL;
 	char *midi_device_name			= NULL;
+	char *pcm_driver_name			= NULL;
 	char *game_name	= NULL;
 	char *savegame_name = NULL;
 	sci_version_t version			= 0;
@@ -817,22 +896,30 @@ main(int argc, char** argv)
 
 	game_name = parse_arguments(argc, argv, &cl_options, &savegame_name);
 
+	init_console_input();
+
 	getcwd(startdir, PATH_MAX);
 	script_debug_flag = cl_options.script_debug_flag;
 
-	printf("FreeSCI %s Copyright (C) 1999, 2000, 2001, 2002\n", VERSION);
+	printf("FreeSCI %s Copyright (C) 1999, 2000-2006\n", VERSION);
 	printf(" Dmitry Jemerov, Christopher T. Lansdown, Sergey Lapin, Rickard Lind,\n"
 		   " Carl Muckenhoupt, Christoph Reichenbach, Magnus Reftel, Lars Skovlund,\n"
-		   " Rink Springer, Petr Vyhnak, Solomon Peachy, Matt Hargett, Alex Angas\n"
+		   " Rink Springer, Petr Vyhnak, Solomon Peachy, Matt Hargett, Alex Angas,\n"
+		   " Walter van Niftrik, Ruediger Hanke, Rainer Canavan, Ismail Khatib\n"
+                   " Hugues Valois\n"
 	       "This program is free software. You can copy and/or modify it freely\n"
 	       "according to the terms of the GNU general public license, v2.0\n"
 	       "or any later version, at your option.\n"
 	       "It comes with ABSOLUTELY NO WARRANTY.\n");
 
 
-	if (game_name) {
+#ifdef _DREAMCAST
+	choose_game();
+#endif
 
-		conf_nr = read_config(game_name, &confs, &conf_entries, &version);
+	conf_nr = read_config(game_name, &confs, &conf_entries, &version);
+
+	if (game_name) {
 		active_conf = confs + conf_nr;
 
 		if (!cl_options.gamedir)
@@ -845,9 +932,6 @@ main(int argc, char** argv)
 			}
 	}
 
-	if (cl_options.version)
-		version = cl_options.version;
-
 	if (cl_options.gamedir)
 	{
 		if (chdir(cl_options.gamedir)) {
@@ -857,6 +941,65 @@ main(int argc, char** argv)
 		free(cl_options.gamedir);
 	}
 
+	/* by now, if the user specified a game name or a game directory, the working dir has been changed */
+	/* so if no resource are found in the working dir, invoke the game selection screen */
+	if (conf_entries > 1 && !game_select_resource_found()) {
+		conf_nr = game_select(gamestate, cl_options, gfx_state, gfx_options,
+				      confs, conf_entries, startdir);
+		if (conf_nr < 0)
+			return 0;
+		else
+			active_conf = confs + conf_nr;
+	}
+
+	if (cl_options.version)
+		version = cl_options.version;
+	else {
+		int got_version = version_detect_from_executable(&version);
+		unsigned int code;
+		sci_version_t hash_version;
+		char *game_name = version_guess_from_hashcode(&hash_version, &code);
+
+		if (got_version) {
+			sciprintf("Interpreter version: %d.%03d.%03d (by executable scan)\n",
+				  SCI_VERSION_MAJOR(version),
+				  SCI_VERSION_MINOR(version),
+				  SCI_VERSION_PATCHLEVEL(version));
+
+			if (!game_name)
+				sciprintf("Could not identify game by hash code: %08X\n"
+					  "Please report the preceding two lines and the name of the game you were trying\n"
+					  "to run to the FreeSCI development team to help other users!\n",
+					  code);
+			else if (hash_version != version)
+				sciprintf("UNEXPECTED INCONSISTENCY: Hash code %08X indicates game version %d.%03d.%03d,\n"
+					  "  but analysis of the executable yields %d.%03d.%03d (for game '%s').\n"
+					  "  Please report this!\n"
+					  "Proceeding with result of the executable analysis.\n",
+					  code,
+					  SCI_VERSION_MAJOR(hash_version),
+					  SCI_VERSION_MINOR(hash_version),
+					  SCI_VERSION_PATCHLEVEL(hash_version),
+					  SCI_VERSION_MAJOR(version),
+					  SCI_VERSION_MINOR(version),
+					  SCI_VERSION_PATCHLEVEL(version));
+		} else {
+			if (game_name) {
+				version = hash_version;
+				sciprintf("Interpreter version: %d.%03d.%03d (by hash)\n",
+					  SCI_VERSION_MAJOR(version),
+					  SCI_VERSION_MINOR(version),
+					  SCI_VERSION_PATCHLEVEL(version));
+			} else
+				sciprintf("Could not determine interpreter version; hash code %08X.\n",
+					  code);
+
+		}
+
+		if (game_name)
+			sciprintf("Game identified as '%s'\n", game_name);
+	}
+
 	getcwd(resource_dir, PATH_MAX); /* Store resource directory */
 
 	sciprintf("Loading resources...\n");
@@ -914,10 +1057,8 @@ main(int argc, char** argv)
 		game_name = (char *) gamestate->game_name;
 
 	/* If no game-specific configuration has been read, then read the non-specific config from file */
-	if (!confs) {
-		conf_nr = read_config(game_name, &confs, &conf_entries, &version);
+	if (!active_conf)
 		active_conf = confs + conf_nr;
-	}
 
 	/* gcc doesn't warn about (void *)s being typecast. If your compiler doesn't like these
 	** implicit casts, don't hesitate to typecast appropriately.  */
@@ -926,6 +1067,11 @@ main(int argc, char** argv)
 		free(cl_options.gfx_driver_name);
 	} /* else it's still NULL */
 
+	if (cl_options.pcmout_driver_name)
+		pcmout_driver = old_lookup_driver((old_lookup_funct_t *)pcmout_find_driver,
+						   MSVC_FUNCTYPECAST_KLUDGE list_pcmout_drivers,
+						   "pcmout driver", cl_options.pcmout_driver_name);
+
 	if (cl_options.midiout_driver_name)
 	{
 		midiout_driver = old_lookup_driver((old_lookup_funct_t *)midiout_find_driver,
@@ -963,6 +1109,8 @@ main(int argc, char** argv)
 			midiout_driver = active_conf->midiout_driver;
 		if (!midi_device)
 			midi_device = active_conf->midi_device;
+		if (!pcmout_driver)
+		  pcmout_driver = active_conf->pcmout_driver;
 	}
 
 	if (confs) {
@@ -988,13 +1136,9 @@ main(int argc, char** argv)
 		return 1;
 	}
 
-	if (confs && active_conf->work_dir)
-		gamestate->work_dir = work_dir;
-
-
 	if (chdir(gamestate->work_dir)) {
 		fprintf(stderr,"Error entering working directory '%s'\n",
-			active_conf->work_dir);
+			gamestate->work_dir);
 		exit(1);
 	}
 
@@ -1046,6 +1190,26 @@ main(int argc, char** argv)
 		}
 	}
 
+	/* Configure the pcmout driver */
+	{
+	        pcmout_sample_rate = active_conf->pcmout_rate;
+	        pcmout_stereo = active_conf->pcmout_stereo;
+	        pcmout_buffer_size = active_conf->pcmout_buffer_size;
+
+		if (pcmout_driver) {
+			driver_option_t *option = get_driver_options(active_conf, FREESCI_DRIVER_SUBSYSTEM_PCM, pcmout_driver->name);
+			while (option) {
+				if ((pcmout_driver->set_parameter)(pcmout_driver, option->option, option->value)) {
+					fprintf(stderr, "Fatal error occured in pcmout driver while processing \"%s = %s\"\n",
+						option->option, option->value);
+					exit(1);
+				}
+
+				option = option->next;
+			}
+		}
+	}
+
 	/* Configure the midiout driver */
 	{
 		driver_option_t *option = get_driver_options(active_conf, FREESCI_DRIVER_SUBSYSTEM_MIDIOUT, midiout_driver->name);
@@ -1072,6 +1236,9 @@ main(int argc, char** argv)
 	if (!sound_server)
 		sound_server = sound_server_find_driver(NULL);
 
+	if (cl_options.show_rooms)
+		set_debug_mode(gamestate, 1, "r");
+
 	gamestate->sound_server = sound_server;
 
 	if (gamestate->sound_server) {
@@ -1086,10 +1253,15 @@ main(int argc, char** argv)
 		}
 		sci_sched_yield();
 
+		if (!soundserver_dead) {
 		poly = gamestate->sound_server->command(gamestate, get_msg_value("SOUND_COMMAND_TEST"), 0, 0);
 
 		printf("Sound server reports polyphony %d\n", poly);
 
+		gamestate->sound_server->command(gamestate, get_msg_value("SOUND_COMMAND_SET_VOLUME"), 0, 0xc);
+
+		}
+		
 		gamestate->sound_server->get_event(gamestate); /* Get init message */
 		/* FIXME: memory allocated that is not freed */
 	}
@@ -1111,6 +1283,9 @@ main(int argc, char** argv)
 	       midiout_driver->name, midiout_driver->version);
 	printf("MIDI-device: Using the %s driver %s\n",
 	       midi_device->name, midi_device->version);
+	printf("PCM-out: Using the %s driver %s\n",
+	       pcmout_driver->name, pcmout_driver->version);
+
 	if (sound_server)
 		printf("Sound server: Using the %s sound server %s\n",
 		       sound_server->name, sound_server->version);
@@ -1120,6 +1295,11 @@ main(int argc, char** argv)
 	gamestate->have_mouse_flag = (cl_options.mouse == DONTCARE)?
 		active_conf->mouse : cl_options.mouse;
 
+#ifdef _DREAMCAST
+	if (fs_unlink("/ram/config"))
+		sciprintf("%s, L%d: fs_unlink(\"/ram/config\") failed!\n", __FILE__, __LINE__);
+	dc_retrieve_mirrored(gamestate->game_name);
+#endif
 
 	if (savegame_name)
 		game_restore(&gamestate, savegame_name);
@@ -1128,6 +1308,12 @@ main(int argc, char** argv)
 	if (gamestate->sound_server)
 		gamestate->sound_server->exit(gamestate); /* Shutdown sound daemon first */
 
+#ifdef _DREAMCAST
+	dc_delete_save_files("/ram");
+	dc_delete_temp_file();
+	dc_store_mirrored(gamestate->game_name);
+#endif
+
 	game_exit(gamestate);
 	script_free_engine(gamestate); /* Uninitialize game state */
 	scir_free_resource_manager(resmgr);
@@ -1139,7 +1325,7 @@ main(int argc, char** argv)
 
 	chdir (startdir); /* ? */
 
-#ifdef HAVE_FORK
+#if defined HAVE_FORK && !(ARM_WINCE)
 	printf("Waiting for sound server to die...");
 	wait(NULL); /* Wait for sound server process to die, if neccessary */
 	printf(" OK.\n");
@@ -1149,12 +1335,14 @@ main(int argc, char** argv)
 
 	sci_free(gamestate);
 
-	other_libs_exit();
+	if (commandline_config_file)
+		sci_free(commandline_config_file);
 
 #ifdef WITH_DMALLOC
 	fprintf(stderr,"--- Everything but the two console buffers should have been freed now ---\n");
 	dmalloc_log_unfreed();
-	BREAKPOINT();
+/*	BREAKPOINT(); */
+	((*(int *)NULL) = 42);
 #endif
 	return 0;
 }
