@@ -1,5 +1,5 @@
 /***************************************************************************
- console.c Copyright (C) 1999,2000,01,02 Christoph Reichenbach
+ console.c Copyright (C) 1999 Christoph Reichenbach, TU Darmstadt
 
 
  This program may be modified and copied freely according to the terms of
@@ -38,23 +38,23 @@ state_t *con_gamestate = NULL;
 
 typedef struct
 {
-  int (*command) ();
-  char *name;
-  char *param;
-  char *description;
+	int (*command) ();
+	char *name;
+	char *param;
+	char *description;
 }
 cmd_command_t;
 
 typedef struct
 {
-  union
-  {
-    int *intp;
-    char **charpp;
-  }
-  var;
-  char *name;
-  char *description;
+	union
+	{
+		int *intp;
+		char **charpp;
+	}
+	var;
+	char *name;
+	char *description;
 }
 cmd_var_t;
 
@@ -72,123 +72,118 @@ FILE *con_file = NULL;
 
 static int _lists_need_sorting = 0;
 
-/* Output buffer */
-char con_outputbuf[SCI_CONSOLE_OUTPUT_BUFFER][SCI_CONSOLE_LINE_WIDTH];
-int con_outputbufpos = 0;	/* buffer line */
-int con_outputbufcolumn = 0;
-int con_outputbuflen = 0;
-int con_outputlookback = 0;
-
-
 unsigned int cmd_paramlength;
 cmd_param_t *cmd_params;
 
+static void(*_con_string_callback)(char*) = NULL;
+static void (*_con_pixmap_callback)(gfx_pixmap_t *) = NULL;
+
+
 /********** dmalloc functions **********/
 
 #ifdef WITH_DMALLOC
 int
 c_dm_log_heap (state_t * s)
 {
-  dmalloc_log_heap_map ();
-  return 0;
+	sciprintf("No longer supported.\n");
+	return 0;
 }
 
 int
 c_dm_stats (state_t * s)
 {
-  dmalloc_log_stats ();
-  return 0;
+	dmalloc_log_stats ();
+	return 0;
 }
 
 int
 c_dm_log_unfreed (state_t * s)
 {
-  dmalloc_log_unfreed ();
-  return 0;
+	dmalloc_log_unfreed ();
+	return 0;
 }
 
 int
 c_dm_verify (state_t * s)
 {
-  unsigned long pointer_var;
-  void *ptr;
+	unsigned long pointer_var;
+	void *ptr;
 
-  pointer_var = strtoul (cmd_params[0].str, NULL, 0);
-  ptr = (void *) pointer_var;
+	pointer_var = strtoul (cmd_params[0].str, NULL, 0);
+	ptr = (void *) pointer_var;
 
-  dmalloc_verify (ptr);
+	dmalloc_verify (ptr);
 
-  return 0;
+	return 0;
 }
 
 int
 c_dm_debug (state_t * s)
 {
-  if (cmd_paramlength)
-  {
-    long newval = strtol (cmd_params[0].str, NULL, 0);
+	if (cmd_paramlength) {
+		long newval = strtol (cmd_params[0].str, NULL, 0);
 
-    sciprintf ("Setting dmalloc_debug(%ld)\n", newval);
-    dmalloc_debug (newval);
-  }
-  else
-    sciprintf ("dmalloc_debug is at 0x%lx\n", dmalloc_debug_current ());
-  return 0;
+		sciprintf ("Setting dmalloc_debug(%ld)\n", newval);
+		dmalloc_debug (newval);
+	}
+	else
+		sciprintf ("dmalloc_debug is at 0x%lx\n", dmalloc_debug_current ());
+	return 0;
 }
 
 int
 c_dm_mark (state_t * s)
 {
-  unsigned long mark = dmalloc_mark ();
+	unsigned long mark = dmalloc_mark ();
 
-  dmalloc_message ("------------- MARK 0x%lx ---------------\n", mark);
-  sciprintf ("mark 0x%lx\n", mark);
-  return 0;
+	dmalloc_message ("------------- MARK 0x%lx ---------------\n", mark);
+	sciprintf ("mark 0x%lx\n", mark);
+	return 0;
 }
 
 int
 c_dm_chmark (state_t * s)
 {
-  unsigned long mark = strtoul (cmd_params[0].str, NULL, 0);
-  sciprintf ("Checking mark 0x%lx\n", mark);
-  dmalloc_message ("--- Mark 0x%lx:\n", mark);
-  dmalloc_log_changed (mark, 1, 1, 1);
-  return 0;
+	unsigned long mark = strtoul (cmd_params[0].str, NULL, 0);
+	sciprintf ("Checking mark 0x%lx\n", mark);
+	dmalloc_message ("--- Mark 0x%lx:\n", mark);
+	dmalloc_log_changed (mark, 1, 1, 1);
+	return 0;
 }
 
 int
 c_dm_print (state_t * s)
 {
-  int i;
-  for (i = 0; i < cmd_paramlength; i++)
-    dmalloc_message ("%s\n", cmd_params[i].str);
-  return 0;
+	int i;
+	for (i = 0; i < cmd_paramlength; i++)
+		dmalloc_message ("%s\n", cmd_params[i].str);
+	return 0;
 }
 
 void
 con_init_dmalloc ()
 {
-  con_hook_command (c_dm_log_heap, "dm_log_heap", "",
-		    "Dumps the heap state to\n  the dmalloc output file\n\nUSAGE\n\n  dm_log_heap");
-  con_hook_command (c_dm_stats, "dm_stats", "",
-		    "Prints memory usage stats\n  to the dmalloc output file\n\n  dm_stats");
-  con_hook_command (c_dm_log_unfreed, "dm_log_unfreed", "",
-		    "Prints unfreed pointer\n  information to the dmalloc\n  output file\n\n"
-		    "USAGE\n\n  dm_log_unfreed");
-  con_hook_command (c_dm_verify, "dm_verify", "s",
-		    "Verifies one pointer,\n  prints output to dmalloc file\n\nUSAGE\n\n"
-		    "  dm_verify <ptr>\n  dm_verify 0\n\n  'dm_verify 0' will verify\n  ALL current pointers.\n");
-  con_hook_command (c_dm_debug, "dm_debug", "s*",
-		    "Sets the dmalloc debug\n  state or displays it\n\nUSAGE\n\n  dm_debug <mode>\n  dm_debug");
-  con_hook_command (c_dm_mark, "dm_mark", "",
-		    "Gets a mark describing\n  the current heap state\n\nUSAGE\n\n  dm_mark\n\n"
-		    "  The mark is written to the\n  dmalloc output file and\n  to sci output.\n\nSEE ALSO\n\n  cm_chmark");
-  con_hook_command (c_dm_chmark, "dm_chmark", "s",
-		    "Checks changes in the\n  heap state since a certain\n  mark was retreived\n\n"
-		    "USAGE\n\n  c_dm_chmark <mark>\n\n  Output is written to the\n  dmalloc output file.\n\n  Use dm_mark to retreive a\n"
-		    "  mark.\n\nSEE ALSO\n\n  c_dm_mark");
-  con_hook_command (c_dm_print, "dm_print", "s*",
-		    "Prints something to the\n  dmalloc output.\n\nUSAGE\n\n  dm_print <text>");
+	con_hook_command (c_dm_log_heap, "dm_log_heap", "",
+			  "Dumps the heap state to\n  the dmalloc output file\n\nUSAGE\n\n  dm_log_heap");
+	con_hook_command (c_dm_stats, "dm_stats", "",
+			  "Prints memory usage stats\n  to the dmalloc output file\n\n  dm_stats");
+	con_hook_command (c_dm_log_unfreed, "dm_log_unfreed", "",
+			  "Prints unfreed pointer\n  information to the dmalloc\n  output file\n\n"
+			  "USAGE\n\n  dm_log_unfreed");
+	con_hook_command (c_dm_verify, "dm_verify", "s",
+			  "Verifies one pointer,\n  prints output to dmalloc file\n\nUSAGE\n\n"
+			  "  dm_verify <ptr>\n  dm_verify 0\n\n  'dm_verify 0' will verify\n  ALL current pointers.\n");
+	con_hook_command (c_dm_debug, "dm_debug", "s*",
+			  "Sets the dmalloc debug\n  state or displays it\n\nUSAGE\n\n  dm_debug <mode>\n  dm_debug");
+	con_hook_command (c_dm_mark, "dm_mark", "",
+			  "Gets a mark describing\n  the current heap state\n\nUSAGE\n\n  dm_mark\n\n"
+			  "  The mark is written to the\n  dmalloc output file and\n  to sci output.\n\nSEE ALSO\n\n  cm_chmark");
+	con_hook_command (c_dm_chmark, "dm_chmark", "s",
+			  "Checks changes in the\n  heap state since a certain\n  mark was retreived\n\n"
+			  "USAGE\n\n  c_dm_chmark <mark>\n\n  Output is written to the\n  dmalloc output file.\n\n  Use dm_mark to retreive a\n"
+			  "  mark.\n\nSEE ALSO\n\n  c_dm_mark");
+	con_hook_command (c_dm_print, "dm_print", "s*",
+			  "Prints something to the\n  dmalloc output.\n\nUSAGE\n\n  dm_print <text>");
 }
 #else /* WITH_DMALLOC */
 
@@ -203,242 +198,229 @@ con_init_dmalloc ()
 void
 _cmd_exit (void)
 {
-  if (_cmd_commands)
-    free (_cmd_commands);
+	if (_cmd_commands)
+		free (_cmd_commands);
 
-  if (_cmd_vars)
-    free (_cmd_vars);
+	if (_cmd_vars)
+		free (_cmd_vars);
 }
 
 static int
 _comp_command (const void *a, const void *b)
 {
-  return strcmp (((cmd_command_t *) a)->name, ((cmd_command_t *) b)->name);
+	return strcmp (((cmd_command_t *) a)->name, ((cmd_command_t *) b)->name);
 }
 
 static int
 _comp_var (const void *a, const void *b)
 {
-  return strcmp (((cmd_var_t *) a)->name, ((cmd_var_t *) b)->name);
+	return strcmp (((cmd_var_t *) a)->name, ((cmd_var_t *) b)->name);
 }
 
 void
 con_sort_all (void)
 {
-  if (_cmd_command_count)
-    qsort (_cmd_commands, _cmd_command_count, sizeof (cmd_command_t),
-	   _comp_command);
-  if (_cmd_var_count)
-    qsort (_cmd_vars, _cmd_var_count, sizeof (cmd_var_t), _comp_var);
-  _lists_need_sorting = 0;
+	if (_cmd_command_count)
+		qsort (_cmd_commands, _cmd_command_count, sizeof (cmd_command_t),
+		       _comp_command);
+	if (_cmd_var_count)
+		qsort (_cmd_vars, _cmd_var_count, sizeof (cmd_var_t), _comp_var);
+	_lists_need_sorting = 0;
 }
 
 void
 con_init (void)
 {
-  if (!_cmd_command_mem)
-  {
-    _cmd_commands =
-      sci_realloc(_cmd_commands, sizeof (cmd_command_t) * (_cmd_command_mem = 32));
-    atexit (_cmd_exit);
+	if (!_cmd_command_mem) {
+		_cmd_commands =
+			sci_realloc(_cmd_commands, sizeof (cmd_command_t)
+				    * (_cmd_command_mem = 32));
+		atexit (_cmd_exit);
 
-    /* Hook up some commands */
-    con_hook_command (&c_version, "version", "",
-		      "Displays the version number");
-    con_hook_command (&c_list, "list", "s*",
-		      "Lists various things (try 'list')");
-    con_hook_command (&c_man, "man", "s",
-		      "Gives a short description of something");
-    con_hook_command (&c_print, "print", "s", "Prints an int variable");
-    con_hook_command (&c_set, "set", "si", "Sets an int variable");
-    con_hook_command (&c_size, "size", "si",
-		      "Displays the size of a resource");
-    con_hook_command (&c_dump, "dump", "si", "HexDumps a resource");
-    con_hook_command (&c_hexgrep, "hexgrep", "shh*",
-		      "Searches some resources for a\n"
-		      "  particular sequence of bytes, re-\n  presented"
-		      " as hexadecimal numbers.\n\n"
-		      "EXAMPLES:\n  hexgrep script e8 03 c8 00\n"
-		      "  hexgrep pic.042 fe");
-    con_hook_command (&c_dissectscript, "dissectscript", "i",
-		      "Examines a script.");
+		/* Hook up some commands */
+		con_hook_command (&c_version, "version", "",
+				  "Displays the version number");
+		con_hook_command (&c_list, "list", "s*",
+				  "Lists various things (try 'list')");
+		con_hook_command (&c_man, "man", "s",
+				  "Gives a short description of something");
+		con_hook_command (&c_print, "print", "s", "Prints an int variable");
+		con_hook_command (&c_set, "set", "si", "Sets an int variable");
+		con_hook_command (&c_size, "size", "si",
+				  "Displays the size of a resource");
+		con_hook_command (&c_dump, "dump", "si", "HexDumps a resource");
+		con_hook_command (&c_hexgrep, "hexgrep", "shh*",
+				  "Searches some resources for a\n"
+				  "  particular sequence of bytes, re-\n  presented"
+				  " as hexadecimal numbers.\n\n"
+				  "EXAMPLES:\n  hexgrep script e8 03 c8 00\n"
+				  "  hexgrep pic.042 fe");
+		con_hook_command (&c_dissectscript, "dissectscript", "i",
+				  "Examines a script.");
 
-    con_init_dmalloc ();
+		con_init_dmalloc ();
 
-    con_hook_int (&con_passthrough, "con_passthrough",
-		  "scicon->stdout passthrough");
-  }
+		con_hook_int (&con_passthrough, "con_passthrough",
+			      "scicon->stdout passthrough");
+	}
 }
 
 void
 con_parse (state_t * s, char *command)
 {
-  int quote = 0;		/* quoting? */
-  int done = 0;			/* are we done yet? */
-  int cdone = 0;		/* Done with the current command? */
-  char *paramt;			/* parameter types */
-  char *cmd = (char *) sci_strdup (command);
-  char *_cmd = cmd;
-  int pos = 0;
+	int quote = 0;		/* quoting? */
+	int done = 0;			/* are we done yet? */
+	int cdone = 0;		/* Done with the current command? */
+	char *paramt;			/* parameter types */
+	char *cmd = (command && command[0])? (char *) sci_strdup (command) : sci_strdup(" ");
+	char *_cmd = cmd;
+	int pos = 0;
 
-  if (!_cmd_command_mem)
-    con_init ();
+	if (!_cmd_command_mem)
+		con_init ();
 
-  while (!done)
-  {
-    int cmdnum = -1;		/* command number */
-    int onvar = 1;		/* currently working on a variable? */
-    unsigned int parammem = 0;
-    unsigned int i;
-    cdone = 0;
-    pos = 0;
+	while (!done) {
+		int cmdnum = -1;		/* command number */
+		int onvar = 1;		/* currently working on a variable? */
+		unsigned int parammem = 0;
+		unsigned int i;
+		cdone = 0;
+		pos = 0;
 
-    /*    cmd_params = sci_realloc(cmd_params, parammem); */
-    cmd_paramlength = 0;
+		/*    cmd_params = sci_realloc(cmd_params, parammem); */
+		cmd_paramlength = 0;
 
-    while (*cmd == ' ')
-      cmd++;
+		while (*cmd == ' ')
+			cmd++;
 
-    while (!cdone)
-    {
-      switch (cmd[pos])
-      {
-      case 0:
-	done = 1;
-      case ';':
-	if (!quote)
-	  cdone = 1;
-      case ' ':
-	if (!quote)
-	  cmd[pos] = onvar = 0;
-	break;
-      case '\\':		/* don't check next char for special meaning */
-	memmove (cmd + pos, cmd + pos + 1, strlen (cmd + pos) - 1);
-	break;
-      case '"':
-	quote ^= 1;
-	memmove (cmd + pos, cmd + pos + 1, strlen (cmd + pos));
-	pos--;
-	break;
-      default:
-	if (!onvar)
-	{
-	  onvar = 1;
-	  if (cmd_paramlength == parammem)
-	    cmd_params = sci_realloc(cmd_params,
-				    sizeof (cmd_param_t) * (parammem += 8));
+		while (!cdone) {
+			switch (cmd[pos]) {
+			case 0:
+				done = 1;
+			case ';':
+				if (!quote)
+					cdone = 1;
+			case ' ':
+				if (!quote)
+					cmd[pos] = onvar = 0;
+				break;
+			case '\\':		/* don't check next char for special meaning */
+				memmove (cmd + pos, cmd + pos + 1, strlen (cmd + pos) - 1);
+				break;
+			case '"':
+				quote ^= 1;
+				memmove (cmd + pos, cmd + pos + 1, strlen (cmd + pos));
+				pos--;
+				break;
+			default:
+				if (!onvar) {
+					onvar = 1;
+					if (cmd_paramlength == parammem)
+						cmd_params = sci_realloc(cmd_params,
+									 sizeof (cmd_param_t)
+									 * (parammem += 8));
+			    
+					cmd_params[cmd_paramlength].str = cmd + pos;
 
-	  cmd_params[cmd_paramlength].str = cmd + pos;
-
-	  cmd_paramlength++;
-	}
-	break;
-      }
-      pos++;
-    }
-
-    if (quote)
-      sciprintf ("unbalanced quotes\n");
-    else if (strcmp (cmd, "") != 0)
-    {
-
-      for (i = 0; i < _cmd_command_count; i++)
-	if (strcmp (_cmd_commands[i].name, cmd) == 0)
-	  cmdnum = i;
-
-      if (cmdnum == -1)
-	sciprintf ("%s: not found\n", cmd);
-      else
-      {
-	unsigned int minparams;
-
-	paramt = _cmd_commands[cmdnum].param;
-
-	minparams = strlen (paramt);
-	if ((paramt[0] != 0) && (paramt[strlen (paramt) - 1] == '*'))
-	  minparams -= 2;
-
-	if (cmd_paramlength < minparams)
-	  sciprintf ("%s: needs more than %d parameters\n",
-		     cmd, cmd_paramlength);
-
-	else if ((cmd_paramlength > strlen (paramt))
-		 && ((strlen (paramt) == 0)
-		     || paramt[strlen (paramt) - 1] != '*'))
-	  sciprintf ("%s: too many parameters", cmd);
-
-	else
-	{
-	  int carry = 1;
-	  char paramtype;
-	  int paramtypepos = 0;
-	  char *endptr;
-
-	  for (i = 0; i < cmd_paramlength; i++)
-	  {
-
-	    paramtype = paramt[paramtypepos];
-
-	    if ((paramt[paramtypepos + 1])
-		&& (paramt[paramtypepos + 1] != '*'))
-	      paramtypepos++;	/* seek next param type unless end of string or '*' */
-
-	    switch (paramtype) {
-		    /* Now turn the parameters into variables of the appropriate types,
-		    ** unless they're strings, and store them in the global cmd_params[]
-		    ** structure  */
-	    case 'i':
-	      {
-		char *orgstr = cmd_params[i].str;
-
-		cmd_params[i].val = strtol (orgstr, &endptr, 0);
-
-		if (*endptr != '\0')
-		{
-		  carry = 0;
-		  sciprintf ("%s: '%s' is not an int\n", cmd, orgstr);
-		}
-	      }
-	      break;
-
-	    case 'h':
-	      {
-		char *orgstr = cmd_params[i].str;
-
-		cmd_params[i].val = strtol (orgstr, &endptr, 16);
-
-		if (*endptr != '\0')
-		{
-		  carry = 0;
-		  sciprintf ("%s: '%s' is not a hex number\n", cmd, orgstr);
+					cmd_paramlength++;
+				}
+				break;
+			}
+			pos++;
 		}
 
-		cmd_params[i].val &= 0xff;	/* Clip hex numbers to 0x00 ... 0xff */
-	      }
-	      break;
+		if (quote)
+			sciprintf ("unbalanced quotes\n");
+		else if (strcmp (cmd, "") != 0) {
 
-	    case 's':
-		    break;
+			for (i = 0; i < _cmd_command_count; i++)
+				if (strcmp (_cmd_commands[i].name, cmd) == 0)
+					cmdnum = i;
 
-	    default:
-		    fprintf(stderr, "Internal error: Heap corruption or prior assertion failed:\n"
-			    "Unknown parameter type '%c' for funtion\n", paramtype);
+			if (cmdnum == -1)
+				sciprintf ("%s: not found\n", cmd);
+			else {
+				unsigned int minparams;
 
-	    }
-	  }
+				paramt = _cmd_commands[cmdnum].param;
 
-	  if (carry == 1)
-	    _cmd_commands[cmdnum].command (s);
+				minparams = strlen (paramt);
+				if ((paramt[0] != 0) && (paramt[strlen (paramt) - 1] == '*'))
+					minparams -= 2;
+
+				if (cmd_paramlength < minparams)
+					sciprintf ("%s: needs more than %d parameters\n",
+						   cmd, cmd_paramlength);
+
+				else if ((cmd_paramlength > strlen (paramt))
+					 && ((strlen (paramt) == 0)
+					     || paramt[strlen (paramt) - 1] != '*'))
+					sciprintf ("%s: too many parameters", cmd);
+				else {
+					int carry = 1;
+					char paramtype;
+					int paramtypepos = 0;
+					char *endptr;
+
+					for (i = 0; i < cmd_paramlength; i++) {
+						paramtype = paramt[paramtypepos];
+
+						if ((paramt[paramtypepos + 1])
+						    && (paramt[paramtypepos + 1] != '*'))
+							paramtypepos++;
+						/* seek next param type unless end of string or '*						   ' */
+
+						switch (paramtype) {
+							/* Now turn the parameters into variables of the appropriate types,
+							** unless they're strings, and store them in the global cmd_params[]
+							** structure  */
+						case 'i': {
+							char *orgstr = cmd_params[i].str;
+
+							cmd_params[i].val = strtol (orgstr, &endptr, 0);
+							if (*endptr != '\0') {
+								carry = 0;
+								sciprintf ("%s: '%s' is not an int\n", cmd, orgstr);
+							}
+						}
+							break;
+
+						case 'h': {
+							char *orgstr = cmd_params[i].str;
+
+							cmd_params[i].val = strtol (orgstr, &endptr, 16);
+
+							if (*endptr != '\0') {
+								carry = 0;
+								sciprintf ("%s: '%s' is not a hex number\n", cmd, orgstr);
+							}
+
+							cmd_params[i].val &= 0xff;	/* Clip hex numbers to 0x00 ... 0xff */
+						}
+							break;
+
+						case 's':
+							break;
+
+						default:
+							fprintf(stderr, "Internal error: Heap corruption or prior assertion failed:\n"
+								"Unknown parameter type '%c' for funtion\n", paramtype);
+
+						}
+					}
+
+					if (carry == 1)
+						_cmd_commands[cmdnum].command (s);
+				}
+			}
+		}
+		cmd += pos;
 	}
-      }
-    }
-    cmd += pos;
-  }
 
-  free (_cmd);
-  if (cmd_params)
-    free (cmd_params);
-  cmd_params = NULL;
+	free (_cmd);
+	if (cmd_params)
+		free (cmd_params);
+	cmd_params = NULL;
 
 }
 
@@ -447,91 +429,88 @@ int
 con_hook_command (int command (state_t *), char *name, char *param,
 		  char *description)
 {
-  unsigned int i;
+	unsigned int i;
 
-  if (NULL == name)
-  {
-    sciprintf("console.c: con_hook_command(): NULL passed for name\n");
-    return -1;
-  }
+	if (NULL == name) {
+		sciprintf("console.c: con_hook_command(): NULL passed for name\n");
+		return -1;
+	}
 
-  if (!_cmd_command_mem)
-    con_init ();
-  if (_cmd_command_mem == _cmd_command_count)
-    _cmd_commands = sci_realloc(_cmd_commands,
-			       sizeof (cmd_command_t) *
-			       (_cmd_command_mem <<= 1));
+	if (!_cmd_command_mem)
+		con_init ();
+	if (_cmd_command_mem == _cmd_command_count)
+		_cmd_commands = sci_realloc(_cmd_commands,
+					    sizeof (cmd_command_t) *
+					    (_cmd_command_mem <<= 1));
 
-  if (command == NULL)
-    return 1;
-  for (i = 0; i < _cmd_command_count; i++)
-    if (strcmp (_cmd_commands[i].name, name) == 0)
-      return 1;
+	if (command == NULL)
+		return 1;
+	for (i = 0; i < _cmd_command_count; i++)
+		if (strcmp (_cmd_commands[i].name, name) == 0)
+			return 1;
 
-  if (param == NULL)
-    param = "";
+	if (param == NULL)
+		param = "";
 
-  if (description == NULL)
-    description = "";
+	if (description == NULL)
+		description = "";
 
-  i = 0;
-  while (param[i] != 0)
-  {
-    switch (param[i])
-    {
-    case '*':
-      if (param[i + 1] != 0)
-	return 1;
-      if (i == 0)
-	return 1;
-    case 'h':
-    case 'i':
-    case 's':
-      break;
-    default:
-      return 1;
-    }
-    i++;
-  }
+	i = 0;
+	while (param[i] != 0) {
+		switch (param[i]) {
+		case '*':
+			if (param[i + 1] != 0)
+				return 1;
+			if (i == 0)
+				return 1;
+		case 'h':
+		case 'i':
+		case 's':
+			break;
+		default:
+			return 1;
+		}
+		i++;
+	}
 
-  _cmd_commands[_cmd_command_count].command = command;
-  _cmd_commands[_cmd_command_count].name = name;
-  _cmd_commands[_cmd_command_count].param = param;
-  _cmd_commands[_cmd_command_count].description = description;
+	_cmd_commands[_cmd_command_count].command = command;
+	_cmd_commands[_cmd_command_count].name = name;
+	_cmd_commands[_cmd_command_count].param = param;
+	_cmd_commands[_cmd_command_count].description = description;
 
-  _cmd_command_count++;
+	_cmd_command_count++;
 
-  _lists_need_sorting = 1;
-  return 0;
+	_lists_need_sorting = 1;
+	return 0;
 }
 
 
 int
 con_hook_int (int *pointer, char *name, char *description)
 {
-  unsigned int i;
+	unsigned int i;
 
-  if (_cmd_var_mem == _cmd_var_count)
-    _cmd_vars = sci_realloc(_cmd_vars,
-			   sizeof (cmd_var_t) * (_cmd_var_mem += 16));
+	if (_cmd_var_mem == _cmd_var_count)
+		_cmd_vars = sci_realloc(_cmd_vars,
+					sizeof (cmd_var_t) * (_cmd_var_mem += 16));
 
-  if (pointer == NULL)
-    return 1;
-  for (i = 0; i < _cmd_var_count; i++)
-    if (strcmp (_cmd_vars[i].name, name) == 0)
-      return 1;
+	if (pointer == NULL)
+		return 1;
+	for (i = 0; i < _cmd_var_count; i++)
+		if (strcmp (_cmd_vars[i].name, name) == 0)
+			return 1;
 
-  if (description == NULL)
-    description = "";
+	if (description == NULL)
+		description = "";
 
-  _cmd_vars[_cmd_var_count].var.intp = pointer;
-  _cmd_vars[_cmd_var_count].name = name;
-  _cmd_vars[_cmd_var_count].description = description;
+	_cmd_vars[_cmd_var_count].var.intp = pointer;
+	_cmd_vars[_cmd_var_count].name = name;
+	_cmd_vars[_cmd_var_count].description = description;
 
-  _cmd_var_count++;
+	_cmd_var_count++;
 
-  _lists_need_sorting = 1;
-  return 0;
+	_lists_need_sorting = 1;
+	return 0;
 }
 
 
@@ -542,95 +521,72 @@ con_hook_int (int *pointer, char *name, char *description)
 
 
 int
-sciprintf (char *fmt, ...)
+sciprintf (const char *fmt, ...)
 {
-  va_list argp;
-  size_t bufsize = 256;
-  unsigned int i;
-  char *buf 	= (char *) sci_malloc (bufsize);
-  char *mbuf	= NULL;
+	va_list argp;
+	size_t bufsize = 256;
+	unsigned int i;
+	char *buf;
 
-  if (NULL == fmt)
-  {
-    fprintf(stderr, "console.c: sciprintf(): NULL passed for parameter fmt\n");
-    return -1;
-  }
+	if (NULL == fmt) {
+		fprintf(stderr, "console.c: sciprintf(): NULL passed for parameter fmt\n");
+		return -1;
+	}
 
-  if (NULL == buf)
-  {
-    fprintf(stderr, "console.c: sciprintf(): malloc failed for buf\n");
-    return -1;
-  }
+	buf = (char *) sci_malloc (bufsize);
 
-  va_start (argp, fmt);
-  while ((i = vsnprintf (buf, bufsize - 1, fmt, argp)) == -1
-	 || (i >= bufsize - 2))
-  {
-    /* while we're out of space... */
-    va_end (argp);
-    va_start (argp, fmt);	/* reset argp */
+	va_start (argp, fmt);
+	while ((i = vsnprintf (buf, bufsize - 1, fmt, argp)) == -1
+	       || (i >= bufsize - 2)) {
+		/* while we're out of space... */
+		va_end (argp);
+		va_start (argp, fmt);	/* reset argp */
 
-    free (buf);
-    buf = (char *) sci_malloc (bufsize <<= 1);
-  }
-  va_end (argp);
+		free (buf);
+		buf = (char *) sci_malloc (bufsize <<= 1);
+	}
+	va_end (argp);
 
-  if (con_passthrough)
-    printf ("%s", buf);
-  if (con_file)
-    fprintf (con_file, "%s", buf);
+	if (con_passthrough)
+		printf ("%s", buf);
+	if (con_file)
+		fprintf (con_file, "%s", buf);
 
-  mbuf = buf;
+	
+	if (_con_string_callback)
+		_con_string_callback(buf);
+	else
+		free(buf);
 
-  i = strlen (mbuf);
+	return 0;
+}
 
-  while (i > 0)
-  {
-    char *seekerpt = strchr (mbuf, '\n');
-    int seeker =
-      seekerpt ? seekerpt - mbuf : SCI_CONSOLE_LINE_WIDTH -
-      con_outputbufcolumn;
-    if (seeker >= SCI_CONSOLE_LINE_WIDTH - con_outputbufcolumn)
-      seeker = (i > SCI_CONSOLE_LINE_WIDTH - con_outputbufcolumn) ?
-	SCI_CONSOLE_LINE_WIDTH - con_outputbufcolumn : i;
+void
+con_set_string_callback(void(*callback)(char *))
+{
+	_con_string_callback = callback;
+}
 
-    memcpy (con_outputbuf[con_outputbufpos] + con_outputbufcolumn, mbuf,
-	    seeker);
-    if (seeker + con_outputbufcolumn < SCI_CONSOLE_LINE_WIDTH)
-      con_outputbuf[con_outputbufpos][seeker + con_outputbufcolumn] = 0;
+void
+con_set_pixmap_callback(void(*callback)(gfx_pixmap_t *))
+{
+	_con_pixmap_callback = callback;
+}
 
-    con_outputbufcolumn += seeker;
+int
+con_can_handle_pixmaps()
+{
+	return _con_pixmap_callback != NULL;
+}
 
-    if (con_outputbufcolumn == SCI_CONSOLE_LINE_WIDTH)
-    {
-      con_outputbufpos++;
-      con_outputbufcolumn = 0;
-    }
-
-    if (seekerpt && (seeker == seekerpt - mbuf))
-    {
-      con_outputbufcolumn = 0;
-      con_outputbufpos++;
-      mbuf++;
-      i--;
-    }
-
-    mbuf += seeker;
-    i -= seeker;
-
-    if (con_outputbufpos >= SCI_CONSOLE_OUTPUT_BUFFER)
-    {
-      con_outputbufpos -= SCI_CONSOLE_OUTPUT_BUFFER;
-      con_outputbuflen = SCI_CONSOLE_OUTPUT_BUFFER;
-    }
-  }
-  if (con_outputbuflen < con_outputbufpos)
-    con_outputbuflen = con_outputbufpos;
-
-  con_outputlookback = con_outputbufpos;
-  free (buf);
-
-  return 1;
+int
+con_insert_pixmap(gfx_pixmap_t *pixmap)
+{
+	if (_con_pixmap_callback)
+		_con_pixmap_callback(pixmap);
+	else
+		return 1;
+	return 0;
 }
 
 
@@ -649,8 +605,11 @@ open_console_file (char *filename)
 #else
   con_file = fopen (filename, "w");
 #endif
-}
 
+  if (NULL == con_file)
+    fprintf(stderr, "console.c: open_console_file(): Could not open output file %s\n", filename);
+
+}
 
 void
 close_console_file ()
@@ -985,7 +944,7 @@ c_hexgrep (state_t * s)
   for (; resnr <= resmax; resnr++)
     if ((script = scir_find_resource(s->resmgr, restype, resnr, 0)))
     {
-      int seeker = 0, seekerold = 0;
+      unsigned int seeker = 0, seekerold = 0;
       int comppos = 0;
       int output_script_name = 0;
 
