@@ -1,6 +1,5 @@
 /***************************************************************************
- midi_mt32.c Copyright (C) 2000,2001,02 Rickard Lind, Solomon Peachy
-
+ midi_mt32.c Copyright (C) 2000,2001 Rickard Lind, Solomon Peachy
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -28,6 +27,7 @@
 
 #ifdef _WIN32
 #	include <windows.h>
+#   include <sci_win32.h>
 #endif
 
 int midi_mt32_poke(guint32 address, guint8 *data, unsigned int n);
@@ -40,14 +40,13 @@ int midi_mt32_patch001_type1_length(guint8 *data, unsigned int length);
 int midi_mt32_sysex_delay();
 int midi_mt32_volume(guint8 volume);
 int midi_mt32_reverb(short param);
-int midi_mt32_event(guint8 command, guint8 note, guint8 velocity);
+int midi_mt32_event(guint8 command, guint8 note, guint8 velocity, guint32 delta);
 int midi_mt32_allstop(void);
 
-static guint8 *data;
-static unsigned int length;
 static int type;
 static guint8 sysex_buffer[266] = {0xF0, 0x41, 0x10, 0x16, 0x12};
 static guint8 default_reverb;
+static char shutdown_msg[20];
 
 static gint8 patch_map[128] = {
   0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
@@ -118,8 +117,8 @@ static gint8 rhythmkey_map[128] = {
 /* timbre, volume, panpot, reverb.  keys 24-87 (64 keys)*/
 static guint8 default_rhythm_keymap[256] = { /* MT-32 default */
   0x7f,0x64,7,1,  0x7f,0x64,7,1,  0x7f,0x64,7,1, 0x7f,0x64,7,1, /* 24-27 */
-  0x7f,0x64,7,1,  0x7f,0x64,7,1,  0x7f,0x64,7,1, 0x7f,0x64,7,1, 
-  0x7f,0x64,7,1,  0x7f,0x64,7,1,  0x7f,0x64,7,1, 0x40,0x64,7,1,  
+  0x7f,0x64,7,1,  0x7f,0x64,7,1,  0x7f,0x64,7,1, 0x7f,0x64,7,1,
+  0x7f,0x64,7,1,  0x7f,0x64,7,1,  0x7f,0x64,7,1, 0x40,0x64,7,1,
   0x40,0x64,7,1,  0x4a,0x64,6,1,  0x41,0x64,7,1, 0x4b,0x64,8,1,
   0x45,0x64,6,1,  0x44,0x64,11,1, 0x46,0x64,6,1, 0x44,0x64,11,1,
   0x5d,0x64,6,1,  0x43,0x64,8,1,  0x47,0x64,6,1, 0x43,0x64,8,1,
@@ -145,7 +144,7 @@ static guint8 default_partial_reserve[9] = {  /* MT-32 DEFAULT */
   gint8 keyshift;
   gint8 volume_adjust;
   guint8 velocity_map_index;
-  guint8 
+  guint8
 } channel[16]; */
 
 static struct {
@@ -158,7 +157,7 @@ static struct {
 /* send default rhythm map and reserve */
 int midi_mt32_defaults(guint8 volume, guint8 reverb) {
   printf("MT-32: Writing Default Rhythm key map\n");
-  midi_mt32_poke(0x030110, default_rhythm_keymap, 256);  
+  midi_mt32_poke(0x030110, default_rhythm_keymap, 256);
 
   printf("MT-32: Writing Default Partial Reserve\n");
   midi_mt32_poke(0x100004, default_partial_reserve, 9);
@@ -167,7 +166,7 @@ int midi_mt32_defaults(guint8 volume, guint8 reverb) {
     mt32_reverb[0].mode = 0;
     mt32_reverb[0].time = 5;
     mt32_reverb[0].level = 3;
-    default_reverb = 0;    
+    default_reverb = 0;
 
     printf("MT-32: Setting up default reverb levels\n");
     midi_mt32_reverb(default_reverb);
@@ -181,7 +180,7 @@ int midi_mt32_defaults(guint8 volume, guint8 reverb) {
   return 0;
 }
 
-int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
+int midi_mt32_open(guint8 *data, unsigned int length)
 {
   guint8 unknown_sysex[6] = {0x16, 0x16, 0x16, 0x16, 0x16, 0x16};
   guint8 i, memtimbres;
@@ -189,8 +188,6 @@ int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
 
   if (midiout_open() < 0)
     return -1;
-  data = data_ptr;
-  length = data_length;
 
   midi_mt32_allstop();
 
@@ -202,6 +199,9 @@ int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
     printf("MT-32: Displaying Text: \"%.20s\"\n", data + 20);
     midi_mt32_poke(0x200000, data + 20, 20);
 
+    /* Cache a copy of the shutdown message */
+    memcpy(shutdown_msg, data + 40, 20);
+
     /* Write Patches (48 or 96) */
     memtimbres = data[491];
     block2 = (memtimbres * 246) + 492;
@@ -211,7 +211,7 @@ int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
 	data[block2] == 0xAB &&
 	data[block2 + 1] == 0xCD) {
       printf("MT-32: Writing Patches #33 - #64\n");
-      midi_mt32_poke_gather(0x050200, data + 363, 128, data + block2 + 2, 128);  
+      midi_mt32_poke_gather(0x050200, data + 363, 128, data + block2 + 2, 128);
       printf("MT-32: Writing Patches #65 - #96\n");
       midi_mt32_poke(0x050400, data + block2 + 130, 256);
       block3 = block2 + 386;
@@ -231,7 +231,7 @@ int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
 	data[block3] == 0xDC &&
 	data[block3 + 1] == 0xBA) {
       printf("MT-32: Writing Rhythm key map\n");
-      midi_mt32_poke(0x030110, data + block3 + 2, 256);  
+      midi_mt32_poke(0x030110, data + block3 + 2, 256);
       printf("MT-32: Writing Partial Reserve\n");
       midi_mt32_poke(0x100004, data + block3 + 258, 9);
     } else {
@@ -264,7 +264,7 @@ int midi_mt32_open(guint8 *data_ptr, unsigned int data_length)
     memcpy(rhythmkey_map, data + 384, 128);
 
     printf("MT-32: Setting up reverb levels\n");
-    default_reverb = data[0x3e]; 
+    default_reverb = data[0x3e];
     memcpy(mt32_reverb,data+ 0x4a, 3 * 11);
     midi_mt32_reverb(default_reverb);
 
@@ -285,8 +285,8 @@ int midi_mt32_close(void)
 {
   midi_mt32_allstop();
   if (type == 0) {
-    printf("MT-32: Displaying Text: \"%.20s\"\n", data + 40);
-    midi_mt32_poke(0x200000, data + 40, 20);
+    printf("MT-32: Displaying Text: \"%.20s\"\n", shutdown_msg);
+    midi_mt32_poke(0x200000, shutdown_msg, 20);
   }
   return midiout_close();
 }
@@ -300,11 +300,11 @@ int midi_mt32_volume(guint8 volume)
   return 0;
 }
 
-int midi_mt32_allstop(void) 
+int midi_mt32_allstop(void)
 {
   int i;
-  for (i = 0; i < 16; i++) 
-    midi_mt32_event(0xb0 | i, 0x7b, 0x00);
+  for (i = 0; i < 16; i++)
+    midi_mt32_event((guint8)(0xb0 | i), 0x7b, 0x00, 0);
 
   return 0;
 }
@@ -328,6 +328,7 @@ int midi_mt32_reverb(short param)
   return 0;
 }
 
+/*
 int midi_mt32_noteoff(guint8 channel, guint8 note, guint8 velocity)
 {
   return midi_mt32_event(channel | 0x80, note, velocity);
@@ -344,8 +345,9 @@ int midi_mt32_noteon(guint8 channel, guint8 note, guint8 velocity)
 
   return midi_mt32_event(channel | 0x90, note, velocity);
 }
+*/
 
-int midi_mt32_event(guint8 command, guint8 note, guint8 velocity)
+int midi_mt32_event(guint8 command, guint8 note, guint8 velocity, guint32 delta)
 {
   guint8 buffer[3] = {0,0,0};
 
@@ -353,18 +355,18 @@ int midi_mt32_event(guint8 command, guint8 note, guint8 velocity)
   buffer[1] = note;
   buffer[2] = velocity;
 
-  return midiout_write_event(buffer, 3);
+  return midiout_write_event(buffer, 3, delta);
 
 }
 
-int midi_mt32_event2(guint8 command, guint8 param)
+int midi_mt32_event2(guint8 command, guint8 param, guint32 delta)
 {
   guint8 buffer[2] = {0,0};
 
   buffer[0] = command;
   buffer[1] = param;
 
-  return midiout_write_event(buffer, 2);
+  return midiout_write_event(buffer, 2, delta);
 }
 
 int midi_mt32_poke(guint32 address, guint8 *data, unsigned int count)
@@ -384,9 +386,9 @@ int midi_mt32_poke(guint32 address, guint8 *data, unsigned int count)
   sysex_buffer[count + 8] = checksum & 0x7F;
   sysex_buffer[count + 9] = 0xF7;
 
-  i = midiout_write_block(sysex_buffer, count + 10);
-  midiout_flush();
-  midi_mt32_sysex_delay();  
+  i = midiout_write_block(sysex_buffer, count + 10, 0);
+  midiout_flush(0);
+  midi_mt32_sysex_delay();
 
   return i;
 
@@ -412,8 +414,8 @@ int midi_mt32_poke_gather(guint32 address, guint8 *data1, unsigned int count1,
   sysex_buffer[count1 + count2 + 8] = checksum & 0x7F;
   sysex_buffer[count1 + count2 + 9] = 0xF7;
 
-  i = midiout_write_block(sysex_buffer, count1 + count2 + 10);
-  midiout_flush();
+  i = midiout_write_block(sysex_buffer, count1 + count2 + 10, 0);
+  midiout_flush(0);
   midi_mt32_sysex_delay();
   return i;
 }
@@ -425,11 +427,11 @@ int midi_mt32_write_block(guint8 *data, unsigned int count)
 
   while (i < count) {
     if ((data[i] == 0xF0) && (i != block_start)) {
-      midiout_write_block(data + block_start, i - block_start);
+      midiout_write_block(data + block_start, i - block_start, 0);
       block_start = i;
     }
     if (data[i] == 0xF7) {
-      midiout_write_block(data + block_start, i - block_start + 1);
+      midiout_write_block(data + block_start, i - block_start + 1, 0);
       midi_mt32_sysex_delay();
       block_start = i + 1;
     }
@@ -437,7 +439,7 @@ int midi_mt32_write_block(guint8 *data, unsigned int count)
   }
   if (count >= block_start)
   {
-    if (midiout_write_block(data + block_start, count - block_start) != (count - block_start))
+    if (midiout_write_block(data + block_start, count - block_start, 0) != (count - block_start))
 	{
 		fprintf(stderr, "midi_mt32_write_block(): midiout_write_block failed!\n");
 		return 1;
@@ -506,12 +508,8 @@ int midi_mt32_sysex_delay()
   /* Under Win32, we won't get any sound, in any case... */
 #ifdef HAVE_USLEEP
   usleep(320 * 63); /* One MIDI byte is 320us, 320us * 63 > 20ms */
-#  else
-#    ifdef _WIN32
-  Sleep(((320 * 63) / 1000) + 1);
 #    else
   sleep(1);
-#  endif
 #endif
   return 0;
 }
