@@ -1,6 +1,6 @@
 /***************************************************************************
- midi.c (C) 1999,2001,02 Christoph Reichenbach, TU Darmstadt
-        (C) 1999-2000,2002 Rickard Lind
+ midi.c (C) 1999,2001 Christoph Reichenbach, TU Darmstadt
+        (C) 1999-2000 Rickard Lind
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -914,7 +914,7 @@ map_MIDI_instruments(resource_mgr_t *resmgr)
   guint8 group, number, keyshift, finetune, bender_range;
   guint8 *patchpointer;
   guint32 pos;
-  FILE *logfile;
+/*  FILE *logfile; */
   resource_t *patch1;
 
   patch1 = scir_find_resource(resmgr, sci_patch, 1, 0);
