@@ -1,5 +1,5 @@
 /***************************************************************************
- polled_ss.c Copyright (C) 1999, 2000, 01, 02 Christoph Reichenbach
+ polled_ss.c Copyright (C) 1999 Christoph Reichenbach, TU Darmstadt
 
  This program may be modified and copied freely according to the terms of
  the GNU general public license (GPL), as long as the above copyright
@@ -37,6 +37,8 @@
 #include <soundserver.h>
 #include <midi_device.h>
 #include <sys/types.h>
+#include <pcmout.h>
+
 #ifdef HAVE_SYS_SOUNDCARD_H
 #  include <sys/stat.h>
 #  include <fcntl.h>
@@ -117,7 +119,7 @@ sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state)
 			/* Just played a note (see bottom of the big outer loop), so we
 			** reset the timer */
 
-			ticks_to_wait = 0; /* Number of ticks until the next song is played */
+			ticks_to_wait = 0; /* Number of ticks until the next note is played */
 
 			old_songpos = ss_state->current_song->pos;	/* Keep a backup of the current song position,
 										** in case we are interrupted and want to resume
@@ -184,7 +186,7 @@ sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state)
 						fprintf(debug_stream, "Initialising song handle %04x: ", event.handle);
 #endif
 						if (modsong) { /* If the song already exists in cache... */
-							int lastmode = song_lib_remove(ss_state->songlib, (word)event.handle); /* remove it */
+							int lastmode = song_lib_remove(ss_state->songlib, (word)event.handle, ss_state->resmgr); /* remove it */
 							if (lastmode == SOUND_STATUS_PLAYING) {
 								newsong = ss_state->songlib[0];	/* If we just retreived the current song,
 														** we'll have to reset the queue for
@@ -238,7 +240,7 @@ sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state)
 									event.handle, sample_rate, pcm_size,
 									(sample_rate)? ((pcm_size * 1.0) / (sample_rate * 1.0)) : 0);
 								fd = open("/dev/dsp", O_WRONLY);
-								if (!fd) {
+								if (!IS_VALID_FD(fd)) {
 									perror("SndSrv: While opening /dev/dsp");
 								} else {
 									tmp = AFMT_U8;
@@ -289,7 +291,7 @@ sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state)
 						fprintf(debug_stream, "Disposing handle %04x (value %04x)\n", event.handle, event.value);
 #endif
 						if (modsong) {
-							int lastmode = song_lib_remove(ss_state->songlib, (word)event.handle);
+							int lastmode = song_lib_remove(ss_state->songlib, (word)event.handle, ss_state->resmgr);
 							if (lastmode == SOUND_STATUS_PLAYING) {
 								newsong = ss_state->songlib[0]; /* Force song detection to start with the highest priority song */
 								global_sound_server->queue_event(event.handle, SOUND_SIGNAL_FINISHED, 0);
@@ -437,6 +439,7 @@ sci0_polled_ss(int reverse_stereo, sound_server_state_t *ss_state)
 
 					case SOUND_COMMAND_SHUTDOWN: {
 						fprintf(debug_stream,"Sound server: Received shutdown signal\n");
+						pcmout_close();
 						midi_close();
 						song_lib_free(ss_state->songlib);
 
@@ -552,7 +555,9 @@ ss_state->current_song->pos);
 						(unsigned char)command,
 						param,
 						param2,
-						&ss_state->sound_cue);
+						ticks_to_wait,
+						&ss_state->sound_cue,
+						ss_state->master_volume);
 				}
 			}
 
@@ -574,7 +579,7 @@ ss_state->current_song->pos);
 				for (i = 0; i < MIDI_CHANNELS; i++) {
 					if (newsong->instruments[i])
 						midi_event2((guint8)(MIDI_INSTRUMENT_CHANGE | i),
-							(guint8)newsong->instruments[i]);
+							(guint8)newsong->instruments[i], 0);
 				}
 			}
 			if (ss_state->current_song)	/* Muting active song: Un-play the last note/MIDI event */
